unit Globals;

{
  BLUE HELL software.

  Copyright (C) 1999 .. 2012, Jan Punter.

  Notes :

    midi componenets needed : http://sourceforge.net/projects/dmidi/
    hires timer needed      : source code included ... I may have stolen that somewhere without
                              acknowleging the source .. somewhere in 1999 or so ...
                              It sure looks a lot like : http://www.programmersheaven.com/download/2818/1/ZipView.aspx
                              but details as to the origin seem absent there too.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

}

interface

Uses

  SysUtils, Windows, Messages, StdCtrls, Forms, Dialogs, Registry;


{$DEFINE DEBUG}
{$UNDEF DEBUG}

Const

  MainTimeResolution = 20; // ms

  // Midi definitions

  CNoteOff         = $8;
  CNoteOn          = $9;
  CKeyPressure     = $a;
  CControlChange   = $b;
  CProgramChange   = $c;
  CChannelPressure = $d;
  CPitchBend       = $e;
  CSysEx           = $f;

  UnityValue       =    1000;
  ShortestTime     =      60; // ms
  LongestTime      = 6000000; // ms = 6000 s = 100 m = 1:40:00

// Our name and path stuff

  Function ApplicationPath    : String;
  Function ApplicationPathURL : String;
  Function ApplicationFileName: String;
  Function IniFileName        : String;
  Function LogFileName        : String;
  Function FileNameToURL( Const aFileName: String): String;

  Function ProgramAssociationFor( anExt: String): String;

  Function  BoolToStr       ( aValue      : Boolean): String;
  Function  ByteToHex       ( aValue: Cardinal): String;
  Function  WordToHex       ( aValue: Cardinal): String;
  Function  LongToHex       ( aValue: Cardinal): String;
  Function  HexToByte       ( Const aValue: String): Cardinal;
  Function  HexToWord       ( Const aValue: String): Cardinal;
  Function  HexToLong       ( Const aValue: String): Cardinal;
  Function  ByteToAscii     ( aValue: Byte): Char;
  Function  Indent          ( anInd: Integer): String;
  Function  TrimStr         ( Const S: String; aLength: Integer): String;
  Function  ExtendStr       ( Const S: String; aLength: Integer): String;

  Function  IntegerToNoteStr( aNoteValue: Integer): String;

// Debug support

{$IFDEF DEBUG}
  Procedure ClearDebugTiming;
  Procedure WriteDebugStr( Const aString: String);
{$ENDIF}


// System stuff

  Function  Clip( var aVal: Integer; aLow, aHigh: Integer): Integer;
  Function  CleanFileName( Const aFileName : String): String;
  // Returns filename only, path and extension are stripped off.


// Miscellaneous

  Procedure NotImplemented( Const aMsg: String);
  Procedure Noop;



implementation


// Our name and path stuff

Var

  strApplicationPath     : String;
  strApplicationPathURL  : String;
  strApplicationFileName : String;
  strIniFileName         : String;
  strLogFileName         : String;
  strIndent              : String;

  Function ApplicationPath: String;
  Begin
    Result := strApplicationPath;
  End;

  Function ApplicationPathURL : String;
  Begin
    Result := strApplicationPathURL;
  End;

  Function ApplicationFileName: String;
  Begin
    Result := strApplicationFileName;
  End;

  Function IniFileName: String;
  Begin
    Result := strIniFileName;
  End;

  Function LogFileName: String;
  Begin
    Result := strLogFileName;
  End;

  Function FileNameToURL( Const aFileName: String): String;
  Var
    i : Integer;
  Begin
    Result := 'file:///';
    For i := 1 To Length( aFileName) Do
      Case aFileName[ i] Of
        ':' : Result := Result + '|';
        '\' : Result := Result + '/';
        Else  Result := Result + aFileName[ i];
      End;
  End;


  Function ProgramAssociationFor( anExt: String): String;
  Var
    Reg : TRegistry;
    S   : String;
  Begin
    S := '';
    Reg := TRegistry.Create;
    Reg.RootKey := HKEY_CLASSES_ROOT;
    If Reg.OpenKey( '.' + anExt + '\Shell\open\command', false)
    Then Begin
      S := Reg.ReadString( '');
      Reg.CloseKey;
    End
    Else Begin
      If Reg.OpenKey( '.' + anExt, False)
      Then Begin
        S := Reg.ReadString( '');
        Reg.CloseKey;
        If S <> ''
        Then Begin
          If Reg.OpenKey( S + '\Shell\Open\Command', False)
          Then S := Reg.ReadString( '');
          Reg.CloseKey;
        End;
      End;
    End;
    If Pos( '%', S) > 0
    Then Begin
      If Pos( '%1', S) > 0
      Then Delete( S, Pos( '%1', S), 2)
      Else Delete( S, Pos( '%', S), 1);
    End;
    While (
        ( Length( S) > 0) And
        (
          ( S[ Length( S)] = ' ') Or
          ( S[ Length( S)] = '"')
        )
      ) Do
      Delete( S, Length( S), 1);
    Result := S;
  End;


  Function  BoolToStr( aValue : Boolean): String;
  Begin
    Case aValue Of
      False : Result := 'False';
      True  : Result := 'True';
    End;
  End;

  Function  ByteToHex( aValue: Cardinal): String;
  Begin
    Result := Format( '%0.2x', [ aValue And $00ff]);
  End;

  Function  WordToHex( aValue: Cardinal): String;
  Begin
    Result := Format( '%0.4x', [ aValue And $ffff]);
  End;

  Function  LongToHex( aValue: Cardinal): String;
  Begin
    Result := Format( '%0.4x.%0.4x', [ aValue And $ffff0000 Shr 16, aValue And $ffff]);
  End;

  Function  HexToByte( Const aValue: String): Cardinal;
  Begin
    Result := StrToInt( Format( '$%s', [ aValue]));
  End;

  Function  HexToWord( Const aValue: String): Cardinal;
  Begin
    Result := StrToInt( Format( '$%s', [ aValue]));
  End;

  Function  HexToLong( Const aValue: String): Cardinal;
  Var
    S : String;
    p : Integer;
  Begin
    S := aValue;
    p := Pos( '.', S);
    While p > 0 Do
    Begin
      S := Copy( S, 1, p - 1) + Copy( S, p + 1, Length( S));
      p := Pos( '.', S);
    End;
    Result := StrToInt( Format( '$%s', [ S]));
  End;

  Function  ByteToAscii( aValue: Byte): Char;
  Begin
    If ( aValue < 32) Or ( aValue >= 127)
    Then Result := '.'
    Else Result := Char( aValue);
  End;

  Function  Indent( anInd: Integer): String;
  Begin
    Result := Copy( strIndent, 1, 2 * anInd);
  End;

  Function  TrimStr( Const S: String; aLength: Integer): String;
  Begin
    Result := Copy( ExtendStr( S, aLength), 1, aLength); // Procrustrus trimming.
  End;

  Function  ExtendStr( Const S: String; aLength: Integer): String;
  Begin
    Result := S;
    While Length( Result) < aLength Do
      Result := Result + ' ';
  End;

  Function  IntegerToNoteStr( aNoteValue: Integer): String;
  Const
    Notes  : Array[ 0 .. 11] Of Char = 'CCDDEFFGGAAB';
    Sharps : Array[ 0 .. 11] Of Char = ' # #  # # # ';
  Var
    i : Integer;
  Begin
    If ( aNoteValue < 0) Or ( aNoteValue > 127)
    Then Result := 'inv'
    Else Begin
      i := aNoteValue Mod 12;
      Result := Format( '%s%d%s', [ Notes[ i], ( aNoteValue Div 12) - 5, Sharps[ i]]);
    End;
  End;


// Debug support

{$IFDEF DEBUG}
Var

  DebugFile : Text;
  DebugName : String;
  DebugTime : DWORD;

  Procedure ClearDebugTiming;
  Begin
    DebugTime := GetTickCount;
  End;

  Function FormatDebugTime( aTime: Integer): String;
  Var
    day, hr, min, sec, ms : Integer;
  Begin
    ms    := aTime Mod 1000;
    aTime := aTime Div 1000;
    sec   := aTime Mod   60;
    aTime := aTime Div   60;
    min   := aTime Mod   60;
    aTime := aTime Div   60;
    hr    := aTime Mod   24;
    aTime := aTime Div   24;
    day   := aTime;
    Result := Format( '%.2d:%.2d:%.2d.%.', [ day, hr, min, sec, ms]);
  End;

  Procedure WriteDebugStr( Const aString: String);
  Begin
    Append( DebugFile);
    WriteLn(
      DebugFile,
      Format(
        '%s - %s', [ FormatDebugTime( GetTickCount - DebugTime), aString ]
      )
    );
    Close( DebugFile);
  End;
{$ENDIF}


  Procedure InitModule;
  Var
    S : String;
    p : Integer;
  Begin
    S := Application.ExeName;
    strApplicationFileName := ExtractFileName( S);
    strApplicationPath     := ExtractFilePath( S);
    strApplicationPathURL  := FileNameToURL( strApplicationPath);
    p := Pos( '.', strApplicationFileName);
    If p = 0
    Then Begin
      strIniFileName := strApplicationPath + strApplicationFileName + '.INI';
      strLogFileName := strApplicationPath + strApplicationFileName + '.LOG';
    End
    Else Begin
      strIniFileName := strApplicationPath + Copy( strApplicationFileName, 0, p - 1) + '.INI';
      strLogFileName := strApplicationPath + Copy( strApplicationFileName, 0, p - 1) + '.LOG';
    End;
    strIndent := ' ';
    While Length( strIndent) < 256 Do
      strIndent := strIndent + strIndent;

{$IFDEF DEBUG}
    DebugName := strApplicationPath + 'debug.txt';
    Assign( DebugFile, DebugName);
    Rewrite( DebugFile);
    Close( DebugFile);
    ClearDebugTiming;
    WriteDebugStr( 'Program start');
{$ENDIF}

  End;


  Function  Clip( var aVal: Integer; aLow, aHigh: Integer): Integer;
  Begin
    if aLow > aHigh
    then begin
      if aVal > aLow
      then aVal := aLow
      else if aVal < aHigh
      then aVal := aHigh;
    end
    else begin
      if aVal < aLow
      then aVal := aLow
      else if aVal > aHigh
      then aVal := aHigh;
    end;
    Result := aVal;
  End;

  Function  CleanFileName( Const aFileName : String): String;
  Var
    i : Integer;
  Begin
    Result := ExtractFileName( aFileName);
    i := Pos( '.', Result);
    If i > 0
    Then Result := Copy( Result, 1, i - 1);
  End;



// Miscellaneous

  Procedure NotImplemented( Const aMsg: String);
  Begin
    MessageDlg(
      Format(
        'Not implemented : %s',
        [ aMsg]
      ),
      mtInformation,
      [ mbOk, mbCancel],
      0
    );
  End;

  Procedure Noop;
  Begin
  End;


// Initialisation:
//   get version info from executable
//   setup debug stuff
//   etc.

Initialization

  InitModule;

Finalization

end.


