Unit Modules;

{
  BLUE HELL software.

  Copyright (C) 1999 .. 2012, Jan Punter.

  Notes :

    midi componenets needed : http://sourceforge.net/projects/dmidi/
    hires timer needed      : source code included ... I may have stolen that somewhere without
                              acknowleging the source .. somewhere in 1999 or so ...
                              It sure looks a lot like : http://www.programmersheaven.com/download/2818/1/ZipView.aspx
                              but details as to the origin seem absent there too.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

}

Interface

Uses

  Windows, Messages, MMSystem, SysUtils, Classes, Graphics, Controls, Forms,
  ComCtrls, ExtCtrls, Buttons, Dialogs, Math, Globals;

Const


  UM_CLICK    = WM_USER + 0;
  UM_DBLCLICK = WM_USER + 1;

  clConnector    = clBlue; // Default color for connectors.
  DefaultSlotExt = '.slt'; // Default extension for slot files.
  MaxPoints      = 8192;   // Max nr of points in a graphsheet.

  // Color settings

  Clr1 = clSilver;
  Clr2 = $00400040;

  ModuleColor     : TColor = Clr1;
  LedColorOn      : TColor = clWhite;
  LedColorOff     : TColor = Clr2;
  KnobColor       : TColor = Clr2;
  ViewerColor     : TColor = Clr2;

  GraphBack       : TColor = Clr2;
  GraphBackPen    : TColor = clGray;
  GraphForePen    : TColor = clWhite;

  SymmetricColor  : TColor = clPurple;
  AsymmetricColor : TColor = clBlue;
  LogicColor      : TColor = clWhite;
  NeuronColor     : TColor = clGray;

const

  InitRepeatPause = 400;  { repeat times (ms) for knob inc/dec operations }
  RepeatPause     = 100;
  FastRepeatPause =  50;

Type

  TMimValue  = Integer;
  TShortMidi = DWORD;

  TConnectorType = (
    ctSymmetric,
    ctAsymmetric,
    ctLogic,
    ctNeuron
  );

    Function  AssembleShortMidi( aCh, anEvent, aData1, aData2: Byte): TShortMidi;
    Procedure DisAssembleShortMidi( aShortMidi: TShortMidi; Var aCh, anEvent, aData1, aData2: Byte);
    Function  ConnectorTypeToColor( aConnectorType: TConnectorType): TColor;

Type

  TLongMidi = Record
    Size : Integer;
    Data : Record {Packed Array[ 0 .. Size - 1] Of Byte} End;
  End;
  PLongMidi = ^TLongMidi;

  EModule       = Class( Exception);
  EBounds       = Class( EModule);
  ECorruptPatch = Class( EModule);
  EInvalidName  = Class( EModule);

  TMimConnector = Class;

  TNotifyCountChange      = Procedure( aSender: TObject; aCount: Integer) Of Object;
  TOnValueValueChange     = Procedure( aSender: TObject; aValue: TMimValue ) Of Object;
  TOnBooleanValueChange   = Procedure( aSender: TObject; aValue: Boolean) Of Object;
  TNotifyConnectionChange = Procedure( aSender: TObject; aSource, aDestination: TMimConnector) Of Object;
  TNotifyActiveSlotChange = Procedure( aSender: TObject; Const aSlotName: String) Of Object;
  TMidiInShortProc        = Procedure( aSender: TObject; aShortMidi: TShortMidi) Of Object;
  TMidiInLongProc         = Procedure( aSender: TObject; Const aLongMidi: TLongMidi) Of Object;
  TMidiOutShortProc       = Procedure( aSender: TObject; aShortMidi: TShortMidi) Of Object;
  TMidiOutLongProc        = Procedure( aSender: TObject; Const aLongMidi: TLongMidi) Of Object;


  TMimCustomModule = Class;

  TMimCustomDetail = Class( TComponent)
  Private
      FLeft        : Integer;
      FTop         : Integer;
      FWidth       : Integer;
      FHeight      : Integer;
      FSnap        : TPoint;
      FDefaultSnap : Boolean;
      FPen         : TPen;
      FBrush       : TBrush;
    Procedure   DoMouseDown( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); Virtual;
    Procedure   DoMouseUp  ( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); Virtual;
    procedure   DoMouseMove( Sender: TObject; Shift: TShiftState; X, Y: Integer);                       Virtual;
    Procedure   DoClick    ( Sender: TObject);
    Procedure   DoDblClick ( Sender: TObject);
    Procedure   SendClick   ( Sender: TObject);
    Procedure   SendDblClick( Sender: TObject);
    Function    CenterPoint: TPoint;
    Procedure   InvalidateRectangle( aRect: TRect; IsOpaque: Boolean); Virtual;
    Procedure   Update;                                                Virtual;
    Procedure   SetSizeAndPos( R: TRect);                              Virtual;
    Procedure   SetLeft  ( aValue: Integer);
    Procedure   SetTop   ( aValue: Integer);
    Procedure   SetWidth ( aValue: Integer);
    Procedure   SetHeight( aValue: Integer);
    Function    GetClientRect: TRect;
    Function    GetParentRect: TRect;
    Function    GetScreenRect: TRect;
    Function    GetClientOrigin: TPoint;
    Procedure   SetColor( aValue: TColor);
    Function    GetColor: TColor;
    Procedure   SetDefaultSnap( aValue: Boolean);
    Function    GetSnapX: Integer;
    Procedure   SetSnapX( aValue: Integer);
    Function    GetSnapY: Integer;
    Procedure   SetSnapY( aValue: Integer);
    Procedure   SetPen  ( aValue: TPen);
    Procedure   SetBrush( aValue: TBrush);
  Protected
    Procedure   BeginMove ( aPoint: TPoint);                           Virtual;
    Procedure   EndMove   ( aPoint: TPoint);                           Virtual;
    Procedure   UpdateMove( aPoint: TPoint);                           Virtual;
    Procedure   Invalidate;                                            Virtual;
    Procedure   PaintOn( aCanvas: TCanvas);                            Virtual;
    Function    ContainsPoint( aPoint: TPoint): Boolean;               Virtual;
    Function    ClientToScreen( aPoint: TPoint): TPoint;               Virtual;
    Function    ScreenToClient( aPoint: TPoint): TPoint;               Virtual;
    Procedure   SetParentComponent( aValue: TComponent);               Override;
    procedure   AssignTo( aDest: TPersistent);                         Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Destructor  Destroy;                                               Override;
    Function    UniqueName: String;                                    Virtual;
    Procedure   Dump( Var aFile: TextFile; anInd, aDepth: Integer);    Virtual;
    Procedure   Recolorize;                                            Virtual;
    Procedure   SnapToGrid( aPoint: TPoint);
    Function    DragAnchor  : TPoint;
    Function    DragDistance: TPoint;
    Function    DragObject    : TComponent;
    Function    LastDragObject: TComponent;
    Function    GetParentComponent: TComponent;                        Override;
    Function    HasParent: Boolean;                                    Override;

    Property    ClientRect   : TRect         Read GetClientRect;
    Property    ParentRect   : TRect         Read GetParentRect;
    Property    ScreenRect   : TRect         Read GetScreenRect;
    Property    ClientOrigin : TPoint        Read GetClientOrigin;
    Property    Width        : Integer       Read FWidth          Write SetWidth;
    Property    Height       : Integer       Read FHeight         Write SetHeight;
    Property    SnapX        : Integer       Read GetSnapX        Write SetSnapX Stored False;
    Property    SnapY        : Integer       Read GetSnapY        Write SetSnapY Stored False;
    Property    Brush        : TBrush        Read FBrush          Write SetBrush;
    Property    Pen          : TPen          Read FPen            Write SetPen;
    Property    Color        : TColor        Read GetColor        Write SetColor;
    Property    DefaultSnap  : Boolean       Read FDefaultSnap    Write SetDefaultSnap Default True;
  Published
    Property    Left         : Integer       Read FLeft           Write SetLeft;
    Property    Top          : Integer       Read FTop            Write SetTop;
  End;

  TMimDetailClass  = Class Of TMimCustomDetail;

  // a semi-abstract base type for details without mouse interaction
  // (such as TextLabels)
  //
  TMimNoMouse = Class( TMimCustomDetail)
  Protected
    Function    ContainsPoint( aPoint: TPoint): Boolean;               Override;
  End;

  // a semi-abstract base type for invisible details
  // (such as TimedEvents).
  //
  TMimInvisible = Class( TMimNoMouse)
  Protected
    Procedure   PaintOn( aCanvas: TCanvas);                            Override;
  End;

  // a semi-abstract base type for all connectors.
  TMimConnector = Class( TMimCustomDetail)
  Private
      FPlugColor       : TColor;
      FConnectionCount : Integer;
      FValue           : TMimValue;
      FCalcCount       : Integer;
      FConnectorType   : TConnectorType;
  Protected
    Procedure   SetPlugColor( aValue: TColor);
    Procedure   SetValue    ( aValue: TMimValue);
    Procedure   SetConnectorType( aValue: TConnectorType);
    Function    GetConnected: Boolean;
    Procedure   EndMove   ( aPoint: TPoint);                           Override;
    Procedure   UpdateMove( aPoint: TPoint);                           Override;
    Procedure   PaintOn( aCanvas: TCanvas);                            Override;
    Procedure   Connect( aConnector: TMimConnector);                      Virtual;
    Function    CanAccept( aConnector: TMimConnector): Boolean;           Virtual;
    Function    SearchDestination: TMimConnector;
    Procedure   ValueChanged( Var aValue: TMimValue);                     Virtual;
    procedure   AssignTo( aDest: TPersistent);                         Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Procedure   Recolorize;                                            Override;
    Procedure   IncConnectionCount;
    Procedure   DecConnectionCount;

    Property    PlugColor     : TColor         Read FPlugColor     Write SetPlugColor Default clConnector;
    Property    Value         : TMimValue         Read FValue         Write SetValue;
    Property    Connected     : Boolean        Read GetConnected;
    Property    ConnectorType : TConnectorType Read FConnectorType Write SetConnectorType;
  End;

  TMimInput = Class( TMimConnector)
  Protected
    Procedure   PaintOn( aCanvas: TCanvas);                            Override;
    Function    CanAccept( aConnector: TMimConnector): Boolean;           Override;
  End;

  TMimOutput = Class( TMimConnector)
  Protected
    Procedure   PaintOn( aCanvas: TCanvas);                            Override;
    Function    CanAccept( aConnector: TMimConnector): Boolean;           Override;
  End;

  TMimTextLabel = Class( TMimNoMouse)
  Private
      FText        : String;
      FFont        : TFont;
      FAutoSize    : Boolean;
      FParentColor : Boolean;
      FClickable   : Boolean;
    Procedure   AdjustBounds;
    Procedure   SetText( Const aValue: String);
    Procedure   SetTextColor( aValue: TColor);
    Function    GetTextColor: TColor;
    Procedure   SetFont( aValue: TFont);
    Procedure   SetAutoSize( aValue: Boolean);
    Procedure   SetParentColor( aValue: Boolean);
  Protected
    Function    ContainsPoint( aPoint: TPoint): Boolean;               Override;
    Procedure   PaintOn( aCanvas: TCanvas);                            Override;
    procedure   AssignTo( aDest: TPersistent);                         Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Destructor  Destroy;                                               Override;
    Procedure   Recolorize;                                            Override;

    Property    Text        : String  Read FText        Write SetText;
    Property    TextColor   : TColor  Read GetTextColor Write SetTextColor;
    Property    Font        : TFont   Read FFont        Write SetFont;
    Property    AutoSize    : Boolean Read FAutoSize    Write SetAutoSize;
    Property    ParentColor : Boolean Read FParentColor Write SetParentColor;
  End;

  TMimViewer = Class( TMimCustomDetail)
  Private
      FText : String;
      FFont : TFont;
    Procedure   SetText( Const aValue: String);
    Procedure   SetBorderColor( aValue: TColor);
    Function    GetBorderColor: TColor;
    Procedure   SetTextColor( aValue: TColor);
    Function    GetTextColor: TColor;
    Procedure   SetFont( aValue: TFont );
  Protected
    Procedure   PaintOn( aCanvas: TCanvas);                            Override;
    procedure   AssignTo( aDest: TPersistent);                         Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Destructor  Destroy;                                               Override;
    Procedure   Recolorize;                                            Override;

    Property    Text        : String Read FText          Write SetText;
    Property    BorderColor : TColor Read GetBorderColor Write SetBorderColor;
    Property    TextColor   : TColor Read GetTextColor   Write SetTextColor;
    Property    Font        : TFont  Read FFont          Write SetFont;
  End;

  TMimKnobMode = (
    kmNone,
    kmTurn,
    kmInc,
    kmDec
  );

  TMimKnob = Class( TMimCustomDetail)
  Private
      FValue         : TMimValue;
      FMinValue      : TMimValue;
      FMaxValue      : TMimValue;
      FStep          : TMimValue;
      FViewer        : TMimViewer;
      FOnValueChange : TOnValueValueChange;
      FOldP1         : TPoint;
      FOldP2         : TPoint;
      FKnobMode      : TMimKnobMode;
      FRepeatTimer   : TTimer;
      FRepeatCount   : Integer;
    Procedure   DoMouseDown( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); Override;
    Procedure   DoMouseUp  ( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); Override;
    procedure   DoMouseMove( Sender: TObject; Shift: TShiftState; X, Y: Integer);                       Override;
    Procedure   DoTimerFired( aSender: TObject);
    Procedure   UpdatePeers;
    Procedure   Update;                                                Override;
    Procedure   SetValue   ( aValue: TMimValue);
    Procedure   SetMinValue( aValue: TMimValue);
    Procedure   SetMaxValue( aValue: TMimValue);
    procedure   SetStep    ( aValue: TMimValue);
    Procedure   SetViewer  ( aValue: TMimViewer);
    Procedure   SetOnValueChange( aValue: TOnValueValueChange);
  Protected
    Procedure   BeginMove ( aPoint: TPoint);                           Override;
    Procedure   EndMove( aPoint: TPoint);                              Override;
    Procedure   UpdateMove( aPoint: TPoint);                           Override;
    Procedure   PaintOn( aCanvas: TCanvas);                            Override;
    procedure   AssignTo( aDest: TPersistent);                         Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Destructor  Destroy;                                               Override;
    Procedure   Recolorize;                                            Override;

    Property    Value        : TMimValue           Read FValue         Write SetValue;
    Property    MinValue     : TMimValue           Read FMinValue      Write SetMinValue;
    Property    MaxValue     : TMimValue           Read FMaxValue      Write SetMaxValue;
    Property    Step         : TMimValue           Read FStep          Write SetStep;
    Property    Viewer       : TMimViewer          Read FViewer        Write SetViewer;
    Property    KnobMode     : TMimKnobMode        Read FKnobMode;
    Property    OnValueChange: TOnValueValueChange Read FOnValueChange Write SetOnValueChange;
  End;

  TMimKnobSmall = class( TMimKnob)
  Public
    Constructor Create( anOwner: TComponent);                          Override;
  end;

  TMimKnobMedium = class( TMimKnob)
  Public
    Constructor Create( anOwner: TComponent);                          Override;
  end;

  TMimExpKnob = Class( TMimKnob)
  Protected
    Procedure   EndMove( aPoint: TPoint);                              Override;
    Procedure   UpdateMove( aPoint: TPoint);                           Override;
  End;

  TMimPushButton = Class( TMimCustomDetail)
  Private
      FDown          : Boolean;
      FOldDown       : Boolean;
      FBitmap        : TBitmap;
      FOnValueChange : TOnBooleanValueChange;
      FGroupId       : Integer;
      FAllowAllUp    : Boolean;
      FOnClick       : TNotifyEvent;
    Procedure   SelectUnique;
    Procedure   SetBitmap       ( aValue: TBitmap);
    Procedure   SetDown         ( aValue: Boolean);
    Procedure   SetGroupId      ( aValue: Integer);
    Procedure   SetAllowAllUp   ( aValue: Boolean);
    Procedure   SetOnValueChange( aValue: TOnBooleanValueChange);
  Protected
    Procedure   BeginMove ( aPoint: TPoint);                           Override;
    Procedure   EndMove   ( aPoint: TPoint);                           Override;
    Procedure   UpdateMove( aPoint: TPoint);                           Override;
    Procedure   PaintOn( aCanvas: TCanvas);                            Override;
    procedure   AssignTo( aDest: TPersistent);                         Override;

    Property    Bitmap: TBitmap Read FBitmap Write SetBitmap;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Destructor  Destroy;                                               Override;
    Procedure   Recolorize;                                            Override;
    Procedure   LoadBitmap( Const aFileName: String);

    Property    Down         : Boolean               Read FDown          Write SetDown;
    Property    GroupId      : Integer               Read FGroupId       Write SetGroupId;
    Property    AllowAllUp   : Boolean               Read FAllowAllUp    Write SetAllowAllUp;
    Property    OnValueChange: TOnBooleanValueChange Read FOnValueChange Write SetOnValueChange;
    Property    OnClick      : TNotifyEvent          Read FOnClick       Write FOnClick;
  End;

  TMimImage = Class( TMimNoMouse)
  Private
      FBitmap      : TBitmap;
      FTransparent : Boolean;
      FAutoSize    : Boolean;
    Procedure   AdjustBounds;
    Procedure   SetBitmap     ( aValue: TBitmap);
    Procedure   SetTransparent( aValue: Boolean);
    Procedure   SetAutoSize   ( aValue: Boolean);
  Protected
    Procedure   PaintOn( aCanvas: TCanvas);                            Override;
    procedure   AssignTo( aDest: TPersistent);                         Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Destructor  Destroy;                                               Override;
    Procedure   LoadBitmap( Const aFileName: String);

    Property    Bitmap      : TBitmap Read FBitmap      Write SetBitmap;
    Property    Transparent : Boolean Read FTransparent Write SetTransparent;
    Property    AutoSize    : Boolean Read FAutoSize    Write SetAutoSize;
  End;

  TMimPointArray = Array[ 0 .. MaxPoints - 1] Of TPoint;
  PMimPointArray = ^TMimPointArray;

  TMimGraphSheet = Class( TMimNoMouse)
  Private
      FData      : PMimPointArray;
      FEltCount  : Integer;
      FXScale    : Integer;
      FYScale    : Integer;
      FDrawColor : TColor;
    Procedure   CheckBounds ( anIndex: Integer);
    Function    GetData     ( anIndex: Integer): TPoint;
    Procedure   SetData     ( anIndex: Integer; aValue: TPoint);
    Function    GetXData    ( anIndex: Integer): Integer;
    Procedure   SetXData    ( anIndex: Integer; aValue: Integer);
    Function    GetYData    ( anIndex: Integer): Integer;
    Procedure   SetYData    ( anIndex: Integer; aValue: Integer);
    Function    GetXScaled  ( anIndex: Integer): Integer;
    Function    GetYScaled  ( anIndex: Integer): Integer;
    Procedure   SetEltCount ( aValue: Integer);
    Procedure   SetXScale   ( aValue: Integer);
    Procedure   SetYScale   ( aValue: Integer);
    Procedure   SetDrawColor( aValue: TColor );
  Protected
    Procedure   PaintOn( aCanvas: TCanvas);                            Override;
    procedure   AssignTo( aDest: TPersistent);                         Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Destructor  Destroy;                                               Override;
    Procedure   Dump( Var aFile: TextFile; anInd, aDepth: Integer);    Override;
    Procedure   Recolorize;                                            Override;
    Procedure   Clear;
    Property    Data   [ anIndex: Integer]: TPoint  Read GetData    Write SetData;
    Property    XData  [ anIndex: Integer]: Integer Read GetXData   Write SetXData;
    Property    YData  [ anIndex: Integer]: Integer Read GetYData   Write SetYData;
    Property    XScaled[ anIndex: Integer]: Integer Read GetXScaled;
    Property    YScaled[ anIndex: Integer]: Integer Read GetYScaled;
    Property    EltCount                  : Integer Read FEltCount  Write SetEltCount;
    Property    XScale                    : Integer Read FXScale    Write SetXScale;
    Property    YScale                    : Integer Read FYScale    Write SetYScale;
    Property    DrawColor                 : TColor  Read FDrawColor Write SetDrawColor;
  End;

  TMimLed = Class( TMimNoMouse)
  Private
      FActive   : Boolean;
      FOnColor  : TColor;
      FOffColor : TColor;
    Procedure   SetActive  ( aValue: Boolean);
    Procedure   SetOnColor ( aValue: TColor);
    Procedure   SetOffColor( aValue: TColor);
  Protected
    Procedure   PaintOn( aCanvas: TCanvas);                            Override;
    procedure   AssignTo( aDest: TPersistent);                         Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Procedure   Recolorize;                                            Override;

    Property    Active   : Boolean Read FActive   Write SetActive;
    Property    OnColor  : TColor  Read FOnColor  Write SetOnColor;
    Property    OffColor : TColor  Read FOffColor Write SetOffColor;
  end;

  TMimCustomModule = Class( TMimCustomDetail)
  Private
      FSelected   : Boolean;
      FInCreation : Boolean;
      FTextLabel  : TMimTextLabel;
    Procedure   DoMouseDown( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); Override;
    Procedure   DoMouseUp  ( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); Override;
    procedure   DoMouseMove( Sender: TObject; Shift: TShiftState; X, Y: Integer);                       Override;
    Procedure   Update;                                                Override;
    Procedure   UpdateDetail;
    Procedure   InvalidateRectangle( aRect: TRect; IsOpaque: Boolean); Override;
    Procedure   SetSelected( aValue: Boolean);
    Function    GetTitle: String;
    Procedure   SetTitle( Const aValue: String);
    Procedure   HandleUpdateMove( aPoint: TPoint);
    Procedure   UnselectPeers;
    Procedure   BringToFront;
    Procedure   Disconnect( aConnector: TMimConnector);                   Virtual;
    Procedure   MarkConnections;
  Protected
    Procedure   Notification( aComponent: TComponent; anOperation: TOperation); Override;
    Procedure   HandleSelectionChange( aValue: Boolean);               Virtual;
    Procedure   EndMove( aPoint: TPoint);                              Override;
    Procedure   UpdateMove( aPoint: TPoint);                           Override;
    Procedure   Move( aPoint: TPoint);                                 Virtual;
    Procedure   PaintOn( aCanvas: TCanvas);                            Override;
    Procedure   PaintDetails( aCanvas: TCanvas);                       Virtual;
    Procedure   CreateDetails;                                         Virtual;
    Function    CreateDetail( aLeft, aTop: Integer; Const aName: String; aDetailClass: TMimDetailClass): TMimCustomDetail;
    Function    SearchDestination( aSource: TMimConnector; aPoint: TPoint): TMimConnector;
    Function    SearchDetail( aPoint: TPoint): TMimCustomDetail;
    Function    ContainsPoint( aPoint: TPoint): Boolean;               Override;
    Function    FindDestination( aSource: TMimConnector; aPoint: TPoint): TMimConnector;
    Procedure   Connect( aSource, aDestination: TMimConnector);           Virtual;
    Function    ClientToScreen( aPoint: TPoint): TPoint;               Override;
    Function    ScreenToClient( aPoint: TPoint): TPoint;               Override;
    Procedure   ScrollSelfInView;
    procedure   AssignTo( aDest: TPersistent);                         Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Destructor  Destroy;                                               Override;
    Procedure   Execute;                                               Virtual;
    Procedure   Dump( Var aFile: TextFile; anInd, aDepth: Integer);    Override;
    Procedure   Recolorize;                                            Override;

    Property    Selected  : Boolean    Read FSelected  Write SetSelected;
    Property    TextLabel : TMimTextLabel Read FTextLabel Write FTextLabel;
  Published
    Property    Title     : String     Read GetTitle   Write SetTitle;
  End;

  TMimModuleClass = Class Of TMimCustomModule;

  TMimModule = Class( TMimCustomModule)
  End;

  TMimModule2 = Class( TMimModule)
  Public
    Constructor Create( anOwner: TComponent);                          Override;
  End;

  TMimModule4 = Class( TMimModule)
  Public
    Constructor Create( anOwner: TComponent);                          Override;
  End;

  TMimModule8 = Class( TMimModule)
  Public
    Constructor Create( anOwner: TComponent);                          Override;
  End;

  TMimMidiModule = Class( TMimCustomModule)
  Private
    FMidiInShort   : TShortMidi;
    FMidiInLong    : TLongMidi;
    FOnMidiInShort : TMidiInShortProc;
    FOnMidiInLong  : TMidiInLongProc;
  Protected
    Procedure   AcceptMidiShort( aShortMidi: TShortMidi);              Virtual;
    Procedure   AcceptMidiLong( aLongMidi: TLongMidi);                 Virtual;
  Public
    Procedure   MidiOutShort( aShortMidi: TShortMidi);                 Virtual;
    Procedure   MidiOutLong( Const aLongMidi: TLongMidi);              Virtual;
    Property    MidiInShort   : TShortMidi       Read FMidiInShort;
    Property    MidiInLong    : TLongMidi        Read FMidiInLong;
    Property    OnMidiInShort : TMidiInShortProc Read FOnMidiInShort Write FOnMidiInShort;
    Property    OnMidiInLong  : TMidiInLongProc  Read FOnMidiInLong  Write FOnMidiInLong;
  End;

  TMimConnection = Class
  Private
      FSource      : TMimConnector;
      FDestination : TMimConnector;
      FMarked      : Boolean;
  Public
    Constructor Create( aSource, aDestination: TMimConnector);
    Destructor  Destroy;                                               Override;

    Property    Source      : TMimConnector Read FSource;
    Property    Destination : TMimConnector Read FDestination;
  End;

  TMimDrawSheet = Class;

  TMimConnections = Class( TList)
  Private
      FOwner : TMimDrawSheet;
    Function    GetConnectionCount: Integer;
    Function    GetConnection ( anIndex: Integer): TMimConnection;
    Function    GetSource     ( anIndex: Integer): TMimConnector;
    Function    GetDestination( anIndex: Integer): TMimConnector;
    Procedure   Disconnect( aConnector: TMimConnector);
  Public
    Function    Connect( aSource, aDestination: TMimConnector): Boolean;
    Function    FindConnection( aSource, aDestination: TMimConnector): Integer;
    Procedure   Execute;                                               Virtual;

    Property    ConnectionCount : Integer Read GetConnectionCount;
    Property    Connection [ anIndex: Integer]: TMimConnection Read GetConnection;
    Property    Source     [ anIndex: Integer] : TMimConnector Read GetSource;
    Property    Destination[ anIndex: Integer] : TMimConnector Read GetDestination;
    Property    Owner: TMimDrawSheet Read FOwner Write FOwner;
  End;

  TMimSlot = Class;

  TMimDrawSheet = Class( TControl)
  Private
      FModules           : TList;
      FConnections       : TMimConnections;
      FSlot              : TMimSlot;
      FCanvas            : TCanvas;
      FPen               : TPen;
      FSlotActive        : Boolean;
      FDoDrawConnections : Boolean;
      FConnectionFixups  : TStringList;
    procedure   WMPaint( Var aMessage: TWMPaint); Message WM_PAINT;
    Procedure   DoMouseDown( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    Procedure   DoMouseUp  ( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure   DoMouseMove( Sender: TObject; Shift: TShiftState; X, Y: Integer);
    Procedure   DoClick    ( Sender: TObject);
    Procedure   DoDblClick ( Sender: TObject);
    Procedure   SendClick   ( Sender: TObject);
    Procedure   SendDblClick( Sender: TObject);
    Procedure   SetPen( aValue: TPen);
    Procedure   SetSlotActive( aValue: Boolean);
    Function    GetModule     ( anIndex: Integer): TMimCustomModule;
    Function    GetConnection ( anIndex: Integer): TMimConnection;
    Function    GetSource     ( anIndex: Integer): TMimConnector;
    Function    GetDestination( anIndex: Integer): TMimConnector;
    Procedure   PerformCountChange;
    Procedure   HandleModuleUpdateMove( aModule: TMimCustomModule; aPoint: TPoint);
    Procedure   HandleModuleMoved( aModule: TMimCustomModule; aPoint: TPoint);
    Procedure   UnselectAllModulesBut( aModule: TMimCustomModule);
    Procedure   Disconnect( aConnector: TMimConnector);
    Procedure   InvalidateMarkedConnections( IsOpaque: Boolean);
    Procedure   MarkConnections( aConnector: TMimConnector);
    Procedure   ClearConnectionFixups;
    Procedure   AddConnectionFixup( Const S: String);
    Procedure   FixupConnections;
    Procedure   CreateConnectionFromFixup( Const S: String);
    Procedure   SyncModules;
    Procedure   PerformMidiOutShort( aSender: TObject; aShortMidi: TShortMidi);
    Procedure   PerformMidiOutLong( aSender: TObject; Const aLongMidi: TLongMidi);
  Protected
    Procedure   BeginMove ( aPoint: TPoint);                           Virtual;
    Procedure   EndMove   ( aPoint: TPoint);                           Virtual;
    Procedure   UpdateMove( aPoint: TPoint);                           Virtual;
    Procedure   Notification( aComponent: TComponent; anOperation: TOperation); Override;
    Procedure   BringModuleToFront( aModule: TMimCustomModule);
    Procedure   SelectModulesInRect( aRect: TRect);
    Procedure   InvalidateRectangleDetail( aRect: TRect; IsOpaque: Boolean);
    Function    SearchDestination( aSource: TMimConnector; aPoint: TPoint): TMimConnector;
    Function    SearchDetail( aPoint: TPoint): TMimCustomDetail;
    Function    ContainsPoint( aPoint: TPoint): Boolean;               Virtual;
    Function    FindDestination( aSource: TMimConnector; aPoint: TPoint): TMimConnector;
    Procedure   Connect( aSource, aDestination: TMimConnector);           Virtual;
    Procedure   Paint;                                                 Virtual;
    Procedure   PaintDetails    ( aCanvas: TCanvas);
    Procedure   PaintConnections( aCanvas: TCanvas);
    Procedure   SizeToFit;
    Procedure   ScrollModuleInView( aModule: TMimCustomModule);
    Procedure   ConnectionMade( aSource, aDestination: TMimConnector);    Virtual;
    Procedure   ConnectionLost( aSource, aDestination: TMimConnector);    Virtual;
    Procedure   GetChildren( aProc: TGetChildProc; Root: TComponent);  Override;
    Procedure   DefineProperties( aFiler : TFiler );                   Override;
    Procedure   WriteConnections( aWriter: TWriter);
    Procedure   ReadConnections ( aReader: TReader);
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Destructor  Destroy;                                               Override;
    Procedure   Execute;                                               Virtual;
    Procedure   Dump( Var aFile: TextFile; anInd, aDepth: Integer);    Virtual;
    Procedure   Recolorize;
    Procedure   AcceptMidiShort( aShortMidi: TShortMidi);
    Procedure   AcceptMidiLong( Const aLongMidi: TLongMidi);
    Function    CreateModule( aModuleClass: TMimModuleClass): TMimCustomModule;
    Procedure   FreeModule( anIndex: Integer);
    Procedure   FreeModules;
    Procedure   FreeSelectedModules;
    Procedure   CopyModulesToClipBoard;
    Procedure   PasteModulesFromClipBoard;
    Procedure   SelectAllModules;
    Procedure   InvertModuleSelection;
    Function    IndexOfModule( aModule: TMimCustomModule): Integer;
    Function    ModuleCount    : Integer;
    Function    ConnectionCount: Integer;
    Function    DragAnchor  : TPoint;
    Function    DragDistance: TPoint;
    Function    DragObject    : TComponent;
    Function    LastDragObject: TComponent;

    Property    Pen        : TPen    Read FPen        Write SetPen;
    Property    SlotActive : Boolean Read FSlotActive Write SetSlotActive;
    Property    Module     [ anIndex: Integer]: TMimCustomModule  Read GetModule;
    Property    Connection [ anIndex: Integer]: TMimConnection    Read GetConnection;
    Property    Source     [ anIndex: Integer]: TMimConnector     Read GetSource;
    Property    Destination[ anIndex: Integer]: TMimConnector     Read GetDestination;
  End;

  TMimSlot = Class( TScrollBox)
  Private
      FOnCountChange      : TNotifyCountChange;
      FPrivateCountChange : TNotifyCountChange;
      FOnConnectionMade   : TNotifyConnectionChange;
      FOnConnectionLost   : TNotifyConnectionChange;
      FOnMidiOutShort     : TMidiOutShortProc;
      FOnMidiOutLong      : TMidiOutLongProc;
      FDrawSheet          : TMimDrawSheet;
      FFileName           : String;
      FActive             : Boolean;
    Procedure   PerformCountChange;
    Procedure   SendClick   ( Sender: TObject);
    Procedure   SendDblClick( Sender: TObject);
    Procedure   UMClick   ( Var aMessage: TMessage); Message UM_CLICK;
    Procedure   UMDblClick( Var aMessage: TMessage); Message UM_DBLCLICK;
    Procedure   DoResize( aSender: TObject);
    Procedure   PerformMidiOutShort( aSender: TObject; aShortMidi: TShortMidi);
    Procedure   PerformMidiOutLong( aSender: TObject; Const aLongMidi: TLongMidi);
    Procedure   SetActive( aValue: Boolean);
  Protected
    Procedure   ConnectionMade( aSource, aDestination: TMimConnector);    Virtual;
    Procedure   ConnectionLost( aSource, aDestination: TMimConnector);    Virtual;

    Property    OnPrivateCountChange : TNotifyCountChange Read FPrivateCountChange Write FPrivateCountChange;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Destructor  Destroy;                                               Override;
    Procedure   Execute;                                               Virtual;
    Procedure   Recolorize;
    Procedure   Save;                                                  Virtual;
    Procedure   SaveAs( Const aFileName: String);                      Virtual;
    Procedure   Load( Const aFileName: String);                        Virtual;
    Procedure   Dump( Var aFile: TextFile; anInd, aDepth: Integer);    Virtual;
    Procedure   AcceptMidiShort( aShortMidi: TShortMidi);
    Procedure   AcceptMidiLong( Const aLongMidi: TLongMidi);
    Function    CreateModule( aModuleClass: TMimModuleClass): TMimCustomModule;
    Procedure   FreeModule( anIndex: Integer);
    Procedure   FreeModules;
    Procedure   FreeSelectedModules;
    Procedure   CopyModulesToClipBoard;
    Procedure   PasteModulesFromClipBoard;
    Procedure   SelectAllModules;
    Procedure   InvertModuleSelection;
    Function    ModuleCount: Integer;
    Procedure   ScrollModuleInView( aModule: TMimCustomModule);
    Function    DragAnchor  : TPoint;
    Function    DragDistance: TPoint;
    Function    DragObject    : TComponent;
    Function    LastDragObject: TComponent;
  Published

    Property    OnCountChange    : TNotifyCountChange      Read FOnCountChange    Write FOnCountChange;
    Property    OnConnectionMade : TNotifyConnectionChange Read FOnConnectionMade Write FOnConnectionMade;
    Property    OnConnectionLost : TNotifyConnectionChange Read FOnConnectionLost Write FOnConnectionLost;
    Property    OnMidiOutShort   : TMidiOutShortProc       Read FOnMidiOutShort   Write FOnMidiOutShort;
    Property    OnMidiOutLong    : TMidiOutLongProc        Read FOnMidiOutLong    Write FOnMidiOutLong;
    Property    SlotFileName     : String                  Read FFileName;
    Property    Active           : Boolean                 Read FActive           Write SetActive;
  End;

  TMimModulePageControl = Class( TPageControl)
  Private
      FOnCountChange       : TNotifyCountChange;
      FOnModuleCountChange : TNotifyCountChange;
      FOnConnectionMade    : TNotifyConnectionChange;
      FOnConnectionLost    : TNotifyConnectionChange;
      FOnActiveSlotChange  : TNotifyActiveSlotChange;
      FOnMidiOutShort      : TMidiOutShortProc;
      FOnMidiOutLong       : TMidiOutLongProc;
      FOnDetailClick       : TNotifyEvent;
      FOnDetailDblClick    : TNotifyEvent;
      FPageCount           : Integer;
      FLockCount           : Integer;
    Function    GetActiveSlot: TMimSlot;
    Procedure   DoPrivateCountChange( Sender: TObject; aCount: Integer);
    Procedure   DoPrivateConnectionMade( aSender: TObject; aSource, aDestination: TMimConnector);
    Procedure   DoPrivateConnectionLost( aSender: TObject; aSource, aDestination: TMimConnector);
    Procedure   DoMidiOutShort( aSender: TObject; aShortMidi: TShortMidi);
    Procedure   DoMidiOutLong( aSender: TObject; Const aLongMidi: TLongMidi);
    Procedure   PerformCountChange;
    Procedure   PerformModuleCountChange( Sender: TMimSlot);
    Procedure   PerformConnectionMade( aSender: TObject; aSource, aDestination: TMimConnector);
    Procedure   PerformConnectionLost( aSender: TObject; aSource, aDestination: TMimConnector);
    Procedure   PerformSlotChange( aSender: TObject);
    Procedure   PerformClick   ( aSender: TObject);
    Procedure   PerformDblClick( aSender: TObject);
    Procedure   Lock;
    Procedure   UnLock;
  Protected
    Procedure   Change;                                                Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Procedure   Execute;                                               Virtual;
    Procedure   Recolorize;
    Procedure   Save;                                                  Virtual;
    Procedure   SaveAll;                                               Virtual;
    Procedure   SaveAs( Const aFileName: String);                      Virtual;
    Procedure   Load( Const aFileName: String);                        Virtual;
    Procedure   Dump( Var aFile: TextFile; anInd, aDepth: Integer);    Virtual;
    Procedure   AddSlot( Const aName: String);
    Procedure   AcceptMidiShort( aShortMidi: TShortMidi);
    Procedure   AcceptMidiLong( Const aLongMidi: TLongMidi);
    Procedure   FreeCurrentSlot;
    Procedure   FreeAllSlots;
    Function    CreateModule( aModuleClass: TMimModuleClass): TMimCustomModule;
    Procedure   FreeModule( anIndex: Integer);
    Procedure   FreeModules;
    Procedure   FreeSelectedModules;
    Procedure   CopyModulesToClipBoard;
    Procedure   PasteModulesFromClipBoard;
    Procedure   SelectAllModules;
    Procedure   InvertModuleSelection;
    Function    ModuleCount : Integer;
    Function    DragAnchor  : TPoint;
    Function    DragDistance: TPoint;
    Function    DragObject    : TComponent;
    Function    LastDragObject: TComponent;

    Property    ActiveSlot: TMimSlot Read GetActiveSlot;
  Published
    Property    OnCountChange       : TNotifyCountChange      Read FOnCountChange       Write FOnCountChange;
    Property    OnModuleCountChange : TNotifyCountChange      Read FOnModuleCountChange Write FOnModuleCountChange;
    Property    OnConnectionMade    : TNotifyConnectionChange Read FOnConnectionMade    Write FOnConnectionMade;
    Property    OnConnectionLost    : TNotifyConnectionChange Read FOnConnectionLost    Write FOnConnectionLost;
    Property    OnActiveSlotChange  : TNotifyActiveSlotChange Read FOnActiveSlotChange  Write FOnActiveSlotChange;
    Property    OnMidiOutShort      : TMidiOutShortProc       Read FOnMidiOutShort      Write FOnMidiOutShort;
    Property    OnMidiOutLong       : TMidiOutLongProc        Read FOnMidiOutLong       Write FOnMidiOutLong;
    Property    OnDetailClick       : TNotifyEvent            Read FOnDetailClick       Write FOnDetailClick;
    Property    OnDetailDblClick    : TNotifyEvent            Read FOnDetailDblClick    Write FOnDetailDblClick;
  End;

  TMimModuleButton = Class( TSpeedButton)
  Private
      FPageName    : String;
      FModuleClass : TMimModuleClass;
  Public
    Procedure   LoadGlyph;
    Procedure   Initialize( Const aPageName: String; aModuleClass: TMimModuleClass);

    Property    PageName   : String       Read FPageName;
    Property    ModuleClass: TMimModuleClass Read FModuleClass;
  End;

  TMimModuleButtonClick = Procedure( aSender: TObject; aModuleClass: TMimModuleClass) Of Object;

  TMimmoduleSelector = Class( TPageControl)
  Private
      FRegisteredModules   : TList;
      FOnModuleButtonClick : TMimModuleButtonClick;
      FButtonsEnabled      : Boolean;
    Procedure   SetButtonsEnabled( aValue: Boolean);
    Function    AddPage( Const aName: String): TTabSheet;
    Function    PageButtonCount( Const aName: String): Integer;
    Function    PageExists( Const aName: String): Boolean;
    Function    PageByName( Const aName: String): TTabSheet;
    Function    FindModulePage( aModuleClass: TMimModuleClass): TTabSheet;
    Function    FindModuleButton( aModuleClass: TMimModuleClass): TMimModuleButton;
    Procedure   DoModuleMouseDown( aSender: TObject; aButton: TMouseButton; aShift: TShiftState; X, Y: Integer);
    Procedure   FormatPage( aPage: TTabSheet);
  Protected
    Procedure   Notification( aComponent: TComponent; anOperation: TOperation); Override;
  Public
    Constructor Create( anOwner: TComponent);                          Override;
    Destructor  Destroy;                                               Override;
    Procedure   RegisterModuleClass( Const aPageName: String; aModuleClass: TMimModuleClass);
    Procedure   UnregisterModuleClass( aModuleClass: TMimModuleClass);
  Published
    Property    OnModuleButtonClick : TMimModuleButtonClick Read FonModuleButtonClick Write FOnModuleButtonClick;
    Property    ButtonsEnabled      : Boolean            Read FButtonsEnabled      Write SetButtonsEnabled;
  End;


  procedure   Register;



implementation

Uses

  ClipBrd;

Const

  NullPoint : TPoint = ( X:0; Y:0);

  Function IsNullPoint( aPoint: TPoint): Boolean;
  Begin
    With aPoint Do
      Result := ( x = 0) And ( y = 0);
  End;

  Function IsNearNullPoint( aPoint: TPoint): Boolean;
  Begin
    With aPoint Do
      Result := ( Abs( x) < 10) And ( Abs( y) < 10);
  End;

  Function  AssembleShortMidi( aCh, anEvent, aData1, aData2: Byte): TShortMidi;
  Begin
    Result := Integer( aCh And $0f)+ Integer( anEvent And $0f) Shl 4 + Integer( aData1) Shl 8 + Integer( aData2) Shl 16;
  End;

  Procedure DisAssembleShortMidi( aShortMidi: TShortMidi; Var aCh, anEvent, aData1, aData2: Byte);
  Begin
    aCh     := ( aShortMidi       ) And $0f;
    anEvent := ( aShortMidi Shr  4) And $0f;
    aData1  := ( aShortMidi Shr  8) And $7f;
    aData2  := ( aShortMidi Shr 16) And $7f;
  End;

  Function  ConnectorTypeToColor( aConnectorType: TConnectorType): TColor;
  Begin
    Case aConnectorType Of
      ctSymmetric  : Result := SymmetricColor;
      ctAsymmetric : Result := AsymmetricColor;
      ctLogic      : Result := LogicColor;
      ctNeuron     : Result := NeuronColor;
      Else           Result := SymmetricColor;
    End;
  End;

  procedure   Register;
  // Althogh seemingly some module stuff can be registered there is no
  // need to do so for the MidiMod project.
  begin
    RegisterComponents(
      'Modules',
      [
        TMimmoduleSelector,
        TMimModulePageControl
      ]
    );
  end;


  Function BoolToStr( aValue: Boolean): String;
  Begin
    Case aValue Of
      False : Result := 'False';
      True  : Result := 'True';
    End;
  End;


// Utility functions.

    Function  CleanFileName( Const aFileName: String): String;
    Var
      p : Integer;
    Begin
      Result := aFileName;
      p := Pos( Uppercase( DefaultSlotExt), UpperCase( Result));
      If p > 0
      Then Result := Copy( Result, 1, p - 1);
      p := Length( Result);
      While p > 0 Do
        If Result[ p] = '\'
        Then Break
        Else Dec( p);
      If p < Length( Result)
      Then Result := Copy( Result, p + 1, Length( Result));
    End;

    Procedure Snap( Var aPoint: TPoint; aSnap: TPoint);
    Begin
      With aSnap Do
      Begin
        If x <>  0
        Then aPoint.x := (( aPoint.x + x Div 2) Div x) * x;
        If y <> 0
        Then aPoint.y := (( aPoint.y + y Div 2) Div y) * y;
      End;
    End;

    Procedure AssertParamInBounds( aParam, aLow, aHigh: Integer; Const aFuncName, aParamName: String);
    Begin
      Assert(
        ( aParam >= aLow) And ( aParam <= aHigh),
        Format(
          '%s : Parameter %s (%d) out of bounds [%d,%d]',
          [
            aFuncName,
            aParamName,
            aParam,
            aLow,
            aHigh
          ]
        )
      )
    End;

    Function  Indent( anInd: Integer; S: String): String;
    Const
      IStr        : String = ' ';
      Initialized : Boolean = False;
    Begin
      If Not Initialized
      Then Begin
        While Length( IStr) < 128 Do
          IStr := IStr + IStr;
        Initialized := True;
      End;
      Result := Format( ' %s%s', [ Copy( IStr, 1, 2 * anInd), S]);
    End;

    Function   RectanglesOverlap( A, B: TRect): Boolean;
    Var
      R : TRect;
    Begin
      Result := IntersectRect( R, A, B);
    End;

    Procedure  Exchange( Var A, B: Integer);
    Var
      T : Integer;
    Begin
      T := A;
      A := B;
      B := T;
    End;

    Function  Distance( aPoint: TPoint): Integer;
    Begin
      With aPoint Do
        Result := Round( Sqrt( x * x + y * y));
    End;

    Function   AddPoints( A, B: TPoint): TPoint;
    Begin
      With Result Do
      Begin
        x := A.x + B.x;
        y := A.y + B.y;
      End;
    End;

    Function   SubPoints( A, B: TPoint): TPoint;
    Begin
      With Result Do
      Begin
        x := A.x - B.x;
        y := A.y - B.y;
      End;
    End;

    Function MoveRect( aRect: TRect; aPoint: TPoint): TRect;
    Begin
      Result := aRect;
      With Result, aPoint Do
      Begin
        Left   := Left   + x;
        Top    := Top    + y;
        Right  := Right  + x;
        Bottom := Bottom + y;
      End;
    End;

    Function InflateRectangle( R: TRect; a: Integer): TRect;
    Begin
      Result := R;
      With Result Do
      Begin
        Dec( Left  , a);
        Dec( Top   , a);
        Inc( Right , a);
        Inc( Bottom, a);
      End;
    End;

    // For angles we we use the mathematical vision, 0 degrees point to the right.

    Function  ValuesToAngle( MinValue, Value, MaxValue: TMimValue): Integer;
    Begin
      If MinValue <> MaxValue
      Then Result := 270 * ( Value - MinValue) Div ( MinValue - MaxValue) + 225
      Else Result := 0; // arbitrary ...
    End;

    Function  PolarToPoint( aCenter: TPoint; anAngle, aLength: Integer): TPoint;
    Begin
      With Result Do
      Begin
        x := Round( aCenter.x + aLength * Cos( DegToRad( anAngle)));
        y := Round( aCenter.y - aLength * Sin( DegToRad( anAngle)));
      End;
    End;

    Function  DistanceToValue( MinValue, MaxValue: TMimValue; aPoint: TPoint): TMimValue;
    Var
      anAngle: Extended;
    Begin
      anAngle := RadToDeg( ArcTan2( - aPoint.y, aPoint.x));
      While anAngle < 0 Do
        anAngle := anAngle + 360;
      While anAngle > 270 Do
        anAngle := anAngle - 360;
      Result := MinValue + Round( (MinValue - MaxValue) * ( anAngle - 225) / 270);
      If MaxValue > MinValue
      Then Begin
        If Result > MaxValue
        Then Result := MaxValue
        Else If Result < MinValue
        Then Result := MinValue;
      End
      Else begin
        If Result > MinValue
        Then Result := MinValue
        Else If Result < MaxValue
        Then Result := MaxValue;
      End
    End;

    Function  DistanceToExpValue( MinValue, MaxValue: TMimValue; aPoint: TPoint): TMimValue;
    Var
      anAngle: Extended;
    Begin
      anAngle := RadToDeg( ArcTan2( - aPoint.y, aPoint.x));
      While anAngle < 0 Do
        anAngle := anAngle + 360;
      While anAngle > 270 Do
        anAngle := anAngle - 360;
      Result := MinValue + Round( (MinValue - MaxValue) * ( Power( 1000, ( anAngle - 225) / 270) - 1) / 999);
      If MaxValue > MinValue
      Then Begin
        If Result > MaxValue
        Then Result := MaxValue
        Else If Result < MinValue
        Then Result := MinValue;
      End
      Else begin
        If Result > MinValue
        Then Result := MinValue
        Else If Result < MaxValue
        Then Result := MaxValue;
      End
    End;

    Function  LinesIntersect( A, B, C, D: TPoint): Boolean;
    //
    // Equations :
    //
    //   l1 :: A + p * ( B - A)
    //   l2 :: C + q + ( D - C)
    //
    //   The lines l1 and l2 intersect when the equation l1 = l2 has
    //   a solution and the point of intersection is between the end
    //   points of both of the lines.
    //
    //   l1 = l2 ->
    //   A + p * ( B - A) = C + q * ( D - C)
    //
    //   p(Bx-Ax) + q(Cx-Dx) = Cx-Ax
    //   p(By-Ay) + q(Cy-Dy) = Cy-Ay
    //
    //   Delta  = (Bx-Ax)(Cy-Dy) - (By-Ay)(Cx-Dx)
    //   Delta1 = (Cx-Ax)(Cy-Dy) - (Cy-Ay)(Cx-Dx)
    //   Delta2 = (Bx-Ax)(Cy-Dy) - (By-Ay)(Cx-Ax)
    //
    //   We have no sulutions if Delta = 0 And ( Delta1 <> 0 Or Delta2 <> 0),
    //   otherwise we have one solution or ininitly many; when we have one solution
    //   we will have to check if this solution falls between the end points
    //   of any of the lines. The solution is the point
    //   (Delta1/Delta, Delta2/Delta)
    Var
      DxBA   : Integer;
      DyBA   : Integer;
      DxCD   : Integer;
      DyCD   : Integer;
      DxCA   : Integer;
      DyCA   : Integer;
      Delta  : Integer;
      Delta1 : Integer;
      Delta2 : Integer;
      p      : Extended;
      q      : Extended;
      Sx1    : Extended;
      Sx2    : Extended;
    Begin
      DxBA   := B.x - A.x;
      DyBA   := B.y - A.y;
      DxCD   := C.x - D.x;
      DyCD   := C.y - D.y;
      DxCA   := C.x - A.x;
      DyCA   := C.y - A.y;
      Delta  := DxBA * DyCD - DyBA * DxCD;
      Delta1 := DxCA * DyCD - DyCA * DxCD;
      Delta2 := DxBA * DyCD - DyBA * DxCA;
      If Delta = 0
      Then Result := ( Delta1 = 0) And ( Delta2 = 0) // this is not ok, we need to check overlap ...
      Else Begin
        p := Delta1 / Delta;
        q := Delta2 / Delta;
        Sx1 := A.x + p * (B.x - A.x);
        Sx2 := C.x + q * (D.x - C.x);
        If A.x > B.x
        Then Exchange( A.x, B.x);
        If C.x > D.x
        Then Exchange( C.x, D.x);
        Result := ( Sx1 >= A.x) And ( Sx1 <= B.x) And (Sx2 >= C.x) And (Sx2 <= D.x);
      End;
    End;

    Function  LineIntersectsRectangle( A, B: TPoint; R: TRect): Boolean;
    //
    //  A line intersects a rectangle when it intersects the rectangles
    //  bounding box, or when the line is somehow contained within
    //  the the rectangle (that is A or B are in R, well actualy for
    //  containment A and B both have to be within R, but A or B is
    //  faster check (and the concern here is not proper contaimnent
    //  per se)).
    //
    Begin
      With R Do
        Result :=
          // Containment of A or B in R
          (
            PtInRect( R, A) Or
            PtInRect( R, B)
          )
          Or
          // Intersection of the bounding box
          (
            LinesIntersect( A, B, Point( Left , Top   ), Point( Right, Top   )) Or
            LinesIntersect( A, B, Point( Right, Top   ), Point( Right, Bottom)) Or
            LinesIntersect( A, B, Point( Right, Bottom), Point( Left , Bottom)) Or
            LinesIntersect( A, B, Point( Left , Bottom), Point( Left , Top   ))
          );
    End;




Var

  GAnchor       : TPoint     = ( x: 0; y: 0);
  GDistance     : TPoint     = ( x: 0; y: 0);
  GDragged      : TComponent = Nil;
  GLastDragged  : TComponent = Nil;
  GAutoCreation : Boolean    = False;



{ ========
  TCustomDetail = Class( TControl)
  Private
      FLeft        : Integer;
      FTop         : Integer;
      FWidth       : Integer;
      FHeight      : Integer;
      FColor       : TColor;
      FSnap        : TPoint;
      FDefaultSnap : Boolean;
  Public
    Property    Left         : Integer       Read FLeft           Write SetLeft;
    Property    Top          : Integer       Read FTop            Write SetTop;
    Property    Width        : Integer       Read FWidth          Write SetWidth;
    Property    Height       : Integer       Read FHeight         Write SetHeight;
    Property    ClientRect   : TRect         Read GetClientRect;
    Property    ParentRect   : TRect         Read GetParentRect;
    Property    ClientOrigin : TPoint        Read GetClientOrigin;
    Property    Color        : TColor        Read GetColor        Write SetColor;
    Property    DefaultSnap  : Boolean       Read FDefaultSnap    Write SetDefaultSnap Default True;
    Property    SnapX        : Integer       Read GetSnapX        Write SetSnapX;
    Property    SnapY        : Integer       Read GetSnapY        Write SetSnapY;
  Private
}

    Procedure   TMimCustomDetail.DoMouseDown( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // Virtual;
    Begin
      If ( GDragged = Nil) And ( Button = mbLeft)
      Then Begin
        GDragged     := Self;
        GLastDragged := GDragged;
        GAnchor      := CenterPoint;
        GDistance    := Point( 0, 0);
        BeginMove ( GAnchor  );
        UpdateMove( GDistance);
      End;
    End;

    Procedure   TMimCustomDetail.DoMouseUp( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // Virtual;
    Begin
      If GDragged = Self
      Then Begin
        UpdateMove( GDistance);
        GDistance.x := X - GAnchor.x;
        GDistance.y := Y - GAnchor.y;
        GDragged    := Nil;
        EndMove( GDistance);
      End;
    End;

    procedure   TMimCustomDetail.DoMouseMove( Sender: TObject; Shift: TShiftState; X, Y: Integer); // Virtual;
    Begin
      If GDragged = Self
      Then Begin
        UpdateMove( GDistance);
        GDistance.x := X - GAnchor.x;
        GDistance.y := Y - GAnchor.y;
        UpdateMove( GDistance);
      End;
    End;

    Procedure   TMimCustomDetail.DoClick( Sender: TObject); // Virtual;
    Begin
      SendClick( Self);
    End;

    Procedure   TMimCustomDetail.DoDblClick( Sender: TObject); // Virtual;
    Begin
      SendDblClick( Self);
    End;

    Procedure   TMimCustomDetail.SendClick( Sender: TObject); // Virtual;
    Begin
      If Assigned( Owner)
      Then
        If Owner Is TMimCustomModule
        Then TMimCustomModule( Owner).SendClick( Sender)
        Else If Owner Is TMimDrawSheet
        Then TMimDrawSheet( Owner).SendClick( Sender);
    End;

    Procedure   TMimCustomDetail.SendDblClick( Sender: TObject); // Virtual;
    Begin
      If Assigned( Owner)
      Then
        If Owner Is TMimCustomModule
        Then TMimCustomModule( Owner).SendDblClick( Sender)
        Else If Owner Is TMimDrawSheet
        Then TMimDrawSheet( Owner).SendDblClick( Sender);
    End;

    Function    TMimCustomDetail.CenterPoint: TPoint;
    Begin
      With Result Do
      Begin
        x := Width  Div 2;
        y := Height Div 2;
      End;
    End;

    Procedure   TMimCustomDetail.InvalidateRectangle( aRect: TRect; IsOpaque: Boolean); // Virtual;
    Begin
      If Assigned( Owner) And ( Owner Is TMimCustomModule)
      Then
        TMimCustomModule( Owner).
          InvalidateRectangle( MoveRect( aRect, Point( Left, Top)), IsOpaque);
    End;

    Procedure   TMimCustomDetail.Update;  // Virtual;
    Begin
      Invalidate;
      If Assigned( Owner) And ( Owner Is TMimCustomModule)
      Then TMimCustomModule( Owner).UpdateDetail;
    End;

    Procedure   TMimCustomDetail.SetSizeAndPos( R: TRect); // Virtual;
    Var
      aPoint : TPoint;
      W      : Integer;
      H      : Integer;
      Old    : TRect;
    Begin
      Old := ParentRect;
      W   := R.Right  - R.Left;
      H   := R.Bottom - R.Top;
      If FDefaultSnap
      Then FSnap := Point( W, H);
      aPoint := Point( R.Left, R.Top);
      Snap( aPoint, FSnap);
      If
        ( FLeft   <> aPoint.x) Or
        ( FTop    <> aPoint.y) Or
        ( FWidth  <> W       ) Or
        ( FHeight <> H       )
      Then Begin
        FLeft   := aPoint.x;
        FTop    := aPoint.y;
        FWidth  := W;
        FHeight := H;
        If Assigned( Owner)
        Then Begin
          If Owner Is TMimCustomModule
          Then Begin
            TMimCustomModule( Owner).InvalidateRectangle( Old, True);
            TMimCustomModule( Owner).UpdateDetail;
          End
          Else If Owner Is TMimDrawSheet
          Then TMimDrawSheet( Owner).InvalidateRectangleDetail( Old, True);
        End;
        Invalidate;
      End;
    End;

    Procedure   TMimCustomDetail.SetDefaultSnap( aValue: Boolean);
    Begin
      If FDefaultSnap <> aValue
      Then Begin
        FDefaultSnap := aValue;
        Invalidate;
      End;
    End;

    Function    TMimCustomDetail.GetSnapX: Integer;
    Begin
      Result := FSnap.x;
    End;

    Procedure   TMimCustomDetail.SetSnapX( aValue: Integer);
    Begin
      If aValue < 1
      Then aValue := 1;
      With FSnap Do
      Begin
        If x <> aValue
        Then Begin
          x := aValue;
          DefaultSnap := False;
        End;
      End;
    End;

    Function    TMimCustomDetail.GetSnapY: Integer;
    Begin
      Result := FSnap.y;
    End;

    Procedure   TMimCustomDetail.SetSnapY( aValue: Integer);
    Begin
      If aValue < 1
      Then aValue := 1;
      With FSnap Do
      Begin
        If y <> aValue
        Then Begin
          y := aValue;
          DefaultSnap := False;
        End;
      End;
    End;

    Procedure   TMimCustomDetail.SetPen( aValue: TPen);
    Begin
      FPen.Assign( aValue);
      Invalidate;
    End;

    Procedure   TMimCustomDetail.SetBrush( aValue: TBrush);
    Begin
      FBrush.Assign( aValue);
      Invalidate;
    End;

    Procedure   TMimCustomDetail.SetLeft( aValue: Integer);
    Var
      R : TRect;
    Begin
      If FLeft <> aValue
      Then Begin
        R := ParentRect;
        R.Left := aValue;
        R.Right := R.Left + Width;
        SetSizeAndPos( R);
      End;
    End;

    Procedure   TMimCustomDetail.SetTop( aValue: Integer);
    Var
      R : TRect;
    Begin
      If FTop <> aValue
      Then Begin
        R := ParentRect;
        R.Top := aValue;
        R.Bottom := R.Top + Height;
        SetSizeAndPos( R);
      End;
    End;

    Procedure   TMimCustomDetail.SetWidth( aValue: Integer);
    Var
      R : TRect;
    Begin
      If FWidth <> aValue
      Then Begin
        R := ParentRect;
        R.Right := R.Left + aValue;
        SetSizeAndPos( R);
      End;
    End;

    Procedure   TMimCustomDetail.SetHeight( aValue: Integer);
    Var
      R : TRect;
    Begin
      If FHeight <> aValue
      Then Begin
        R := ParentRect;
        R.Bottom := R.Top + aValue;
        SetSizeAndPos( R);
      End;
    End;

    Function    TMimCustomDetail.GetParentRect: TRect;
    Begin
      Result := Rect( Left, Top, Left + Width, Top + Height);
    End;

    Function    TMimCustomDetail.GetClientRect: TRect;
    Begin
      Result := Rect( 0, 0, Width, Height);
    End;

    Function    TMimCustomDetail.GetScreenRect: TRect;
    Begin
      With Result Do
      Begin
        TopLeft     := ClientOrigin;
        BottomRight := AddPoints( ClientOrigin, Point( Width, Height));
      End;
    End;

    Function    TMimCustomDetail.GetClientOrigin: TPoint;
    Begin
//      Result := ClientToScreen( Point( Left, Top));
      Result := ClientToScreen( Point( 0, 0));
    End;

    Procedure   TMimCustomDetail.SetColor( aValue: TColor);
    Begin
      If Assigned( Brush)
      Then Begin
        Brush.Color := aValue;
        Invalidate;
      End;
    End;

    Function    TMimCustomDetail.GetColor: TColor;
    Begin
      If Assigned( Brush)
      Then Result := Brush.Color
      Else Result := clBlack;
    End;

{  Protected }

    Procedure   TMimCustomDetail.BeginMove( aPoint: TPoint); // Virtual;
    Begin
    End;

    Procedure   TMimCustomDetail.EndMove( aPoint: TPoint); // Virtual;
    Begin
    End;

    Procedure   TMimCustomDetail.UpdateMove( aPoint: TPoint); // Virtual
    Begin
    End;

    Procedure   TMimCustomDetail.Invalidate; // Virtual;
    Begin
      InvalidateRectangle( ClientRect, False);
    End;

    Procedure   TMimCustomDetail.PaintOn( aCanvas: TCanvas); // Virtual;
    Begin
      With aCanvas Do
      Begin
        Pen   := Self.Pen;
        Brush := Self.Brush;
        Rectangle( 0, 0, Width, Height);
      End;
    End;

    Function    TMimCustomDetail.ContainsPoint( aPoint: TPoint): Boolean; // Virtual;
    Begin
      Result := PtInRect( ClientRect, aPoint);
    End;

    Function    TMimCustomDetail.ClientToScreen( aPoint: TPoint): TPoint;
    Begin
      If Assigned( Owner) And ( Owner Is TMimCustomModule) And HasParent And ( Owner.HasPArent)
      Then Result := AddPoints( TMimCustomModule( Owner).ClientToScreen( aPoint), Point( Left, Top))
      Else Result := AddPoints( aPoint, Point( Left, Top));
    End;

    Function    TMimCustomDetail.ScreenToClient( aPoint: TPoint): TPoint;
    Begin
      If Assigned( Owner) And ( Owner Is TMimCustomModule) And HasPArent And Owner.HasParent
      Then Result := AddPoints( TMimCustomModule( Owner).ScreenToClient( aPoint), Point( - Left, - Top))
      Else Result := AddPoints( aPoint, Point( - Left, - Top));
    End;

    Procedure   TMimCustomDetail.SetParentComponent( aValue: TComponent); // Override;
    Begin
      Assert( aValue = Owner, 'Didn''t expect parent to be not the owner ...')
    End;

    procedure   TMimCustomDetail.AssignTo( aDest: TPersistent); // Override;
    Begin
      If Assigned( aDest) And ( aDest Is TMimCustomDetail)
      Then Begin
        TMimCustomDetail( aDest).Left  := Left;
        TMimCustomDetail( aDest).Top   := Top;
        TMimCustomDetail( aDest).Pen   .Assign( Pen);
        TMimCustomDetail( aDest).Brush .Assign( Brush);
      End;
    End;

{  Public }

    Constructor TMimCustomDetail.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      FPen   := TPen.Create;
      FBrush := TBrush.Create;
      Brush.Color := clBlue;
      Pen  .Color := clBlack;
      FSnap  := Point( 1, 1);
      Width  := 25;
      Height := 25;
      Name   := UniqueName;
    End;

    Destructor  TMimCustomDetail.Destroy; // Override;
    Begin
      FPen  .Free;
      FBrush.Free;
      Inherited Destroy;
    End;

    Function    TMimCustomDetail.UniqueName: String; // Virtual;
    Var
      i : Integer;
      S : String;
    Begin
      Result := '';
      S      := Copy( ClassName, 5, Length( ClassName));
      If Owner <> Nil
      Then With Owner Do
        For i := 0 To ComponentCount - 1 Do
          If FindComponent( Format( '%s%d', [ S, i])) = Nil
          Then Begin
            Result := Format( '%s%d', [ S, i]);
            Break;
          End;
    End;

    Procedure   TMimCustomDetail.Dump( Var aFile: TextFile; anInd, aDepth: Integer); // Virtual;
    Begin
      WriteLn( aFile, Indent( anInd, Format( '%s : ''%s''', [ ClassName, Name])));
    End;

    Procedure   TMimCustomDetail.Recolorize; // Virtual;
    Begin
      Invalidate;
    End;

    Procedure   TMimCustomDetail.SnapToGrid( aPoint: TPoint);
    Begin
      Snap( aPoint, FSnap);
      Left := aPoint.x;
      Top  := aPoint.y;
    End;

    Function    TMimCustomDetail.DragAnchor: TPoint;
    Begin
      Result := GAnchor;
    End;

    Function    TMimCustomDetail.DragDistance: TPoint;
    Begin
      Result := GDistance;
    End;

    Function    TMimCustomDetail.DragObject: TComponent;
    Begin
      Result := GDragged;
    End;

    Function    TMimCustomDetail.LastDragObject: TComponent;
    Begin
      Result := GLastDragged;
    End;

    Function    TMimCustomDetail.GetParentComponent: TComponent; // Override;
    Begin
      Result := Owner;
    End;

    Function    TMimCustomDetail.HasParent: Boolean; // Override;
    Begin
      Result := True;
    End;

{
  // a semi-abstract base type for details without mouse interaction
  // (such as labels).
  //
  TNoMouse = Class( TCustomDetail)
  Protected
}

    Function    TMimNoMouse.ContainsPoint( aPoint: TPoint): Boolean; // Override;
    Begin
      // No mouse interaction.
      Result := False;
    End;

{ ========
  // a semi-abstract base type for invisible details
  // (such as timers).
  //
  TInvisible = Class( TCustomDetail)
  Protected
}

    Procedure   TMimInvisible.PaintOn( aCanvas: TCanvas); // Override;
    Begin
      If csDesigning In ComponentState
      Then Inherited PaintOn( aCanvas);
    End;


{ ========
  TConnector = Class( TCustomDetail)
  Private
      FPlugColor       : TColor;
      FConnectionCount : Integer;
      FValue           : TValue;
      FCalcCount       : Integer;
      FConnectorType   : TConnectorType;
  Public
    Property    PlugColor     : TColor         Read FPlugColor     Write SetPlugColor Default clConnector;
    Property    Value         : TValue         Read FValue         Write SetValue;
    Property    Connected     : Boolean        Read GetConnected;
    Property    ConnectorType : TConnectorType Read FConnectorType Write SetConnectorType;
  Private
}

{  Protected }

    Procedure   TMimConnector.SetPlugColor( aValue: TColor);
    Begin
      If aValue <> FPlugColor
      Then Begin
        FPlugColor := aValue;
        Invalidate;
      End;
    End;

    Procedure   TMimConnector.SetValue( aValue: TMimValue);
    Begin
      If FValue <> aValue
      Then Begin
        FValue := aValue;
        ValueChanged( FValue);
      End;
    End;

    Procedure   TMimConnector.SetConnectorType( aValue: TConnectorType);
    Begin
      If aValue <> FConnectorType
      Then Begin
        FConnectorType := aValue;
        PlugColor := ConnectorTypeToColor( FConnectorType);
      End;
    End;

    Function    TMimConnector.GetConnected: Boolean;
    Begin
      Result := FConnectionCount > 0;
    End;

    Procedure   TMimConnector.EndMove( aPoint: TPoint); // Override;
    Var
      aDst : TMimConnector;
    Begin
      aDst := SearchDestination;
      If Assigned( aDst) And aDst.CanAccept( Self)
      Then Connect( aDst);
    End;

    Procedure   TMimConnector.UpdateMove( aPoint: TPoint); // Override;
    Var
      aDC    : HDC;
      aPen   : HPen;
      OldObj : HGDIOBJ;
      aDst   : TMimConnector;
      aColor : TColor;
    Begin
      aDst := SearchDestination;
      If Assigned( aDst) And aDst.CanAccept( Self)
      Then aColor := clGreen
      Else aColor := clWhite;
      aDC := GetDC( 0);

      If Assigned( Owner) And ( Owner Is TMimCustomModule)
      Then
        With TMimCustomModule( Owner) Do
          If Assigned( Owner)
          Then Begin
            If Owner Is TMimCustomModule
            Then
              With TMimCustomModule( Owner) Do
                With ScreenRect Do
                  IntersectClipRect( aDC, Left, Top, Right, Bottom)
            Else If Owner Is TMimDrawSheet
            Then
              With TMimDrawSheet( Owner) Do
                With ClientOrigin Do
                  IntersectClipRect( aDC, X, Y, X + ClientWidth, Y + ClientHeight)
          End;

      Try
        SetRop2( aDC, R2_XORPEN);
        aPen   := CreatePen( PS_SOLID, 0, ColorToRGB( aColor));
        OldObj := SelectObject( aDC, aPen);
        Try
          With ClientToScreen( DragAnchor) Do
          Begin
            MoveToEx( aDC, x, y, Nil);
            LineTo( aDC, x + aPoint.x, y + aPoint.y);
          End;
        Finally
          SelectObject( aDC, OldObj);
          DeleteObject( aPen);
        End;
      Finally
        ReleaseDC( 0, aDC);
      End;
    End;

    Procedure   TMimConnector.PaintOn( aCanvas: TCanvas); // Override;
    Begin
      Brush.Color := PlugColor;
      With aCanvas Do
      Begin
        Brush := Self.Brush;
        Pen   := Self.Pen;
        Rectangle( 0, 0, Width, Height);
      End;
    End;

    Procedure   TMimConnector.Connect( aConnector: TMimConnector); // Virtual;
    Begin
      If
        Assigned( Owner)          And
        ( Owner Is TMimCustomModule) And
        ( aConnector <> Self    )
      Then
        With TMimCustomModule( Owner) Do
          Connect( Self, aConnector);
    End;

    Function    TMimConnector.CanAccept( aConnector: TMimConnector): Boolean; // Virtual;
    Begin
      Result := aConnector Is TMimConnector;
    End;

    Function    TMimConnector.SearchDestination: TMimConnector;
    Var
      aPoint: TPoint;
    Begin
      aPoint := AddPoints( DragAnchor, DragDistance);
      If ContainsPoint( aPoint)
      Then Result := Self
      Else
        If Assigned( Owner) And ( Owner Is TMimCustomModule)
        Then
          Result :=
            TMimCustomModule( Owner).
              SearchDestination( Self, AddPoints( aPoint, Point( Left, Top)))
        Else
          Result := Nil;
    End;

    Procedure   TMimConnector.ValueChanged( Var aValue: TMimValue); // Virtual;
    Begin
      // Nothing here, it's for the children.
    End;

    procedure   TMimConnector.AssignTo( aDest: TPersistent); // Override;
    Begin
      Inherited AssignTo( aDest);
      If Assigned( aDest) And ( aDest Is TMimConnector)
      Then TMimConnector( aDest).ConnectorType := ConnectorType;
    End;

{   Public }

    Constructor TMimConnector.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      Width         := 15;
      Height        := 15;
      FPlugColor    := clConnector;
      ConnectorType := ctAsymmetric;
      Brush.Color   := FPlugColor;
      Pen  .Color   := clBlack;
    End;

    Procedure   TMimConnector.Recolorize; // Override;
    Begin
      PlugColor := ConnectorTypeToColor( ConnectorType);
      Inherited;
    End;

    Procedure   TMimConnector.IncConnectionCount;
    Begin
      Inc( FconnectionCount);
      If FConnectionCount = 1
      Then Invalidate;
    End;

    Procedure   TMimConnector.DecConnectionCount;
    Begin
      If FConnectionCount > 0
      Then Begin
        Dec( FconnectionCount);
        If FConnectionCount = 0
        Then Invalidate;
      End;
    End;

{ ========
  TInput = Class( TConnector)
  Protected
}

    Procedure   TMimInput.PaintOn( aCanvas: TCanvas); // Override;
    Begin
      Brush.Color := PlugColor;
      With aCanvas Do
      Begin
        Pen   := Self.Pen;
        Brush := Self.Brush;
        Ellipse( 0, 0, Width, Height);
        With Brush Do
          If FConnectionCount = 0
          Then Color := clBlack
          Else Color := FPlugColor;
        Ellipse( 4, 4, Width - 4, Height - 4);
      End;
    End;

    Function    TMimInput.CanAccept( aConnector: TMimConnector): Boolean; // Override;
    Begin
      CanAccept := aConnector Is TMimOutput;
    End;


{ ========
  TOutput = Class( TConnector)
  Protected
}

    Procedure   TMimOutput.PaintOn( aCanvas: TCanvas); // Override;
    Begin
      Brush.Color := PlugColor;
      With aCanvas Do
      Begin
        Pen   := Self.Pen;
        Brush := Self.Brush;
        Rectangle( 0, 0, Width    , Height    );
        Ellipse  ( 1, 1, Width - 1, Height - 1);
        With Brush Do
          If FConnectionCount = 0
          Then Color := clBlack
          Else Color := FPlugColor;
        Ellipse( 4, 4, Width - 4, Height - 4);
      End;
    End;

    Function    TMimOutput.CanAccept( aConnector: TMimConnector): Boolean; // Override;
    Begin
      Result := aConnector Is TMimInput;
    End;

{
  TTextLabel = Class( TNoMouse)
  Private
      FText        : String;
      FFont        : TFont;
      FAutoSize    : Boolean;
      FParentColor : Boolean;
      FClickable   : Boolean;
  Public
    Property    Text        : String  Read FText        Write SetText;
    Property    TextColor   : TColor  Read GetTextColor Write SetTextColor;
    Property    Font        : TFont   Read FFont        Write SetFont;
    Property    AutoSize    : Boolean Read FAutoSize    Write SetAutoSize;
    Property    ParentColor : Boolean Read FParentColor Write SetParentColor;
  Private
}

    Procedure   TMimTextLabel.AdjustBounds;
    var
      aDC       : HDC;
      aSaveFont : HFont;
      aTextSize : TSize;
    begin
      aDC := GetDC( 0);
      aSaveFont := SelectObject( aDC, Font.Handle);
      GetTextExtentPoint32( aDC, PChar( FText), Length( FText), aTextSize);
      SelectObject( aDC, aSaveFont);
      ReleaseDC( 0, aDC);
      SetSizeAndPos(
        Rect(
          Left,
          Top,
          Left + aTextSize.cx + ( GetSystemMetrics( SM_CXBORDER) * 2),
          Top  + aTextSize.cy + ( GetSystemMetrics( SM_CYBORDER) * 1)
        )
      );
    End;

    Procedure   TMimTextLabel.SetText( Const aValue: String);
    Begin
      If FText <> aValue
      Then Begin
        FText := aValue;
        If FAutoSize
        Then AdjustBounds;
        InvalidateRectangle( InflateRectangle( ClientRect, - 1), False);
      End;
    End;

    Procedure   TMimTextLabel.SetTextColor( aValue: TColor);
    Begin
      If TextColor <> aValue
      Then Begin
        Font.Color := aValue;
        InvalidateRectangle( InflateRectangle( ClientRect, - 1), False);
      End;
    End;

    Function    TMimTextLabel.GetTextColor: TColor;
    Begin
      Result := Font.Color;
    End;

    Procedure   TMimTextLabel.SetFont( aValue: TFont);
    Begin
      Font.Assign( aValue);
      If FAutoSize
      Then AdjustBounds;
      InvalidateRectangle( InflateRectangle( ClientRect, - 1), False);
    End;

    Procedure   TMimTextLabel.SetAutoSize( aValue: Boolean);
    Begin
      If FAutoSize <> aValue
      Then Begin
        FAutoSize := aValue;
        If FAutoSize
        Then Begin
          AdjustBounds;
          Invalidate;
        End;
      End;
    End;

    Procedure   TMimTextLabel.SetParentColor( aValue: Boolean);
    Begin
      If aValue <> FParentColor
      Then Begin
        FParentColor := aValue;
        Invalidate;
      End;
    End;

{  Protected }

    Function    TMimTextLabel.ContainsPoint( aPoint: TPoint): Boolean; // Override;
    Begin
      Result := FClickable And PtInRect( ClientRect, aPoint);
    End;

    Procedure   TMimTextLabel.PaintOn( aCanvas: TCanvas); // Override;
    Begin
      Pen.Color := Brush.Color;
      Inherited PaintOn( aCanvas);
      With aCanvas Do
      Begin
        If Not FParentColor
        Then Brush := Self.Brush;
        Font := Self.Font;
        If FAutoSize
        Then Begin
          With TextExtent( FText) Do
          Begin
            Self.FWidth  := cx + 6;
            Self.FHeight := cy + 2;
          End;
        End;
        TextOut( 0, 0, FText);
      End;
    End;

    procedure   TMimTextLabel.AssignTo( aDest: TPersistent); // Override;
    Begin
      Inherited AssignTo( aDest);
      If Assigned( aDest) And ( aDest Is TMimTextLabel)
      Then Begin
        TMimTextLabel( aDest).Text        := Text;
        TMimTextLabel( aDest).Font        .Assign( Font);
        TMimTextLabel( aDest).AutoSize    := AutoSize;
        TMimTextLabel( aDest).ParentColor := ParentColor;
      End;
    End;

{  Public }

    Constructor TMimTextLabel.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      FFont := TFont.Create;
      With FFont Do
      Begin
        Name   := 'Arial';
        Height := 13;
        Style  := [];
      End;
      Color       := ModuleColor;
      TextColor   := clBlack;
      Width       := 90;
      Height      := 15;
      AutoSize    := False;
      ParentColor := True;
    End;

    Destructor  TMimTextLabel.Destroy; // Override;
    Begin
      FFont.Free;
      Inherited Destroy;
    End;

    Procedure   TMimTextLabel.Recolorize; // Override;
    Begin
      Color := ModuleColor;
      Inherited;
    End;

{ ========
  TViewer = Class( TCustomDetail)
  Private
      FText : String;
      FFont : TFont;
  Public
    Property    Text        : String Read FText          Write SetText;
    Property    BorderColor : TColor Read GetBorderColor Write SetBorderColor;
    Property    TextColor   : TColor Read GetTextColor   Write SetTextColor;
    Property    Font        : TFont  Read FFont          Write SetFont;
  Private
}

    Procedure   TMimViewer.SetText( Const aValue: String);
    Begin
      If FText <> aValue
      Then Begin
        FText := aValue;
        InvalidateRectangle( InflateRectangle( ClientRect, - 2), True);
        If Assigned( Owner) And ( Owner Is TMimCustomModule)
        Then TMimCustomModule( Owner).UpdateDetail;
      End;
    End;

    Procedure   TMimViewer.SetBorderColor( aValue: TColor);
    Begin
      If BorderColor <> aValue
      Then Begin
        Pen.Color := aValue;
        Invalidate;
      End;
    End;

    Function    TMimViewer.GetBorderColor: TColor;
    Begin
      Result := Pen.Color;
    End;

    Procedure   TMimViewer.SetTextColor( aValue: TColor);
    Begin
      If TextColor <> aValue
      Then Begin
        Font.Color := aValue;
        InvalidateRectangle( InflateRectangle( ClientRect, - 1), False);
      End;
    End;

    Function    TMimViewer.GetTextColor: TColor;
    Begin
      Result := Font.Color;
    End;

    Procedure   TMimViewer.SetFont( aValue: TFont);
    Begin
      Font.Assign( aValue);
      InvalidateRectangle( InflateRectangle( ClientRect, - 1), False);
    End;

{  Protected }

    Procedure   TMimViewer.PaintOn( aCanvas: TCanvas); // Override;
    Begin
      With aCanvas Do
      Begin
        Pen   := Self.Pen;
        Brush := Self.Brush;
        Font  := Self.Font;
        Rectangle( 0, 0, Width, Height);
        TextOut( 4, 1, FText);
      End;
    End;

    procedure   TMimViewer.AssignTo( aDest: TPersistent); // Override;
    Begin
      Inherited AssignTo( aDest);
      If Assigned( aDest) And ( aDest Is TMimViewer)
      Then Begin
        TMimViewer( aDest).Text        := Text;
        TMimViewer( aDest).BorderColor := BorderColor;
        TMimViewer( aDest).TextColor   := TextColor;
      End;
    End;

{  Public }

    Constructor TMimViewer.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      FFont       := TFont.Create;
      Width       := 35;
      Height      := 15;
      Color       := ViewerColor;
      BorderColor := clWhite;
      TextColor   := clWhite;
      With Font Do
      Begin
        Height := 12;
        Style  := [];
      End;
    End;

    Destructor  TMimViewer.Destroy; // Override;
    Begin
      FFont.Free;
      Inherited Destroy;
    End;

    Procedure   TMimViewer.Recolorize; // Override;
    Begin
      Color       := ViewerColor;
      BorderColor := clWhite;
      TextColor   := clWhite;
      Inherited;
    End;

{ ========
  TKnob = Class( TCustomDetail)
  Private
      FValue         : Integer;
      FMinValue      : Integer;
      FMaxValue      : Integer;
      FViewer        : TViewer;
      FOnValueChange : TOnIntegerValueChange;
      FOldP1         : TPoint;
      FOldP2         : TPoint;
      FKnobMode      : TKnobMode;
      FRepeatTimer   : TTimer;
      FRepeatCount   : Integer;
  Public
    Property    Value        : Integer               Read FValue         Write SetValue;
    Property    MinValue     : Integer               Read FMinValue      Write SetMinValue;
    Property    MaxValue     : Integer               Read FMaxValue      Write SetMaxValue;
    Property    Viewer       : TViewer               Read FViewer        Write SetViewer;
    Property    KnobMode     : TKobMode              Read FKnobMode;
    Property    OnValueChange: TOnIntegerValueChange Read FOnValueChange Write SetOnValueChange;
  Private
}

    Procedure   TMimKnob.DoMouseDown( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // Virtual;
    Begin
      If ( GDragged = Nil) And ( Button = mbLeft)
      Then Begin
        GDragged     := Self;
        GLastDragged := GDragged;
        GAnchor      := Point( X, Y);
        GDistance    := Point( X, Y);
        BeginMove ( GAnchor  );
        UpdateMove( GDistance);
      End
      Else If Button = mbRight
      Then Begin
        FKnobMode := kmNone;
        DoDblClick( Self);
      End;
    End;

    Procedure   TMimKnob.DoMouseUp( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // Virtual;
    Begin
      If GDragged = Self
      Then Begin
        UpdateMove( GDistance);
        GDistance.x := X - GAnchor.x;
        GDistance.y := Y - GAnchor.y;
        GDragged    := Nil;
        EndMove( GDistance);
      End;
    End;

    procedure   TMimKnob.DoMouseMove( Sender: TObject; Shift: TShiftState; X, Y: Integer); // Virtual;
    Begin
      If GDragged = Self
      Then Begin
        UpdateMove( GDistance);
        GDistance.x := X - GAnchor.x;
        GDistance.y := Y - GAnchor.y;
        UpdateMove( GDistance);
      End;
    End;

    Procedure   TMimKnob.DoTimerFired( aSender: TObject);
    Begin
      Inc( FRepeatCount);
      If FRepeatCount < 50
      Then FRepeatTimer.Interval := RepeatPause
      Else FRepeatTimer.Interval := FastRepeatPause;
      If FKnobMode In [ kmInc,kmDec]
      Then Begin
        Try
          Case FKnobMode Of
            kmInc  : Value := Value + 1;
            kmDec  : Value := Value - 1;
          End;
        Except
          FRepeatTimer.Enabled := False;
          FRepeatCount := 0;
          Raise;
        End;
      End;
    End;

    Procedure   TMimKnob.UpdatePeers;
    Begin
      If Assigned( FViewer)
      Then FViewer.Text := IntToStr( Value);
      If Assigned( FOnValueChange)
      Then FOnValueChange( Self, FValue);
    End;

    Procedure   TMimKnob.Update; // Override;
    Begin
      UpdatePeers;
      Inherited Update;
    End;

    Procedure   TMimKnob.SetValue( aValue: TMimValue);
    Var
      R1  : TRect;
      R2  : TRect;
      P1  : TPoint;
      P2  : TPoint;
      Ang : Integer;
    Begin
      If aValue <> FValue
      Then Begin
        Ang := ValuesToAngle( MinValue, Value, MaxValue);
        P1  := PolarToPoint( CenterPoint, Ang, 3);
        P2  := PolarToPoint( CenterPoint, Ang, 9);
        If P1.x > P2.x
        Then Exchange( P1.x, P2.x);
        If P1.y > P2.y
        Then Exchange( P1.y, P2.y);

        If aValue < FMinValue
        Then FValue := FMinValue
        Else If aValue > FMaxValue
        Then FValue := FMaxValue
        Else FValue := aValue;

        UpdatePeers;

        If
          ( P1.x  <> FOldP1.x) Or
          ( P1.y  <> FOldP1.y) Or
          ( P2.x  <> FOldP2.x) Or
          ( P2.y  <> FOldP2.y)
        Then Begin
          R1 := InflateRectangle( Rect( P1.x    , P1.y    , P2.x    , P2.y)    , 2);
          R2 := InflateRectangle( Rect( FOldP1.x, FOldP1.y, FOldP2.x, FOldP2.y), 2);
          InvalidateRectangle( R2, False);
          InvalidateRectangle( R1, False);
          FOldP1 := P1;
          FOldP2 := P2;
        End;

//        Inherited Update;
      End;
    End;

    Procedure   TMimKnob.SetStep( aValue: TMimValue);
    begin
      if aValue <> FStep 
      then begin
        FStep := aValue;
      end;
    end;
    
    Procedure   TMimKnob.SetMinValue( aValue: TMimValue);
    Begin
      If aValue <> FMinValue
      Then Begin
        FMinValue := aValue;
        If FValue < FMinValue
        Then FValue := FMinValue;
        Update;
      End;
    End;

    Procedure   TMimKnob.SetMaxValue( aValue: TMimValue);
    Begin
      If aValue <> FMaxValue
      Then Begin
        FMaxValue := aValue;
        If FValue > FMaxValue
        Then FValue := FMaxValue;
        Update;
      End;
    End;

    Procedure   TMimKnob.SetViewer( aValue: TMimViewer);
    Begin
      If aValue <> FViewer
      Then Begin
        FViewer := aValue;
        If Assigned( FViewer)
        Then Update;
      End;
    End;

    Procedure   TMimKnob.SetOnValueChange( aValue: TOnValueValueChange);
    Begin
      FOnValueChange := aValue;
      If Assigned( FOnValueChange)
      Then Update;
    End;

{  Protected }

    Procedure   TMimKnob.BeginMove( aPoint: TPoint); // Override;
    Const
      CRange   = 8;
      CRadius  = 25;
    Var
      R1 : TRect;
      R2 : TRect;
      P  : TPoint;
    Begin
      R1 := Rect( 0             , Height - CRange, CRange, Height);
      R2 := Rect( Width - CRange, Height - CRange, Width , Height);

      If PtInRect( R1, aPoint)
      Then FKnobMode := kmDec
      Else If PtInRect( R2, aPoint)
      Then FKnobMode := kmInc
      Else FKnobMode := kmTurn;

      If FKnobMode = kmTurn
      Then Begin
        P          := PolarToPoint( CenterPoint, ValuesToAngle( MinValue, Value, MaxValue), CRadius);
        With ClientToScreen( P) Do SetCursorPos( X, Y);
        GDistance  := SubPoints( CenterPoint, P);
        GAnchor    := CenterPoint;
      End
      Else Begin
        If Not Assigned( FRepeatTimer)
        Then FRepeatTimer := TTimer.Create( Self);
        With FRepeatTimer Do
        Begin
          OnTimer  := DoTimerFired;
          Interval := InitRepeatPause;
          Enabled  := True;
        End;
        FRepeatCount := 0;
      End;
    End;

    Procedure   TMimKnob.EndMove( aPoint: TPoint); // Override;
    Var
      aValue : TMimValue;
    Begin
      FRepeatCount := 0;
      If Assigned( FRepeatTimer)
      Then FRepeatTimer.Enabled := False;
      Case FKnobMode Of
        kmInc  : aValue := Value + 1;
        kmDec  : aValue := Value - 1;
        Else     aValue := DistanceToValue( MinValue, MaxValue, aPoint);
      End;
      If aValue < MinValue
      Then Value := MinValue
      Else If aValue > MaxValue
      Then Value := MaxValue
      Else Value := aValue;
      FKnobMode := kmNone;
    End;

    Procedure   TMimKnob.UpdateMove( aPoint: TPoint); // Override;
    Begin
      Case FKnobMode Of
        kmTurn : Value := DistanceToValue( MinValue, MaxValue, aPoint);
        kmInc  : {Nothing};
        kmDec  : {Nothing};
      End;
    End;

    Procedure   TMimKnob.PaintOn( aCanvas: TCanvas); // Override;
    Var
      S : TPoint;
      E : TPoint;
      A : Integer;
    Begin
      With aCanvas Do
      Begin
        Pen   := Self.Pen;
        Brush := Self.Brush;
        Ellipse( 0, 0, Width, Height);
        Pen.Width := 2;
        A := ValuesToAngle( MinValue, Value, MaxValue);
        S := PolarToPoint( CenterPoint, A, 3);
        E := PolarToPoint( CenterPoint, A, 9);
        With aCanvas Do
        Begin
          MoveTo( S.x, S.y);
          LineTo( E.x, E.y);
        End;
        Pen.Width := 1;

        MoveTo( 0        , Height - 4);
        LineTo( 0        , Height - 1);
        LineTo( 4        , Height - 1);

        MoveTo( Width - 5, Height - 1);
        LineTo( Width - 1, Height - 1);
        LineTo( Width - 1, Height - 4);
      End;
    End;

    procedure   TMimKnob.AssignTo( aDest: TPersistent); // Override;
    Begin
      Inherited AssignTo( aDest);
      If Assigned( aDest) And ( aDest Is TMimKnob)
      Then Begin
        TMimKnob( aDest).Value    := Value;
        TMimKnob( aDest).MinValue := MinValue;
        TMimKnob( aDest).MaxValue := MaxValue;
      End;
    End;

{  Public }

    Constructor TMimKnob.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      Width       :=   28;
      Height      :=   28;
      Value       :=  500;
      MinValue    :=    0;
      MaxValue    := 1000;
      Color       := KnobColor;
      Brush.Color := Color;
      Pen  .Color := clWhite;
    End;

    Destructor  TMimKnob.Destroy; // Override;
    Begin
      If Assigned( FRepeatTimer)
      Then FRepeatTimer.Free;
      Inherited Destroy;
    End;

    Procedure   TMimKnob.Recolorize; // Override;
    Begin
      Color       := KnobColor;
      Brush.Color := Color;
      Pen  .Color := clWhite;
      Inherited;
    End;


{ ========    
  TMimKnobSmall = class( TMimKnob)
  Public
}
  
    Constructor TMimKnobSmall.Create( anOwner: TComponent); // Override;
    begin
      inherited;
      Width  :=   16;
      Height :=   16;
    end;

    
{ ========
  TMimKnobMedium = class( TMimKnob)
  Public
}
  
    Constructor TMimKnobMedium.Create( anOwner: TComponent); // Override;
    begin
      inherited;
      Width  :=   20;
      Height :=   20;
    end;

    
{ ========
  TExpKnob = Class( TKnob)
  Protected
}

    Procedure   TMimExpKnob.EndMove( aPoint: TPoint); // Override;
    Begin
      Value := DistanceToExpValue( MinValue, MaxValue, aPoint);
    End;

    Procedure   TMimExpKnob.UpdateMove( aPoint: TPoint); // Override;
    Begin
      Value := DistanceToExpValue( MinValue, MaxValue, aPoint);
    End;


{ ========
  TPushButton = Class( TCustomDetail)
  Private
      FDown          : Boolean;
      FOldDown       : Boolean;
      FBitmap        : TBitmap;
      FOnValueChange : TOnBooleanValueChange;
      FGroupId       : Integer;
      FAllowAllUp    : Boolean;
      FOnClick       : TNotifyEvent;
  Protected
    Property    Bitmap: TBitmap Read FBitmap Write SetBitmap;
  Public
    Property    Down         : Boolean               Read FDown          Write SetDown;
    Property    GroupId      : Integer               Read FGroupId       Write SetGroupId;
    Property    AllowAllUp   : Boolean               Read FAllowAllUp    Write SetAllowAllUp;
    Property    OnValueChange: TOnBooleanValueChange Read FOnValueChange Write SetOnValueChange;
    Property    OnClick      : TNotifyEvent          Read FOnClick       Write FOnClick;
  Private
}

    Procedure   TMimPushButton.SelectUnique;
    Var
      i    : Integer;
      aBut : TMimPushButton;
    Begin
      If Assigned( Owner) And ( Owner Is TMimCustomDetail)
      Then
        With TMimCustomDetail( Owner) Do
          For i := 0 To ComponentCount - 1 Do
            If ( Components[ i] Is TMimPushButton) And ( Components[ i] <> Self)
            Then Begin
              aBut := TMimPushButton( Components[ i]);
              With aBut Do
                If ( GroupId = Self.GroupId) And FDown
                Then Begin
                  FDown := False;
                  If Assigned( FOnValueChange)
                  Then FOnValueChange( aBut, False);
                  Invalidate;
                End;
            End;
    End;

    Procedure   TMimPushButton.SetBitmap( aValue: TBitmap);
    Begin
      With FBitmap Do
      Begin
        Assign( aValue);
        TransparentMode := tmAuto;
        Transparent     := True;
      End;
      Invalidate;
    End;

    Procedure   TMimPushButton.SetDown( aValue: Boolean);
    Begin
      If FDown <> aValue
      Then Begin
        If ( GroupId > 0) And Not aValue And Not AllowAllup
        Then Exit;
        FDown := aValue;
        If Down And ( GroupId > 0)
        Then SelectUnique;
        If Assigned( FOnValueChange)
        Then FOnValueChange( Self, Down);
        Invalidate;
      End;
    End;

    Procedure   TMimPushButton.SetOnValueChange( aValue: TOnBooleanValueChange);
    Begin
      FOnValueChange := aValue;
      If Assigned( FOnValueChange) And ( DragObject <> Self)
      Then FOnValueChange( Self, Down);
    End;

    Procedure   TMimPushButton.SetGroupId( aValue: Integer);
    Begin
      If FGroupId <> aValue
      Then Begin
        FGroupId := aValue;
        If FGroupId > 0
        Then Begin
          If Down
          Then SelectUnique
          Else
            If Not AllowAllUp
            Then Down := True;
        End;
      End;
    End;

    Procedure   TMimPushButton.SetAllowAllUp( aValue: Boolean);
    Begin
      If FAllowAllUp <> aValue
      Then Begin
        FAllowAllUp := aValue;
        If ( GroupId > 0) And Not FAllowAllUp And Not Down
        Then Down := True;
      End;
    End;

{  Protected }

    Procedure   TMimPushButton.BeginMove( aPoint: TPoint); // Override;
    Begin
      GAnchor  := Point( 0, 0);
      FOldDown := Down;
      Down := Not Down;
      Update;
    End;

    Procedure   TMimPushButton.EndMove( aPoint: TPoint); // Override;
    Begin
      If ContainsPoint( aPoint)
      Then Begin
        Down := Not FOldDown;
        If Assigned( FOnClick)
        Then FOnClick( Self);
      End
      Else Down := FOldDown;
      Update;
    End;

    Procedure   TMimPushButton.UpdateMove( aPoint: TPoint); // Override;
    Var
      WasDown : Boolean;
    Begin
      WasDown := Down;
      If ContainsPoint( aPoint)
      Then Down := Not FOldDown
      Else Down := FOldDown;
      If WasDown <> Down
      then Update;
    End;

    Procedure   TMimPushButton.PaintOn( aCanvas: TCanvas); // Override;
    Var
      R1 : TRect;
      R2 : TRect;
      C1 : TColor;
      C2 : TColor;
    Begin
      Pen.Color := clBlack;
      R1 := InflateRectangle( ClientRect, - 1);
      With R1 Do
      Begin
        Dec( Right );
        Dec( Bottom);
      End;
      R2 := InflateRectangle( ClientRect, - 2);
      With aCanvas Do
      Begin
        Self.Pen.Color := ModuleColor;
        Brush := Self.Brush;
        Pen   := Self.Pen;
        Rectangle( 0, 0, Width, Height);
        If Down
        Then Begin
          C1 := clGray;
          C2 := clWhite;
          R2 := MoveRect( R2, Point( 1, 1));
        End
        Else Begin
          C1 := clWhite;
          C2 := clGray;
        End;
        Self.Pen.Color := C1;
        Pen   := Self.Pen;
        With R1 Do
        Begin
          MoveTo( Left , Bottom);
          LineTo( Left , Top   );
          LineTo( Right, Top   );
          Self.Pen.Color := C2;
          Pen := Self.Pen;
          MoveTo( Left , Bottom);
          LineTo( Right, Bottom);
          LineTo( Right, Top   );
        End;
        With R2 Do
        Begin
          MoveWindowOrg( Handle, Left, Top);
          IntersectClipRect( Handle, 0, 0, Right - 4, Bottom - 4);
        End;
        Draw( 0, 0, FBitmap);
      End;
    End;

    procedure   TMimPushButton.AssignTo( aDest: TPersistent); // Override;
    Begin
      Inherited AssignTo( aDest);
      If Assigned( aDest) And ( aDest Is TMimPushButton)
      Then Begin
        TMimPushButton( aDest).Down   := Down;
        TMimPushButton( aDest).Bitmap .Assign( Bitmap);
      End;
    End;

{  Public }

    Constructor TMimPushButton.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      FBitmap := TBitmap.Create;
      FBitmap.TransparentMode := tmAuto;
      Brush.Color := ModuleColor;
      Pen  .Color := ModuleColor;
    End;

    Destructor  TMimPushButton.Destroy; // Override;
    Begin
      FBitmap.Free;
      Inherited Destroy;
    End;

    Procedure   TMimPushButton.Recolorize; // Override;
    Begin
      Brush.Color := ModuleColor;
      Pen  .Color := ModuleColor;
      Inherited;
    End;

    Procedure   TMimPushButton.LoadBitmap( Const aFileName: String);
    Var
      Bm : TBitmap;
    Begin
      Bm := TBitmap.Create;
      Try
        Bm.LoadFromFile( ApplicationPath + aFilename);
        Bitmap := Bm;
      Finally
        Bm.Free;
      End;
    End;

{ ========
  TIllustration = Class( TNoMouse)
  Private
      FBitmap      : TBitmap;
      FTransparent : Boolean;
      FAutoSize    : Boolean;
  Public
    Property    Bitmap      : TBitmap Read FBitmap      Write SetBitmap;
    Property    Transparent : Boolean Read FTransparent Write SetTransparent;
    Property    AutoSize    : Boolean Read FAutoSize    Write SetAutoSize;
  Private
}

    Procedure   TMimImage.AdjustBounds;
    Var
      R : TRect;
    Begin
      R := ParentRect;
      With R Do
      Begin
        Bottom := Top  + FBitmap.Height;
        Right  := Left + FBitmap.Width;
      End;
      SetSizeAndPos( R);
      Invalidate;
    End;

    Procedure   TMimImage.SetBitmap( aValue: TBitmap);
    Begin
      With FBitmap Do
      Begin
        Assign( aValue);
        TransparentMode := tmAuto;
        Transparent     := FTransparent;
      End;
      If FAutoSize
      Then AdjustBounds
      Else Invalidate;
    End;

    Procedure   TMimImage.SetTransparent( aValue: Boolean);
    Begin
      If FTransparent <> aValue
      Then Begin
        FTransparent := aValue;
        FBitmap.Transparent := FTransparent;
        Invalidate;
      End;
    End;

    Procedure   TMimImage.SetAutoSize( aValue: Boolean);
    Begin
      If FAutoSize <> aValue
      Then Begin
        FAutoSize := aValue;
        If FAutoSize
        Then AdjustBounds;
      End;
    End;

{  Protected }

    Procedure   TMimImage.PaintOn( aCanvas: TCanvas); // Override;
    Begin
      With aCanvas Do
        Draw( 0, 0, FBitmap);
    End;

    procedure   TMimImage.AssignTo( aDest: TPersistent); // Override;
    Begin
      Inherited AssignTo( aDest);
      If Assigned( aDest) And ( aDest Is TMimImage)
      Then Begin
        TMimImage( aDest).Bitmap      .Assign( Bitmap);
        TMimImage( aDest).Transparent := Transparent;
        TMimImage( aDest).AutoSize    := AutoSize;
      End;
    End;

{  Public }

    Constructor TMimImage.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      Width  := 20;
      Height := 20;
      FBitmap := TBitmap.Create;
      FTransparent := True;
      Transparent  := True;
      AutoSize     := False;
    End;

    Destructor  TMimImage.Destroy; // Override;
    Begin
      FBitmap.Free;
      Inherited Destroy;
    End;

    Procedure   TMimImage.LoadBitmap( Const aFileName: String);
    Var
      Bm : TBitmap;
    Begin
      Bm := TBitmap.Create;
      Try
        Bm.LoadFromFile( ApplicationPath + aFilename);
        Bitmap := Bm;
      Finally
        Bm.Free;
      End;
    End;


{ ========
Const

  MaxPoints = 8192;

Type

  TPointArray = Array[ 0 .. MaxPoints - 1] Of TPoint;
  PPointArray = ^TPointArray;

  TGraphSheet = Class( TNoMouse)
  Private
      FData      : PPointArray;
      FEltCount  : Integer;
      FXScale    : Integer;
      FYScale    : Integer;
      FDrawColor : TColor;
  Public
    Property    Data   [ anIndex: Integer]: TPoint  Read GetData    Write SetData;
    Property    XData  [ anIndex: Integer]: Integer Read GetXData   Write SetXData;
    Property    YData  [ anIndex: Integer]: Integer Read GetYData   Write SetYData;
    Property    XScaled[ anIndex: Integer]: Integer Read GetXScaled;
    Property    YScaled[ anIndex: Integer]: Integer Read GetYScaled;
    Property    EltCount                  : Integer Read FEltCount  Write SetEltCount;
    Property    XScale                    : Integer Read FXScale    Write SetXScale;
    Property    YScale                    : Integer Read FYScale    Write SetYScale;
    Property    DrawColor                 : TColor  Read FDrawColor Write SetDrawColor;
  Private
}

    Procedure   TMimGraphSheet.CheckBounds( anIndex: Integer);
    Begin
      If ( anIndex < 0) Or ( anIndex >= EltCount)
      Then Raise EBounds.CreateFmt( 'GraphSheet::Index %d out of bounds [%d,%d]', [ anIndex, 0, EltCount - 1])
    End;

    Function    TMimGraphSheet.GetData( anIndex: Integer): TPoint;
    Begin
      CheckBounds( anIndex);
      Result := FData^[ anIndex];
    End;

    Procedure   TMimGraphSheet.SetData( anIndex: Integer; aValue: TPoint);
    Var
      R : TRect;
    Begin
      CheckBounds( anIndex);
      If ( FData^[ anIndex].x <> aValue.x) Or ( FData^[ anIndex].y <> aValue.y)
      Then Begin
        R :=
          Rect(
            XScaled[ anIndex],
            YScaled[ anIndex],
            XScaled[ anIndex] + 1,
            YScaled[ anIndex] + 1
          );
        FData^[ anIndex] := aValue;
        InvalidateRectangle( R, False);
        InvalidateRectangle(
          Rect(
            XScaled[ anIndex],
            YScaled[ anIndex],
            XScaled[ anIndex] + 1,
            YScaled[ anIndex] + 1
          ),
          False
        );
      End;
    End;

    Function    TMimGraphSheet.GetXData( anIndex: Integer): Integer;
    Begin
      Result := Data[ anIndex].x;
    End;

    Procedure   TMimGraphSheet.SetXData( anIndex: Integer; aValue: Integer);
    Begin
      With Data[ anIndex] Do
        x := aValue;
    End;

    Function    TMimGraphSheet.GetYData( anIndex: Integer): Integer;
    Begin
      Result := Data[ anIndex].y;
    End;

    Procedure   TMimGraphSheet.SetYData( anIndex: Integer; aValue: Integer);
    Begin
      With Data[ anIndex] Do
        y := aValue;
    End;

    Function    TMimGraphSheet.GetXScaled( anIndex: Integer): Integer;
    Begin
      Result := Round( 1 + ( XData[ anIndex] / 1000) * ( XScale / 1000)* ( Width - 2));
    End;

    Function    TMimGraphSheet.GetYScaled( anIndex: Integer): Integer;
    Begin
      Result := Round( 1 - (( YData[ anIndex] - 1000) / 2000) * ( YScale / 1000) * ( Height - 2));
    End;

    Procedure   TMimGraphSheet.SetEltCount ( aValue: Integer);
    Var
      NewData : PMimPointArray;
      i       : Integer;
    Begin
      If FEltCount <> aValue
      Then Begin
        NewData := Nil;
        If aValue > 0
        Then Begin
          GetMem( NewData, aValue * SizeOf( TPoint));
          For i := 0 To aValue - 1 Do
            NewData^[ i] := Point( 0, 0);
        End;
        If Assigned( FData) And ( FEltCount > 0)
        Then Begin
          If Assigned( NewData)
          Then Move( FData^, NewData^, Min( FEltCount, aValue));
          FreeMem( FData, FEltCount * SizeOf( TPoint));
        End;
        FEltCount := aValue;
        FData     := NewData;
      End;
    End;

    Procedure   TMimGraphSheet.SetXScale( aValue: Integer);
    Begin
      If FXScale <> aValue
      Then Begin
        FXScale := aValue;
        Invalidate;
      End;
    End;

    Procedure   TMimGraphSheet.SetYScale( aValue: Integer);
    Begin
      If FYScale <> aValue
      Then Begin
        FYScale := aValue;
        Invalidate;
      End;
    End;

    Procedure   TMimGraphSheet.SetDrawColor( aValue: TColor);
    Begin
      If FDrawColor <> aValue
      Then Begin
        FDrawColor := aValue;
        Invalidate;
      End;
    End;

//   Protected

    Procedure   TMimGraphSheet.PaintOn( aCanvas: TCanvas); // Override;
    Const
      VM = 8;
      HM = 8;
    Var
      i : Integer;
    Begin
      Inherited PaintOn( aCanvas);
      With aCanvas Do
      Begin
//        MoveTo( 0, Height Div 2);
//        LineTo( Width, Height Div 2);
        For i := 1 To VM - 1 Do
        Begin
          MoveTo( i * Width Div VM, 0     );
          LineTo( i * Width Div VM, Height);
        End;
        For i := 1 To HM - 1 Do
        Begin
          MoveTo( 0    , i * Height Div HM);
          LineTo( Width, i * Height Div HM);
        End;
        If EltCount > 0
        Then Begin
          For i := 0 To EltCount - 1 Do
            Pixels[ XScaled[ i], YScaled[ i]] := DrawColor;
        End;
      End;
    End;

    procedure   TMimGraphSheet.AssignTo( aDest: TPersistent); // Override;
    Begin
      Inherited AssignTo( aDest);
      If Assigned( aDest) And ( aDest Is TMimGraphSheet)
      Then Begin
        TMimGraphSheet( aDest).EltCount  := EltCount;
        TMimGraphSheet( aDest).XScale    := XScale;
        TMimGraphSheet( aDest).YScale    := YScale;
        TMimGraphSheet( aDest).DrawColor := DrawColor;
      End;
    End;

//   Public

    Constructor TMimGraphSheet.Create( anOwner: TComponent); // Override;
    Var
      i : Integer;
    Begin
      Inherited Create( anOwner);
      EltCount    := 128;
      XScale      := 1000;
      YScale      := 1000;
      DrawColor   := GraphForePen;
      Brush.Color := GraphBack;
      Pen  .Color := GraphBackPen;
      For i := 0 To EltCount - 1 Do
      Begin
        XData[ i] := Random( 1001);
        YData[ i] := Random( 1001);
      End;
    End;

    Destructor  TMimGraphSheet.Destroy; // Override;
    Begin
      EltCount := 0;
      Inherited Destroy;
    End;

    Procedure   TMimGraphSheet.Dump( Var aFile: TextFile; anInd, aDepth: Integer); // Override;
    Var
      i : Integer;
    Begin
      Inherited Dump( aFile, anInd, aDepth);
      If aDepth > 0
      Then Begin
        WriteLn( aFile, Indent( anInd, 'Data'));
        For i := 0 To EltCount - 1 Do
          WriteLn(
            aFile,
            Indent( anInd + 1,
              Format(
                '( %.5d, %.5d) ( %.5d, %.5d)',
                [
                  XData  [ i],
                  YData  [ i],
                  XScaled[ i],
                  YScaled[ i]
                ]
              )
            )
          );
        WriteLn( aFile, Indent( anInd, 'End data'));
      End;
    End;

    Procedure   TMimGraphSheet.Recolorize; // Override;
    Begin
      DrawColor   := GraphForePen;
      Brush.Color := GraphBack;
      Pen  .Color := GraphBackPen;
      Inherited;
    End;

    Procedure   TMimGraphSheet.Clear;
    Begin
    End;


{ ========
  TIndicator = Class( TNoMouse)
  Private
      FActive   : Boolean;
      FOnColor  : TColor;
      FOffColor : TColor;
    Property    Active   : Boolean Read FActive   Write SetActive;
    Property    OnColor  : TColor  Read FOnColor  Write SetOnColor;
    Property    OffColor : TColor  Read FOffColor Write SetOffColor;
}

    Procedure   TMimLed.SetActive( aValue: Boolean);
    Begin
      If FActive <> aValue
      Then Begin
        FActive := aValue;
        Invalidate;
      End;
    End;

    Procedure   TMimLed.SetOnColor( aValue: TColor);
    Begin
      If FOnColor <> aValue
      Then Begin
        FOnColor := aValue;
        If Active
        Then Invalidate;
      End;
    End;

    Procedure   TMimLed.SetOffColor( aValue: TColor);
    Begin
      If FOffColor <> aValue
      Then Begin
        FOffColor := aValue;
        If Not Active
        Then Invalidate;
      End;
    End;

{  Protected }

    Procedure   TMimLed.PaintOn( aCanvas: TCanvas); // Override;
    Begin
      If Active
      Then Brush.Color := OnColor
      Else Brush.Color := OffColor;
      With aCanvas Do
      Begin
        Pen   := Self.Pen;
        Brush := Self.Brush;
        Rectangle( 0, 0, Width, Height);
      End;
    End;

    procedure   TMimLed.AssignTo( aDest: TPersistent); // Override;
    Begin
      Inherited AssignTo( aDest);
      If Assigned( aDest) And ( aDest Is TMimLed)
      Then Begin
        TMimLed( aDest).OnColor  := OnColor;
        TMimLed( aDest).OffColor := OffColor;
      End;
    End;

{  Public }

    Constructor TMimLed.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      OnColor  := LedColorOn;
      OffColor := LedColorOff;
      Width    := 9;
      Height   := 4;
    End;

    Procedure   TMimLed.Recolorize; // Override;
    Begin
      OnColor  := LedColorOn;
      OffColor := LedColorOff;
      Inherited;
    End;

{ ========
  TCustomModule = Class( TCustomPanel)
  Private
      FSelected   : Boolean;
      FInCreation : Boolean;
      FTextLabel  : TTextLabel;
  Public
    Property    Selected  : Boolean    Read FSelected  Write SetSelected;
    Property    TextLabel : TTextLabel Read FTextLabel Write FTextLabel;
  Private
}

    Procedure   TMimCustomModule.DoMouseDown( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // Override;
    Var
      aDetail : TMimCustomDetail;
    Begin
      If ( GDragged = Nil) And ( Button = mbLeft)
      Then Begin
        aDetail := SearchDetail( Point( X, Y));
        If Assigned( aDetail) And ( aDetail <> Self)
        Then
          With aDetail Do
            DoMouseDown( Sender, Button, Shift, X - Left, Y - Top)
        Else Begin
          GDragged     := Self;
          GLastDragged := Self;
          GAnchor      := Point( X, Y);
          GDistance    := Point( 0, 0);
          BringToFront;
          If ssCtrl In Shift
          Then Selected := Not Selected
          Else
            If Not Selected
            Then Begin
              Selected := True;
              UnselectPeers;
            End;
          BeginMove       ( GAnchor  );
          HandleUpdateMove( GDistance);
        End;
      End
      Else If Button = mbRight
      Then Begin
        aDetail := SearchDetail( Point( X, Y));
        If Assigned( aDetail) And ( aDetail <> Self)
        Then
          With aDetail Do
            DoMouseDown( Sender, Button, Shift, X - Left, Y - Top)
      End;
    End;

    Procedure   TMimCustomModule.DoMouseUp( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); // Override;
    Begin
      If GDragged = Self
      Then Begin
        HandleUpdateMove( GDistance);
        GDistance.x := X - GAnchor.x;
        GDistance.y := Y - GAnchor.y;
        GDragged    := Nil;
        EndMove( GDistance);
        ScrollSelfInView;
        If FInCreation
        Then Begin
          FInCreation := False;
          Invalidate;
        End;
      End;
    End;

    procedure   TMimCustomModule.DoMouseMove( Sender: TObject; Shift: TShiftState; X, Y: Integer); // Override;
    Begin
      If GDragged = Self
      Then Begin
        HandleUpdateMove( GDistance);
        GDistance.x := X - GAnchor.x;
        GDistance.y := Y - GAnchor.y;
        HandleUpdateMove( GDistance);
      End;
    End;

    Procedure   TMimCustomModule.UpdateDetail;
    Begin
      If Assigned( Owner) And ( Owner Is TMimDrawSheet)
      Then TMimDrawSheet( Owner).Update;
    End;

    Procedure   TMimCustomModule.Update;  // Override;
    Begin
      Invalidate;
      If Assigned( Owner) And ( Owner Is TMimDrawSheet)
      Then TMimDrawSheet( Owner).Update;
    End;

    Procedure   TMimCustomModule.InvalidateRectangle( aRect: TRect; IsOpaque: Boolean); // Override;
    Begin
      If Assigned( Owner) And ( Owner Is TMimDrawSheet)
      Then
        TMimDrawSheet( Owner).
          InvalidateRectangleDetail(
            MoveRect( aRect, Point( Left, Top)),
            IsOpaque
          );
    End;

    Procedure   TMimCustomModule.SetSelected( aValue: Boolean);
    Begin
      If aValue <> FSelected
      Then Begin
        FSelected := aValue;
        HandleSelectionChange( FSelected);
      End;
    End;

    Function    TMimCustomModule.GetTitle: String;
    Begin
      Result := TextLabel.Text;
    End;

    Procedure   TMimCustomModule.SetTitle( Const aValue: String);
    Begin
      TextLabel.Text := aValue;
    End;

    Procedure   TMimCustomModule.HandleUpdateMove( aPoint: TPoint);
    Begin
      If Assigned( Owner) And ( Owner Is TMimDrawSheet)
      Then TMimDrawSheet( Owner).HandleModuleUpdateMove( Self, aPoint)
      Else UpdateMove( aPoint);
    End;

    Procedure   TMimCustomModule.UnselectPeers;
    Begin
      If Assigned( Owner) And ( Owner Is TMimDrawSheet)
      Then TMimDrawSheet( Owner).UnselectAllModulesBut( Self);
    End;

    Procedure   TMimCustomModule.BringToFront;
    Begin
      If Assigned( Owner) And ( Owner Is TMimDrawSheet)
      Then TMimDrawSheet( Owner).BringModuleToFront( Self)
    End;

    Procedure   TMimCustomModule.Disconnect( aConnector: TMimConnector); // Virtual;
    Begin
      If Assigned( Owner) And ( Owner Is TMimDrawSheet)
      Then TMimDrawSheet( Owner).Disconnect( aConnector)
    End;

    Procedure   TMimCustomModule.MarkConnections;
    Var
      i : Integer;
    Begin
      If Assigned( Owner) And ( Owner Is TMimDrawSheet)
      Then Begin
        For i := 0 To ComponentCount - 1 Do
          If Components[ i] Is TMimConnector
          Then TMimDrawSheet( Owner).MarkConnections( TMimConnector( Components[ i]));
      End;
    End;

{  Protected }

    Procedure   TMimCustomModule.Notification( aComponent: TComponent; anOperation: TOperation); // Override;
    Begin
      If ( aComponent Is TMimConnector) And (  anOperation = opRemove)
      Then Disconnect( TMimConnector( aComponent));
    End;

    Procedure   TMimCustomModule.HandleSelectionChange( aValue: Boolean); // Virtual;
    Begin
      If Assigned( FTextLabel) And Not ( csDestroying In ComponentState)
      Then
        With FTextLabel Do
          Case aValue Of
            False : Color := ModuleColor;
            True  : Color := clWhite;
          End;
    End;

    Procedure   TMimCustomModule.EndMove( aPoint: TPoint); // Override;
    Begin
      If Assigned( Owner) And ( Owner Is TMimDrawSheet)
      Then TMimDrawSheet( Owner).HandleModuleMoved( Self, aPoint)
      Else Move( aPoint);
    End;

    Procedure   TMimCustomModule.PaintOn( aCanvas: TCanvas); // Override;
    Begin
      If Not FInCreation
      Then Begin
        With aCanvas Do
        Begin
          Pen   := Self.Pen;
          Brush := Self.Brush;
          Rectangle( 0, 0, Width, Height);
          Pen.Color := clWhite;
          MoveTo( 0    , Height);
          LineTo( 0    , 0     );
          LineTo( Width, 0     );
          Pen.Color := clBlack;
        End;
        PaintDetails( aCanvas);
      End;
    End;

    Procedure   TMimCustomModule.PaintDetails( aCanvas: TCanvas); // Virtual;
    Var
      i   : Integer;
      Sdc : Integer;
      Res : Bool;
    Const
      Recur  : Boolean = False;
    Begin
      Assert( Not Recur, 'Oops ... recursion ...');
      Recur := True;
      With aCanvas Do
      Begin
        For i := 0 To ComponentCount - 1 Do
          If Components[ i] Is TMimCustomDetail
          Then
            With TMimCustomDetail( Components[ i]) Do
            Begin
              Sdc := SaveDC( Handle);
              Assert( Sdc <> 0, 'No more save DC ...');
              MoveWindowOrg( Handle, Left, Top);
              IntersectClipRect( Handle, 0, 0, Width, Height);
              Refresh;
              Try
                If RectanglesOverlap( ClientRect, ClipRect)
                Then PaintOn( aCanvas);
              Finally
                Res := RestoreDC( Handle, - 1);
                Assert( Res, 'No more restore DC ...');
              End;
            End;
      End;
      Recur := False;
    End;

    Procedure   TMimCustomModule.UpdateMove( aPoint: TPoint); // Override;
    Var
      aDC   : HDC;
      aRect : TRect;
    Begin
      aRect := ClientRect;
      If Assigned( Owner)
      Then Begin
        aDC := GetDC( 0);
        If Owner Is TMimCustomModule
        Then Begin
          With TMimCustomModule( Owner) Do
          Begin
            With ScreenRect Do
              IntersectClipRect( aDC, Left, Top, Right, Bottom);
            With aRect Do
            Begin
              Left   := ClientOrigin.x + Self.Left + aPoint.x;
              Top    := ClientOrigin.y + Self.Top  + aPoint.y;
              Right  := Left + Self.Width;
              Bottom := Top  + Self.Height;
            End;
          End;
        End
        Else If Owner Is TMimDrawSheet
        Then Begin
          With TMimDrawSheet( Owner) Do
          Begin
            With ClientOrigin Do
              IntersectClipRect( aDC, X, Y, X + ClientWidth, Y + ClientHeight);
            With aRect Do
            Begin
              Left   := ClientOrigin.x + Self.Left + aPoint.x;
              Top    := ClientOrigin.y + Self.Top  + aPoint.y;
              Right  := Left + Self.Width;
              Bottom := Top  + Self.Height;
            End;
          End;
        End;
        Try
          DrawFocusRect( aDC, aRect);
        Finally
          ReleaseDC( 0, aDC);
        End;
      End;
    End;

    Procedure   TMimCustomModule.Move( aPoint: TPoint); // Virtual;
    Begin
      SetSizeAndPos( MoveRect( ParentRect, aPoint));
    End;

    Procedure   TMimCustomModule.CreateDetails; // Virtual;
    Begin
      FTextLabel := TMimTextLabel( CreateDetail( 2, 2, '', TMimTextLabel));
      With FTextLabel Do
      Begin
        AutoSize    := False;
        Width       := 55;
        Height      := 12;
        Text        := Copy( Self.ClassName, 5, Length( ClassName));
        ParentColor := False;
        Color       := ModuleColor;
        FClickable  := True;
      End;
    End;

    Function    TMimCustomModule.CreateDetail( aLeft, aTop: Integer; Const aName: String; aDetailClass: TMimDetailClass): TMimCustomDetail;
    Begin
      Result := aDetailClass.Create( Self);

      If ( Result Is TMimConnector) And ( aName = '')
      Then Raise EInvalidName.Create( 'Connectors must have a non-empty name !');

      With Result Do
      Begin
        Left := aLeft;
        Top  := aTop;
        Name := aName;
      End;
    End;

    Function    TMimCustomModule.SearchDestination( aSource: TMimConnector; aPoint: TPoint): TMimConnector;
    Begin
      Result := FindDestination( aSource, aPoint);
      If ( Result = Nil) And Assigned( Owner) And ( Owner Is TMimDrawSheet)
      Then
        Result :=
          TMimDrawSheet( Owner).
            SearchDestination(
              aSource,
              AddPoints(
                Point(
                  Left,
                  Top
                ),
                aPoint
              )
            );
    End;

    Function    TMimCustomModule.SearchDetail( aPoint: TPoint): TMimCustomDetail;
    Var
      i : Integer;
      C : TMimCustomDetail;
    Begin
      Result := Nil;
      If ContainsPoint( aPoint)
      Then Begin
        For i := ComponentCount - 1 DownTo 0 Do
          If Components[ i] Is TMimCustomDetail
          Then Begin
            C := TMimCustomDetail( Components[ i]);
            With C Do
              If ContainsPoint( AddPoints( aPoint, Point( - Left, - Top)))
              Then Begin
                Result := C;
                Break;
              End;
          End;
        If Not Assigned( Result)
        Then Result := Self;
      End;
    End;

    Function    TMimCustomModule.ContainsPoint( aPoint: TPoint): Boolean; // Override;
    Begin
      Result := PtInRect( ClientRect, aPoint);
    End;

    Function    TMimCustomModule.FindDestination( aSource: TMimConnector; aPoint: TPoint): TMimConnector;
    Var
      i : Integer;
      C : TMimConnector;
    Begin
      Result := Nil;
      If ContainsPoint( aPoint)
      Then
        For i := ComponentCount - 1 Downto 0 Do
          If ( Components[ i] Is TMimConnector) And ( Components[ i] <> aSource)
          Then Begin
            C := TMimConnector( Components[ i]);
            With C Do
              If ContainsPoint( AddPoints( aPoint, Point( - Left, - Top)))
              Then Begin
                Result := C;
                Break;
              End;
          End;
    End;

    Procedure   TMimCustomModule.Connect( aSource, aDestination: TMimConnector);  // Virtual;
    Begin
      If Assigned( Owner) And ( Owner Is TMimDrawSheet)
      Then TMimDrawSheet( Owner).Connect( aSource, aDestination);
    End;

    Function    TMimCustomModule.ClientToScreen( aPoint: TPoint): TPoint; // Override;
    Begin
      Result := aPoint;
      If Assigned( Owner) And Owner.HasParent
      Then
        If Owner Is TMimCustomModule
        Then Result := AddPoints( TMimCustomModule( Owner).ClientToScreen( aPoint), Point( Left, Top))
        Else If Owner Is TMimDrawSheet
        Then Result := AddPoints( TMimDrawSheet( Owner).ClientToScreen( aPoint), Point( Left, Top));
    End;

    Function    TMimCustomModule.ScreenToClient( aPoint: TPoint): TPoint; // Override;
    Begin
      Result := aPoint;
      If Assigned( Owner) And Owner.HasPArent
      Then
        If Owner Is TMimCustomModule
        Then Result := AddPoints( TMimCustomModule( Owner).ScreenToClient( aPoint), Point( - Left, - Top))
        Else If Owner Is TMimDrawSheet
        Then Result := AddPoints( TMimDrawSheet( Owner).ScreenToClient( aPoint), Point( - Left, - Top));
    End;

    Procedure   TMimCustomModule.ScrollSelfInView;
    Begin
      If Assigned( Owner) And ( Owner Is TMimDrawSheet)
      Then TMimDrawSheet( Owner).ScrollModuleInView( Self);
    End;

    procedure   TMimCustomModule.AssignTo( aDest: TPersistent); // Override;
    Var
      i : Integer;
    Begin
      Inherited AssignTo( aDest);
      If Assigned( aDest) And ( aDest Is TMimCustomModule)
      Then Begin
        For i := 0 To ComponentCount - 1 Do
          TMimCustomModule( aDest).Components[ i].Assign( Components[ i]);
        TMimCustomModule( aDest).Selected := Selected;
      End;
    End;

{  Public }

    Constructor TMimCustomModule.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      FInCreation := GAutoCreation;
      DefaultSnap := False;
      FSnap       := Point( 250, 50);
      FSelected   := True;
      Selected    := False;
      Left        := 0;
      Top         := 0;
      Width       := 250;
      Height      := 50;
      Color       := ModuleColor;
      Brush.Color := Color;
      Pen  .Color := clBlack;
      CreateDetails;
    End;

    Destructor  TMimCustomModule.Destroy; // Override;
    Var
      i : Integer;
    Begin
      For i := 0 To ComponentCount - 1 Do
        If Components[ i] Is TMimCustomDetail
        Then Notification( Components[ i], opRemove);
      Inherited Destroy;
    End;

    Procedure   TMimCustomModule.Execute; // Virtual;
    Begin
      // Nothing here, it's for the children.
    End;

    Procedure   TMimCustomModule.Dump( Var aFile: TextFile; anInd, aDepth: Integer); // Override;
    Var
      i : Integer;
    Begin
      Inherited Dump( aFile, anInd, aDepth);
      If aDepth > 0
      Then
        For i := 0 To ComponentCount - 1 Do
          If Components[ i] Is TMimCustomDetail
          Then TMimCustomDetail( Components[ i]).Dump( aFile, anInd + 1, aDepth - 1);
    End;

    Procedure   TMimCustomModule.Recolorize; // Override;
    Var
      i : Integer;
    Begin
      Color       := ModuleColor;
      Brush.Color := Color;
      For i := 0 To ComponentCount - 1 Do
        If Components[ i] Is TMimCustomDetail
        Then TMimCustomDetail( Components[ i]).Recolorize;
    End;

{ ========
  TModule2 = Class( TCustomModule)
  Public
}

    Constructor TMimModule2.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      Height := Height * 2;
    End;

{ ========
  TModule4 = Class( TCustomModule)
  Public
}

    Constructor TMimModule4.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      Height := Height * 4;
    End;

{ ========
  TModule8 = Class( TCustomModule)
  Public
}

    Constructor TMimModule8.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      Height := Height * 8;
    End;

{ ========
  TMidiModule = Class( TCustomModule)
  Private
    FMidiInShort   : TShortMidi;
    FMidiInLong    : TLongMidi;
    FOnMidiInShort : TMidiInShortProc;
    FOnMidiInLong  : TMidiInLongProc;
  Public
    Property    MidiInShort   : TShortMidi       Read FMidiInShort;
    Property    MidiInLong    : TLongMidi        Read FMidiInLong;
    Property    OnMidiInShort : TMidiInShortProc Read FOnMidiInShort Write FOnMidiInShort;
    Property    OnMidiInLong  : TMidiInLongProc  Read FOnMidiInLong  Write FOnMidiInLong;
  Protected
}

    Procedure   TMimMidiModule.AcceptMidiShort( aShortMidi: TShortMidi); // Virtual;
    Begin
      FMidiInShort := aShortMidi;
      If Assigned( FOnMidiInShort)
      Then FOnMidiInShort( Self, FMidiInShort);
    End;

    Procedure   TMimMidiModule.AcceptMidiLong( aLongMidi: TLongMidi); // Virtual;
    Begin
      FMidiInLong := aLongMidi;
      If Assigned( FOnMidiInLong)
      Then FOnMidiInLong( Self, FMidiInLong);
    End;

{  Public }

    Procedure   TMimMidiModule.MidiOutShort( aShortMidi: TShortMidi); // Virtual;
    Begin
      If Owner Is TMimDrawSheet
      Then TMimDrawSheet( Owner).PerformMidiOutShort( Self, aShortMidi);
    End;

    Procedure   TMimMidiModule.MidiOutLong( Const aLongMidi: TLongMidi); // Virtual;
    Begin
      If Owner Is TMimDrawSheet
      Then TMimDrawSheet( Owner).PerformMidiOutLong( Self, aLongMidi);
    End;


{ ========
  TConnection = Class
  Private
      FSource      : TConnector;
      FDestination : TConnector;
      FMarked      : Boolean;
  Public
    Property    Source      : TConnector Read FSource;
    Property    Destination : TConnector Read FDestination;
}

    Constructor TMimConnection.Create( aSource, aDestination: TMimConnector);
    Begin
      Inherited Create;
      FSource      := aSource;
      FDestination := aDestination;
    End;

    Destructor  TMimConnection.Destroy; // Override;
    Begin
      If Assigned( FSource) And ( FSource Is TMimInput)
      Then Source.Value := 0;
      If Assigned( FDestination) And ( FDestination Is TMimInput)
      Then Destination.Value := 0;
      Inherited Destroy;
    End;

{ ========
  TConnections = Class( TList)
  Private
      FOwner : TDrawSheet;
  Public
    Property    ConnectionCount : Integer Read GetConnectionCount;
    Property    Connection [ anIndex: Integer] : TConnection Read GetConnection;
    Property    Source     [ anIndex: Integer] : TConnector Read GetSource;
    Property    Destination[ anIndex: Integer] : TConnector Read GetDestination;
    Property    Owner: TDrawSheet Read FOwner Write FOwner;
  Private
}

    Procedure   TMimConnections.Disconnect( aConnector: TMimConnector);
    Var
      i : Integer;
    Begin
      For i := ConnectionCount - 1 Downto 0 Do
        With Connection[ i] Do
          If ( Source = aConnector) Or ( Destination = aConnector)
          Then Begin
            Source     .DecConnectionCount;
            Destination.DecConnectionCount;
            Connection[ i].Free;
            Delete( i);
            If Assigned( FOwner)
            Then FOwner.ConnectionLost( Source, Destination)
          End;
      Pack;
    End;

    Function    TMimConnections.GetConnectionCount: Integer;
    Begin
      Result := Count;
    End;

    Function    TMimConnections.GetConnection( anIndex: Integer): TMimConnection;
    Begin
      Result := TMimConnection( Items[ anIndex]);
    End;

    Function    TMimConnections.GetSource( anIndex: Integer): TMimConnector;
    Begin
      Result := Connection[ anIndex].Source;
    End;

    Function    TMimConnections.GetDestination( anIndex: Integer): TMimConnector;
    Begin
      Result := Connection[ anIndex].Destination;
    End;

{  Public }

    Function   TMimConnections.Connect( aSource, aDestination: TMimConnector): Boolean;
    Var
      i : Integer;
    Begin
      i := FindConnection( aSource, aDestination);
      If i < 0
      Then i := FindConnection( aDestination, aSource);
      If i >= 0
      Then Begin
        aSource     .DecConnectionCount;
        aDestination.DecConnectionCount;
        Connection[ i].Free;
        Delete( i);
        Pack;
        Result := False;
      End
      Else Begin
        Add( TMimConnection.Create( aSource, aDestination));
        aSource     .IncConnectionCount;
        aDestination.IncConnectionCount;
        Result := True;
      End;
    End;

    Function    TMimConnections.FindConnection( aSource, aDestination: TMimConnector): Integer;
    Var
      i : Integer;
    Begin
      Result := - 1;
      For i := ConnectionCount - 1 DownTo 0 Do
        With Connection[ i] Do
          If ( aSource = Source) And ( aDestination = Destination)
          Then Begin
            Result := i;
            Break;
          End;
    End;

    Procedure   TMimConnections.Execute; // Virtual;
    Var
      i : Integer;
    Begin
      // Clear all input's FCalcCount's and values.
      For i := 0 To ConnectionCount - 1 Do
        If Connection[ i].Source Is TMimInput
        Then
          With Connection[ i].Source Do
          Begin
            FCalcCount := 0;
            FValue     := 0;
          End
        Else If Connection[ i].Destination Is TMimInput
        Then
          With Connection[ i].Destination Do
          Begin
            FCalcCount := 0;
            FValue     := 0;
          End;
      // Transfer outputvalues to input(value)s.
      For i := 0 To ConnectionCount - 1 Do
        With Connection[ i] Do
          If ( Source Is TMimOutput) And ( Destination Is TMimInput)
          Then
            With Destination Do
            Begin
              Inc( FCalcCount);
              FValue := FValue + Source.FValue;
            End
          Else If ( Destination Is TMimOutput) And ( Source Is TMimInput)
          Then
            With Source Do
            Begin
              Inc( FCalcCount);
              FValue := FValue + Destination.FValue;
            End
          Else If ( Destination Is TMimInput) And ( Source Is TMimInput)
          Then
            With Source Do
            Begin
              Inc( FCalcCount);
              FValue := FValue + Destination.FValue;
            End;
      // Then normalize the input values (and notify value changes)
      {
      For i := 0 To ConnectionCount - 1 Do
        If Connection[ i].Source Is TInput
        Then
          With Connection[ i].Source Do
          Begin
            If FCalcCount > 0
            Then Value := FValue Div FCalcCount;
          End
        Else If Connection[ i].Destination Is TInput
        Then
          With Connection[ i].Destination Do
          Begin
            If FCalcCount > 0
            Then Value := FValue Div FCalcCount;
          End;
      }
    End;

{ ========
  TDrawSheet = Class( TGraphicControl)
  Private
      FModules           : TList;
      FConnections       : TConnections;
      FSlot              : TSlot;
      FCanvas            : TCanvas;
      FPen               : TPen;
      FDoDrawConnections : Boolean;
      FConnectionFixups  : TStringList;
  Public
    Property    Pen : TPen Read FPen Write SetPen;
    Property    Module     [ anIndex: Integer]: TCustomModule  Read GetModule;
    Property    Connection [ anIndex: Integer]: TConnection    Read GetConnection;
    Property    Source     [ anIndex: Integer]: TConnector     Read GetSource;
    Property    Destination[ anIndex: Integer]: TConnector     Read GetDestination;
  Private
}

    procedure   TMimDrawSheet.WMPaint( Var aMessage: TWMPaint); // Message WM_PAINT;
    Begin
      If aMessage.DC <> 0
      then begin
        FCanvas.Handle := aMessage.DC;
        try
          Paint;
        finally
          FCanvas.Handle := 0;
        end;
      end;
    End;

    Procedure   TMimDrawSheet.DoMouseDown( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    Var
      aDetail : TMimCustomDetail;
    Begin
      If ( GDragged = Nil) And ( Button = mbLeft)
      Then Begin
        aDetail := SearchDetail( Point( X, Y));
        If Assigned( aDetail)
        Then
          With aDetail Do
            DoMouseDown( Sender, Button, Shift, X - Left, Y - Top)
        Else Begin
          GDragged     := Self;
          GLastDragged := Self;
          GAnchor      := Point( X, Y);
          GDistance    := Point( 0, 0);
          BeginMove ( GAnchor  );
          UpdateMove( GDistance);
        End;
      End
      Else If Button = mbRight
      Then Begin
        aDetail := SearchDetail( Point( X, Y));
        If Assigned( aDetail)
        Then
          With aDetail Do
            DoMouseDown( Sender, Button, Shift, X - Left, Y - Top)
      End
    End;

    Procedure   TMimDrawSheet.DoMouseUp( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    Var
      aPoint: TPoint;
    Begin
      If GDragged = Self
      Then Begin
        UpdateMove( GDistance);
        GDistance.x := X - GAnchor.x;
        GDistance.y := Y - GAnchor.y;
        GDragged := Nil;
        EndMove( GDistance);
      End
      Else
        If Assigned( GDragged) And ( GDragged Is TMimCustomDetail)
        Then
          With TMimCustomDetail( GDragged) Do
          Begin
            aPoint := ScreenToClient( Self.ClientToScreen( Point( X, Y)));
            With aPoint Do
              DoMouseUp( Sender, Button, Shift, X, Y);
          End;
    End;

    procedure   TMimDrawSheet.DoMouseMove( Sender: TObject; Shift: TShiftState; X, Y: Integer);
    Var
      aPoint : TPoint;
    Begin
      If GDragged = Self
      Then Begin
        UpdateMove( GDistance);
        GDistance.x := X - GAnchor.x;
        GDistance.y := Y - GAnchor.y;
        UpdateMove( GDistance);
      End
      Else
        If Assigned( GDragged) And ( GDragged Is TMimCustomDetail)
        Then
          With TMimCustomDetail( GDragged) Do
          Begin
            aPoint := ScreenToClient( Self.ClientToScreen( Point( X, Y)));
            With aPoint Do
              DoMouseMove( Sender, Shift, X, Y);
          End;
    End;

    Procedure   TMimDrawSheet.DoClick( Sender: TObject);
    Begin
      If Assigned( GDragged) And ( GDragged Is TMimCustomDetail)
      Then TMimCustomDetail( GDragged).DoClick( GDragged)
      Else If GDragged = Self
      Then SendClick( Self);
    End;

    Procedure   TMimDrawSheet.DoDblClick( Sender: TObject);
    Begin
      If Assigned( GLastDragged) And ( GLastDragged Is TMimCustomDetail)
      Then TMimCustomDetail( GLastDragged).DoDblClick( GLastDragged)
      Else If GLastDragged = Self
      Then SendDblClick( Self);
{
      GDragged := GLastDragged;
      EndMove( DragAnchor);
      GDragged := Nil;
}
    End;

    Procedure   TMimDrawSheet.SendClick( Sender: TObject);
    Begin
      If Assigned( FSlot)
      Then FSlot.SendClick( Sender);
    End;

    Procedure   TMimDrawSheet.SendDblClick( Sender: TObject);
    Begin
      If Assigned( FSlot)
      Then FSlot.SendDblClick( Sender);
    End;

    Procedure   TMimDrawSheet.SetPen( aValue: TPen);
    Begin
      FPen.Assign( aValue);
      Invalidate;
    End;

    Procedure   TMimDrawSheet.SetSlotActive( aValue: Boolean);
    Begin
      If ( aValue <> FSlotActive) And Assigned( FSlot)
      Then Begin
        FSlot.Active := aValue;
        FSlotActive := aValue;
      End;
    End;

    Function    TMimDrawSheet.GetModule( anIndex: Integer): TMimCustomModule;
    Begin
      If ( anIndex >= 0) And ( anIndex < ModuleCount)
      Then Result := TMimCustomModule( FModules[ anIndex])
      Else Result := Nil;
    End;

    Function    TMimDrawSheet.GetConnection( anIndex: Integer): TMimConnection;
    Begin
      If ( anIndex >= 0) And ( anIndex < ConnectionCount)
      Then Result := FConnections[ anIndex]
      Else Result := Nil;
    End;

    Function    TMimDrawSheet.GetSource( anIndex: Integer): TMimConnector;
    Begin
      If ( anIndex >= 0) And ( anIndex < ConnectionCount)
      Then Result := Connection[ anIndex].Source
      Else Result := Nil;
    End;

    Function    TMimDrawSheet.GetDestination( anIndex: Integer): TMimConnector;
    Begin
      If ( anIndex >= 0) And ( anIndex < ConnectionCount)
      Then Result := Connection[ anIndex].Destination
      Else Result := Nil;
    End;

    Procedure   TMimDrawSheet.PerformCountChange;
    Begin
      If Assigned( FSlot)
      Then FSlot.PerformCountChange;
    End;

    Procedure   TMimDrawSheet.HandleModuleUpdateMove( aModule: TMimCustomModule; aPoint: TPoint);
    Var
      i : Integer;
    Begin
      For i := 0 to ModuleCount - 1 Do
        If Module[ i].Selected
        Then Module[ i].UpdateMove( aPoint);
    End;

    Procedure   TMimDrawSheet.HandleModuleMoved( aModule: TMimCustomModule; aPoint: TPoint);
    Var
      i : Integer;
    Begin
      If IsNearNullPoint( aPoint)
      Then Exit;
      FDoDrawConnections := False;
      For i := 0 to ModuleCount - 1 Do
        If Module[ i].Selected
        Then Module[ i].MarkConnections;
      InvalidateMarkedConnections( True);
      For i := 0 to ModuleCount - 1 Do
        If Module[ i].Selected
        Then Begin
          With Module[ i] Do
          Begin
            MarkConnections;
            Module[ i].Move( aPoint);
          End;
        End;
      FDoDrawConnections := True;
      InvalidateMarkedConnections( False);
      SizeToFit;
    End;

    Procedure   TMimDrawSheet.UnselectAllModulesBut( aModule: TMimCustomModule);
    Var
      i : Integer;
    Begin
      For i := 0 to ModuleCount - 1 Do
        If Module[ i].Selected And ( Module[ i] <> aModule)
        Then Module[ i].Selected := False;
    End;

    Procedure   TMimDrawSheet.Disconnect( aConnector: TMimConnector);
    Begin
      If Assigned( FConnections)
      Then Begin
        FConnections.Disconnect( aConnector);
        Invalidate;
      End;
    End;

    Procedure   TMimDrawSheet.InvalidateMarkedConnections( IsOpaque: Boolean);
    Var
      i : Integer;
      R : TRect;
    Begin
      If HasParent
      Then Begin
        For i := 0 To ConnectionCount - 1 Do
        Begin
          With Connection[ i] Do
            If FMarked
            Then Begin
              UnionRect( R, Source.ScreenRect, Destination.ScreenRect);
              FMarked := False;
            End;
          R := MoveRect( R, ScreenToClient( Point( Left, Top)));
          InvalidateRect( Parent.Handle, @ R, IsOpaque);
        End;
      End;
    End;

    Procedure   TMimDrawSheet.MarkConnections( aConnector: TMimConnector);
    Var
      i : Integer;
    Begin
      For i := 0 To ConnectionCount - 1 Do
        With Connection[ i] Do
          If ( Source = aConnector) Or ( Destination = aConnector)
          Then FMarked := True;
    End;

    Procedure   TMimDrawSheet.ClearConnectionFixups;
    Begin
      If Assigned( FConnectionFixUps)
      Then FConnectionFixups.Clear
      Else FConnectionFixups := TStringList.Create;
    End;

    Procedure   TMimDrawSheet.AddConnectionFixup( Const S: String);
    Begin
      FConnectionFixups.Add( S);
    End;

    Procedure   TMimDrawSheet.FixupConnections;
    Var
      i            : Integer;
      HadException : Boolean;
    Begin
      Try
        HadException := False;
        With FConnectionFixups Do
          For i := 0 To Count - 1 Do
            Try
              CreateConnectionFromFixup( Strings[ i]);
            Except
              On E:ECorruptPatch Do
                If Not HadException
                Then Begin
                  HadException := True;
                  ShowMessage( E.Message);
                End;
            End;
      Finally
        FConnectionFixups.Free;
        FConnectionFixups := Nil;
      End;
    End;

    Procedure   TMimDrawSheet.CreateConnectionFromFixup( Const S: String);
    Const
      Nums : Set Of AnsiChar = [ '0' .. '9'];
      Ascs : Set Of AnsiChar = [ '0' .. '9', 'A' .. 'Z', 'a' .. 'z'];
    Var
      p        : Integer;
      i        : Integer;
      SrcModId : Integer;
      SrcConId : Integer;
      DstModId : Integer;
      DstConId : Integer;
      SrcModNm : String;
      SrcConNm : String;
      DstModNm : String;
      DstConNm : String;
      SrcMod   : TMimCustomModule;
      SrcCon   : TMimConnector;
      DstMod   : TMimCustomModule;
      DstCon   : TMimConnector;
    Begin
      p := 1;
      i := 1;

      If ( Length( S) > 0) And CharInSet( S[ 1], Nums)
      Then Begin
        // Do it the old way.

        While ( p <= Length( S)) And CharInSet( S[ p], Nums) Do
          Inc( p);
        SrcModId := StrToInt( Copy( S, i, p - i));
        If ( p > Length( S)) Or ( S[ p] <> '.')
        Then Exit
        Else Inc( p);

        i := p;
        While ( p <= Length( S)) And CharInSet( S[ p], Nums) Do
          Inc( p);
        SrcConId := StrToInt( Copy( S, i, p - i));
        If ( p > Length( S)) Or ( S[ p] <> ' ')
        Then Exit
        Else Inc( p);

        i := p;
        While ( p <= Length( S)) And CharInSet( S[ p], Nums) Do
          Inc( p);
        DstModId := StrToInt( Copy( S, i, p - i));
        If ( p > Length( S)) Or ( S[ p] <> '.')
        Then Exit
        Else Inc( p);

        i := p;
        While ( p <= Length( S)) And CharInSet( S[ p], Nums) Do
          Inc( p);
        DstConId := StrToInt( Copy( S, i, p - i));

        Try
          SrcMod := Components[ SrcModId] As TMimCustomModule;
          SrcCon := SrcMod.Components[ SrcConId] As TMimConnector;
          DstMod := Components[ DstModId] As TMimCustomModule;
          DstCon := DstMod.Components[ DstConId] As TMimConnector;
          Connect( SrcCon, DstCon);
        Except
          On Exception Do
            Raise
              ECorruptPatch.
                Create(
                  'The (old format) patch is corrupted or incompatible:'^M +
                  'can''t make all the connections, and some connections will probably be wrong.'
                );
        End;
      End

      Else If Length( S) > 0
      Then Begin
        // Do it the new way ....

        While ( p <= Length( S)) And CharInSet( S[ p], Ascs) Do
          Inc( p);
        SrcModNm := Copy( S, i, p - i);
        If ( p > Length( S)) Or ( S[ p] <> '.')
        Then Exit
        Else Inc( p);

        i := p;
        While ( p <= Length( S)) And CharInSet( S[ p], Ascs) Do
          Inc( p);
        SrcConNm := Copy( S, i, p - i);
        If ( p > Length( S)) Or ( S[ p] <> ' ')
        Then Exit
        Else Inc( p);

        i := p;
        While ( p <= Length( S)) And CharInSet( S[ p], Ascs) Do
          Inc( p);
        DstModNm := Copy( S, i, p - i);
        If ( p > Length( S)) Or ( S[ p] <> '.')
        Then Exit
        Else Inc( p);

        i := p;
        While ( p <= Length( S)) And CharInSet( S[ p], Ascs) Do
          Inc( p);
        DstConNm := Copy( S, i, p - i);

        Try
          SrcMod :=        FindComponent( SrcModNm) As TMimCustomModule;
          SrcCon := SrcMod.FindComponent( SrcConNm) As TMimConnector;
          DstMod :=        FindComponent( DstModNm) As TMimCustomModule;
          DstCon := DstMod.FindComponent( DstConNm) As TMimConnector;
          Connect( SrcCon, DstCon);
        Except
          On Exception Do
            Raise
              ECorruptPatch.
                Create(
                  'The (new format) patch is corrupted or incompatible:'^M +
                  'can''t make all the connections.'
                );
        End;

      End
      Else
        Raise
          ECorruptPatch.
            Create(
              'The patch is corrupted or incompatible:'^M +
              '- this can''t be repaired, don''t know what will happen next ... crash ?'^M +
              'The patch format is not recognised ...'
            );
    End;

    Procedure   TMimDrawSheet.SyncModules;
    Var
      i : Integer;
    Begin
      FModules.Clear;
      For i := 0 To ComponentCount - 1 Do
        If Components[ i] Is TMimCustomModule
        Then Begin
          FModules.Add( Components[ i]);
          With TMimCustomModule( Components[ i]) Do
            Name := UniqueName;
        End;
    End;

    Procedure   TMimDrawSheet.PerformMidiOutShort( aSender: TObject; aShortMidi: TShortMidi);
    Begin
      If Assigned( FSlot)
      Then FSlot.PerformMidiOutShort( aSender, aShortMidi);
    End;

    Procedure   TMimDrawSheet.PerformMidiOutLong( aSender: TObject; Const aLongMidi: TLongMidi);
    Begin
      If Assigned( FSlot)
      Then FSlot.PerformMidiOutLong( aSender, aLongMidi);
    End;

{  Protected }

    Procedure   TMimDrawSheet.BeginMove( aPoint: TPoint); // Virtual;
    Begin
    End;

    Procedure   TMimDrawSheet.EndMove( aPoint: TPoint); // Virtual;
    Begin
      SelectModulesInRect(
        Rect(
          DragAnchor.x,
          DragAnchor.y,
          DragAnchor.x + aPoint.x,
          DragAnchor.y + aPoint.y
        )
      );
    End;

    Procedure   TMimDrawSheet.UpdateMove( aPoint: TPoint); // Virtual;
    Var
      aDC   : HDC;
      aRect : TRect;
    Begin
      aDC := GetDC( 0);
      With ClientOrigin Do
        IntersectClipRect( aDC, X, Y, X + ClientWidth, Y + ClientHeight);
      Try
        With aRect Do
        Begin
          Left   := Self.ClientOrigin.x + DragAnchor.x;
          Top    := Self.ClientOrigin.y + DragAnchor.y;
          Right  := Left + aPoint.x;
          Bottom := Top  + aPoint.y;
          If Left > Right
          Then Exchange( Left, Right);
          If Top > Bottom
          Then Exchange( Bottom, Top);
        End;
        DrawFocusRect( aDC, aRect);
      Finally
        ReleaseDC( 0, aDC);
      End;
    End;

    Procedure   TMimDrawSheet.Notification( aComponent: TComponent; anOperation: TOperation); // Override;
    Var
      i : Integer;
    Begin
      If aComponent Is TMimCustomModule
      Then Begin
        If anOperation = opInsert
        Then Begin
          i := IndexOfModule( TMimCustomModule( aComponent));
          If i < 0
          Then FModules.Add( aComponent);
        End
        Else If anOperation = opRemove
        Then Begin
          i := IndexOfModule( TMimCustomModule( aComponent));
          If i >= 0
          Then
            With FModules Do
            Begin
              Delete( i);
              Pack;
            End;
        End;
        PerformCountChange;
        SizeToFit;
      End
      Else Inherited Notification( aComponent, anOperation);
    End;

    Procedure   TMimDrawSheet.BringModuleToFront( aModule: TMimCustomModule);
    Begin
      If Assigned( FModules)
      Then
        With FModules Do
        Begin
          Remove( aModule);
          Pack;
          Add( aModule);
        End;
    End;

    Procedure   TMimDrawSheet.SelectModulesInRect( aRect: TRect);
    Var
      i : Integer;
    Begin
      With aRect Do
      Begin
        If Left > Right
        Then Exchange( Left, Right);
        If Top > Bottom
        Then Exchange( Bottom, Top);
      End;
      For i := 0 To ComponentCount - 1 Do
        If Components[ i] Is TMimCustomModule
        Then
          With TMimCustomModule( Components[ i]) Do
            Selected := RectanglesOverlap( aRect, Rect( Left, Top, Left + Width, Top + Height));
    End;

    Procedure   TMimDrawSheet.InvalidateRectangleDetail( aRect: TRect; IsOpaque: Boolean);
    Var
      R : TRect;
    Begin
      If Assigned( Parent) And Parent.HandleAllocated
      Then Begin
        R := MoveRect( aRect, Point( Left, Top));
        InvalidateRect( Parent.Handle, @ R, IsOpaque);
      End;
    End;

    Function    TMimDrawSheet.SearchDestination( aSource: TMimConnector; aPoint: TPoint): TMimConnector;
    Begin
      Result := FindDestination( aSource, aPoint);
    End;

    Function    TMimDrawSheet.SearchDetail( aPoint: TPoint): TMimCustomDetail;
    Var
      i : Integer;
    Begin
      Result := Nil;
      If ContainsPoint( aPoint)
      Then Begin
        For i := ModuleCount - 1 DownTo 0 Do
          With Module[ i] Do
            If ContainsPoint( AddPoints( aPoint, Point( - Left, - Top)))
            Then Begin
              Result := Module[ i];
              Break;
            End;
      End;
    End;

    Function    TMimDrawSheet.ContainsPoint( aPoint: TPoint): Boolean; // Virtual;
    Begin
      Result := PtInRect( ClientRect, aPoint);
    End;

    Function    TMimDrawSheet.FindDestination( aSource: TMimConnector; aPoint: TPoint): TMimConnector;
    Var
      i : Integer;
    Begin
      Result := Nil;
      If ContainsPoint( aPoint)
      Then
        For i := ModuleCount - 1 DownTo 0 Do
        Begin
          Result :=
            Module[ i].
              FindDestination(
                aSource,
                AddPoints(
                  aPoint,
                  Point(
                    - Module[ i].Left,
                    - Module[ i].Top
                  )
                )
              );
          If Result <> Nil
          Then Break;
        End;
    End;

    Procedure   TMimDrawSheet.Connect( aSource, aDestination: TMimConnector); // Virtual;
    Var
      R : TRect;
    Begin
      If Assigned( FConnections)
      Then Begin
        If Assigned( aDestination)
        Then Begin
          if fconnections.connect( asource, adestination)
          then connectionmade( asource, adestination)
          else connectionlost( asource, adestination);
          if
            unionrect( r, asource.screenrect, adestination.screenrect) and
            hasparent
          then begin
            r := moverect( r, screentoclient( point( left, top)));
            invalidaterect( parent.handle, @ r, true);
          end;
        End
        Else Begin
        End;
      End;
    End;

    Procedure   TMimDrawSheet.Paint; // Virtual;
    Begin
      PaintDetails    ( FCanvas);
      If FDoDrawConnections
      Then PaintConnections( FCanvas);
    End;

    Procedure   TMimDrawSheet.PaintDetails( aCanvas: TCanvas);
    Var
      i   : Integer;
      Sdc : Integer;
      Res : Bool;
    Const
      Recur  : Boolean = False;
    Begin
      Assert( Not Recur, 'Oops ... recursion ...');
      Recur := True;
      Try
        For i := 0 to ModuleCount - 1 Do
          With aCanvas, Module[ i] Do
          Begin
            Sdc := SaveDC( Handle);
            Assert( Sdc <> 0, 'No more save DC ...');
            MoveWindowOrg( Handle, Left, Top);
            IntersectClipRect( Handle, 0, 0, Width, Height);
            Refresh;
            Try
              If RectanglesOverlap( ClientRect, ClipRect)
              Then PaintOn( aCanvas);
            Finally
              Res := RestoreDC( Handle, - 1);
              Assert( Res, 'No more restore DC ...');
            End;
          End;
      Finally
        Recur := False;
      End;
    End;

    Procedure   TMimDrawSheet.PaintConnections( aCanvas: TCanvas);
    Var
      i : Integer;
      A : TPoint;
      B : TPoint;
    Begin
      With aCanvas Do
      Begin
        Pen := Self.Pen;

        ///////////////////////////////////////////////////////////////////////
        //
        // The following assignment eats resources when executed.
        //
        // Pen.Color := clRed;
        //
        ///////////////////////////////////////////////////////////////////////

        For i := 0 To ConnectionCount - 1 Do
        Begin

          /////////////////////////////////////////////////////////////////////
          //
          //   The following lines eat resources when executed ... why ?
          //
          //   Pen.Color := Source[ i].PlugColor;
          //
          //   or
          //
          //   Pen.Color := clRed;
          //
          /////////////////////////////////////////////////////////////////////
          //
          // Lets try something else, this works ... why !
          //   (see actual implementation)
          //
          // Could the following section from the graphics.pas module
          // be the problem ? Or is it a locking problem perhaps ?
          //
          //   procedure TCanvas.PenChanged(APen: TObject);
          //   begin
          //     if csPenValid in State then
          //     begin
          //       Exclude(State, csPenValid);
          //       SelectObject(FHandle, StockPen);
          //     end;
          //   end;
          //
          /////////////////////////////////////////////////////////////////////


          // Actual implementation ////////////////////////////////////////////
          //
          Self.Pen.Color := Source[ i].PlugColor;
          Pen := Self.Pen;
          //
          // End //////////////////////////////////////////////////////////////

          With Source[ i] Do
            A := ClientToScreen( CenterPoint);
          A := ScreenToClient( A);
          With Destination[ i] Do
            B := ClientToScreen( CenterPoint);
          B := ScreenToClient( B);
          If LineIntersectsRectangle( A, B, ClipRect)
          Then Begin
            With A Do
              MoveTo( X, Y);
            With B Do
              LineTo( X, Y);
          End;
        End;
      End;
    End;

    Procedure   TMimDrawSheet.SizeToFit;
    Var
      BR : TPoint;
      i  : Integer;
    Begin
      If Assigned( FSlot)
      Then Begin
        Align := alNone;
        With FSlot Do
          BR := Point( ClientWidth, ClientHeight);
        For i := 0 To ModuleCount - 1 Do
          With Module[ i] Do
          Begin
            If Left + Width > BR.x
            Then BR.x := Left + Width;
            If Top + Height > BR.y
            Then BR.y := Top + Height;
          End;
        With BR Do
        Begin
          Width  := x;
          Height := y;
        End;
      End;
    End;

    Procedure   TMimDrawSheet.ScrollModuleInView( aModule: TMimCustomModule);
    Begin
      If Assigned( FSlot)
      Then FSlot.ScrollModuleInView( aModule);
    End;


    Procedure   TMimDrawSheet.ConnectionMade( aSource, aDestination: TMimConnector); // Virtual;
    Begin
      If Assigned( FSlot)
      Then FSlot.ConnectionMade( aSource, aDestination);
    End;

    Procedure   TMimDrawSheet.ConnectionLost( aSource, aDestination: TMimConnector);  //Virtual;
    Begin
      If Assigned( FSlot)
      Then FSlot.ConnectionLost( aSource, aDestination);
    End;

    ///////////////////////////////////////////////////////////////////////////
    //
    // Reader / Writer / Persistance - stuff
    //
    ///////////////////////////////////////////////////////////////////////////


    Procedure   TMimDrawSheet.GetChildren( aProc: TGetChildProc; Root: TComponent); // Override;
    var
      i       : Integer;
      aModule : TMimCustomModule;
    begin
      For i := 0 To ModuleCount - 1 Do
      Begin
        aModule := Module[ i];
        If aModule.Owner = Root
        Then aProc( aModule);
      End;
    End;

    Procedure   TMimDrawSheet.DefineProperties( aFiler: TFiler); // Override;
    Begin
      Inherited DefineProperties( aFiler);
      aFiler.DefineProperty( 'Connections', ReadConnections, WriteConnections, True);
    End;

    Procedure   TMimDrawSheet.WriteConnections( aWriter: TWriter);
    Var
      i : Integer;
    Begin
      With aWriter Do
      Begin
        WriteListBegin;
        For i := 0 To ConnectionCount - 1 Do
          WriteString(
            Format(
              '%s.%s %s.%s',
              [
                Source     [ i].Owner.Name,
                Source     [ i]      .Name,
                Destination[ i].Owner.Name,
                Destination[ i]      .Name
              ]
            )

{
            Old code, index based

            Format(
              '%d.%d %d.%d',
              [
                Source     [ i].Owner.ComponentIndex,
                Source     [ i]      .ComponentIndex,
                Destination[ i].Owner.ComponentIndex,
                Destination[ i]      .ComponentIndex
              ]
            )
}

          );
        WriteListEnd;
      End;
    End;

    Procedure   TMimDrawSheet.ReadConnections( aReader: TReader);
    Begin
      ClearConnectionFixups;
      With aReader Do
      Begin
        ReadListBegin;
        While Not EndOfList Do
          AddConnectionFixup( ReadString);
        ReadListEnd;
      End;
    End;

    //
    //
    // Reader / Writer / Persistance - stuff
    //
    ///////////////////////////////////////////////////////////////////////////


{  Public }

    Constructor TMimDrawSheet.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      FPen               := TPen.Create;
      ControlStyle       := ControlStyle + [ csReplicatable];
      FCanvas            := TCanvas.Create;
      Align              := alClient;
      FModules           := TList.Create;
      FConnections       := TMimConnections.Create;
      FConnections.Owner := Self;
      OnMouseDown        := DoMouseDown;
      OnMouseUp          := DoMouseUp;
      OnMouseMove        := DoMouseMove;
      OnClick            := DoClick;
      OnDblClick         := DoDblClick;
      FDoDrawConnections := True;
      FSlotActive        := True;
    End;

    Destructor  TMimDrawSheet.Destroy; // Override;
    Begin
      FConnections.Free;
      FModules    .Free;
      FCanvas     .Free;
      FPen        .Free;
      If Assigned( FConnectionFixups)
      Then FConnectionFixups.Free;
      Inherited Destroy;
    End;

    Procedure   TMimDrawSheet.Execute; // Virtual;
    Var
      i : Integer;
    Begin
      If Not ( csDestroying In ComponentState)
      Then Begin
        For i := 0 To ModuleCount - 1 Do
          Module[ i].Execute;
        If Assigned( FConnections)
        Then FConnections.Execute;
      End;
    End;

    Procedure   TMimDrawSheet.Dump( Var aFile: TextFile; anInd, aDepth: Integer); // Virtual;
    Var
      i : Integer;
    Begin
      WriteLn( aFile, Indent( anInd, ClassName));
      If aDepth > 0
      Then Begin
        WriteLn( aFile, Indent( anInd, 'Begin'));
        For i := 0 To ModuleCount - 1 Do
          Module[ i].Dump( aFile, anInd + 1, aDepth - 1);
        WriteLn( aFile, Indent( anInd, 'End'));
      End;
    End;

    Procedure   TMimDrawSheet.Recolorize;
    Var
      i : Integer;
    Begin
      For i := 0 To ModuleCount - 1 Do
        Module[ i].Recolorize;
    End;

    Procedure   TMimDrawSheet.AcceptMidiShort( aShortMidi: TShortMidi);
    Var
      i : Integer;
    Begin
      For i := 0 To ModuleCount - 1 Do
        If Module[ i] Is TMimMidiModule
        Then TMimMidiModule( Module[ i]).AcceptMidiShort( aShortMidi);
    End;

    Procedure   TMimDrawSheet.AcceptMidiLong( Const aLongMidi: TLongMidi);
    Var
      i : Integer;
    Begin
        For i := 0 To ModuleCount - 1 Do
          If Module[ i] Is TMimMidiModule
          Then TMimMidiModule( Module[ i]).AcceptMidiLong( aLongMidi);
    End;

    Function    TMimDrawSheet.CreateModule( aModuleClass: TMimModuleClass): TMimCustomModule;
    Begin
      GAutoCreation := True;
      Try
//        ReleaseCapture;
//        If HasParent
//        Then SetCapture( Parent.Handle);
        Result := aModuleClass.Create( Self);
        With Result Do
        Begin
          GDragged    := Result;
          GAnchor     := CenterPoint;
          GDistance   := Point( 0, 0);
          BringToFront;
          Selected := True;
          UnselectPeers;
          HandleUpdateMove( GDistance);
        End;
      Finally
        GAutoCreation := False;
      End;
    End;

    Procedure   TMimDrawSheet.FreeModule( anIndex: Integer);
    Begin
      AssertParamInBounds(
        anIndex,
        0,
        ModuleCount - 1,
        'TDrawSheet.FreeModule( anIndex: Integer)',
        'anIndex'
      );
      Module[ anIndex].Free;
      Invalidate;
    End;

    Procedure   TMimDrawSheet.FreeModules;
    Var
      i : Integer;
    Begin
      With FModules Do
        For i := Count - 1 Downto 0 Do
          FreeModule( i);
    End;

    Procedure   TMimDrawSheet.FreeSelectedModules;
    Var
      i : Integer;
    Begin
      With FModules Do
        For i := Count - 1 Downto 0 Do
          If Module[ i].Selected
          Then FreeModule( i);
    End;



{
  Procedure DebugSave( aDrawSheet: TDrawSheet; Const aFileName: String);
  Var
    Sb : TStream;
    St : TStream;
  Begin
    aDrawSheet.SyncModules;
    WriteComponentResFile( ApplicationPath + 'tmps.slb', aDrawSheet);
    Sb := TFileStream.Create( ApplicationPath + 'tmps.slb', fmOpenRead);
    St := TFileStream.Create( aFileName, fmCreate);
    Try
      ObjectResourceToText( Sb, St);
    Finally
      St.Free;
      Sb.Free;
      DeleteFile( ApplicationPath + 'tmps.slb');
    End;
  End;
}


    Procedure   TMimDrawSheet.CopyModulesToClipBoard;
    Var
      i          : Integer;
      aDrawSheet : TMimDrawSheet;
      aModule    : TMimCustomModule;
      aSrcTag    : LongInt;
      aDstTag    : LongInt;
    Begin
      aDrawSheet := TMimDrawSheet.Create( Nil);
      Try
        With FModules Do
          For i := 0 To Count - 1 Do
            If Module[ i].Selected
            Then Begin
              aModule := ( TMimModuleClass( Module[ i].ClassType)).Create( aDrawSheet);
              Module[ i].Tag := LongInt( aModule);
              With aModule Do
              Begin
                Assign( Module[ i]);
                Module[ i].MarkConnections;
              End;
            End
            Else Module[ i].Tag := 0;
        For i := 0 To ConnectionCount - 1 Do
          With Connection[ i] Do
            If FMarked
            Then Begin
              aSrcTag := ( Source.     Owner As TMimCustomModule).Tag;
              aDstTag := ( Destination.Owner As TMimCustomModule).Tag;
              If ( aSrcTag <> 0) And ( aDstTag <> 0)
              Then Begin
                With aDrawSheet Do
                Begin
                  FConnections.Connect(
                    TMimCustomModule( aSrcTag).Components[ Self.Source     [ i].ComponentIndex] As TMimConnector,
                    TMimCustomModule( aDstTag).Components[ Self.Destination[ i].ComponentIndex] As TMimConnector
                  );
                End;
              End;
            End;
        ClipBoard.SetComponent( aDrawSheet);
      Finally
        aDrawSheet.Free;
      End;
    End;

    Procedure   TMimDrawSheet.PasteModulesFromClipBoard;
    Var
      aDummy    : TComponent;
      aComponent: TComponent;
      aModule   : TMimCustomModule;
      i         : Integer;
      aSrcTag   : Integer;
      aDstTag   : Integer;
    Begin
      For i := 0 To ModuleCount - 1 Do
        Module[ i].Selected := False;
      aDummy     := TComponent.Create( Application);
      aComponent := ClipBoard.GetComponent( aDummy, aDummy);
      Try
        If Assigned( aComponent) And ( aComponent Is TMimDrawSheet)
        Then Begin
          With TMimDrawSheet( aComponent) Do
          Begin
            FixUpConnections;
            GAutoCreation := True;
            For i := 0 To ModuleCount - 1 Do
            Begin
              aModule := ( TMimModuleClass( Module[ i].ClassType)).Create( Nil);
              With aModule Do
              Begin
                Assign( Module[ i]);
                Selected    := True;
                FInCreation := False;
              End;
              Module[ i].Tag := Integer( aModule);
              Self.InsertComponent( aModule);
            End;
            GAutoCreation := False;
            // Still working with TDrawSheet( aComponent), not with Self ...
            For i := 0 To ConnectionCount - 1 Do
              With Connection[ i] Do
              Begin
                aSrcTag := ( Source.     Owner As TMimCustomModule).Tag;
                aDstTag := ( Destination.Owner As TMimCustomModule).Tag;
                If ( aSrcTag <> 0) And ( aDstTag <> 0)
                Then
                  Self.FConnections.Connect(
                    TMimCustomModule( aSrcTag).Components[ Source     .ComponentIndex] As TMimConnector,
                    TMimCustomModule( aDstTag).Components[ Destination.ComponentIndex] As TMimConnector
                  );
              End;
          End;
        End
        Else ShowMessage( 'Clipboard has no modules.');
      Finally
        aDummy.Free;
        Invalidate;
      End;
    End;

    Procedure   TMimDrawSheet.SelectAllModules;
    Var
      i : Integer;
    Begin
      For i := 0 To ModuleCount - 1 Do
       Module[ i].Selected := True;
    End;

    Procedure   TMimDrawSheet.InvertModuleSelection;
    Var
      i : Integer;
    Begin
      For i := 0 To ModuleCount - 1 Do
       With Module[ i] Do
         Selected := Not Selected;
    End;

    Function    TMimDrawSheet.IndexOfModule( aModule: TMimCustomModule): Integer;
    Begin
      Result := FModules.IndexOf( aModule);
    End;

    Function    TMimDrawSheet.ModuleCount: Integer;
    Begin
      If Assigned( FModules)
      Then Result := FModules.Count
      Else Result := 0;
    End;

    Function    TMimDrawSheet.ConnectionCount: Integer;
    Begin
      If Assigned( FConnections)
      Then Result := FConnections.Count
      Else Result := 0;
    End;

    Function    TMimDrawSheet.DragAnchor: TPoint;
    Begin
      Result := GAnchor;
    End;

    Function    TMimDrawSheet.DragDistance: TPoint;
    Begin
      Result := GDistance;
    End;

    Function    TMimDrawSheet.DragObject: TComponent;
    Begin
      Result := GDragged;
    End;

    Function    TMimDrawSheet.LastDragObject: TComponent;
    Begin
      Result := GLastDragged;
    End;

{ ========
  TSlot = Class( TScrollBox)
  Private
      FOnCountChange      : TNotifyCountChange;
      FPrivateCountChange : TNotifyCountChange;
      FOnConnectionMade   : TNotifyConnectionChange;
      FOnConnectionLost   : TNotifyConnectionChange;
      FDrawSheet          : TDrawSheet;
      FFileName           : String;
  Protected
    Property    OnPrivateCountChange : TNotifyCountChange Read FPrivateCountChange Write FPrivateCountChange;
  Published
    Property    OnCountChange    : TNotifyCountChange      Read FOnCountChange    Write FOnCountChange;
    Property    OnConnectionMade : TNotifyConnectionChange Read FOnConnectionMade Write FOnConnectionMade;
    Property    OnConnectionLost : TNotifyConnectionChange Read FOnConnectionLost Write FOnConnectionLost;
    Property    SlotFileName     : String                  Read FFileName;
  Private
}

    Procedure   TMimSlot.PerformCountChange;
    Begin
      If Not ( csDestroying In ComponentState)
      Then Begin
        If Assigned( FPrivateCountChange)
        Then FPrivateCountChange( Self, ModuleCount);
        If Assigned( FOnCountChange)
        Then FOnCountChange( Self, ModuleCount);
      End;
    End;

    Procedure   TMimSlot.SendClick( Sender: TObject);
    Begin
      PostMessage( Handle, UM_CLICK, 0, LongInt( Sender));
    End;

    Procedure   TMimSlot.SendDblClick( Sender: TObject);
    Begin
      PostMessage( Handle, UM_DBLCLICK, 0, LongInt( Sender));
    End;

    Procedure   TMimSlot.UMClick( Var aMessage: TMessage); // Message UM_CLICK;
    Begin
      If Assigned( OnClick)
      Then OnClick( TObject( aMessage.LParam));
    End;

    Procedure   TMimSlot.UMDblClick( Var aMessage: TMessage); // Message UM_DBLCLICK;
    Begin
      If Assigned( OnDblClick)
      Then OnDblClick( TObject( aMessage.LParam));
    End;

    Procedure   TMimSlot.DoResize( aSender: TObject);
    Begin
      If Assigned( FDrawSheet)
      Then FDrawSheet.SizeToFit;
    End;

    Procedure   TMimSlot.PerformMidiOutShort( aSender: TObject; aShortMidi: TShortMidi);
    Begin
      If Assigned( FOnMidiOutShort) And Active
      Then FOnMidiOutShort( aSender, aShortMidi);
    End;

    Procedure   TMimSlot.PerformMidiOutLong( aSender: TObject; Const aLongMidi: TLongMidi);
    Begin
      If Assigned( FOnMidiOutLong) And Active
      Then FOnMidiOutLong( aSender, aLongMidi);
    End;

    Procedure   TMimSlot.SetActive( aValue: Boolean);
    Begin
      If FActive <> aValue
      Then Begin
        FActive := aValue;
      End;
    End;

{  Protected }

    Procedure   TMimSlot.ConnectionMade( aSource, aDestination: TMimConnector); // Virtual;
    Begin
      If Assigned( FOnConnectionMade)
      Then FOnConnectionMade( Self, aSource, aDestination);
    End;

    Procedure   TMimSlot.ConnectionLost( aSource, aDestination: TMimConnector);  // Virtual;
    Begin
      If Assigned( FOnConnectionLost)
      Then FOnConnectionLost( Self, aSource, aDestination);
    End;

{  Public }

    Constructor TMimSlot.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      BorderStyle := bsNone;
      OnResize    := DoResize;
      FFileName   := '';
      FDrawSheet  := TMimDrawSheet.Create( Self);
      With FDrawSheet Do
      Begin
        Parent := Self;
        FSlot  := Self;
      End;
      Active := True;
    End;

    Procedure   TMimSlot.Execute; // Virtual;
    Begin
      If Assigned( FDrawSheet) And Active
      Then FDrawSheet.Execute;
    End;

    Destructor  TMimSlot.Destroy; // Override;
    Begin
      Inherited Destroy;
    End;

    Procedure   TMimSlot.Recolorize;
    Begin
      If Assigned( FDrawSheet)
      Then FDrawSheet.Recolorize;
    End;

    Procedure   TMimSlot.Save; // Virtual;
    Begin
      SaveAs( SlotFileName);
    End;

    Procedure   TMimSlot.SaveAs( Const aFileName: String); // Virtual;
    Var
      Sb : TStream;
      St : TStream;
    Begin
      FDrawSheet.SyncModules;
      WriteComponentResFile( ApplicationPath + 'tmps.slb', FDrawSheet);
      Sb := TFileStream.Create( ApplicationPath + 'tmps.slb', fmOpenRead);
      St := TFileStream.Create( aFileName, fmCreate);
      Try
        ObjectResourceToText( Sb, St);
      Finally
        St.Free;
        Sb.Free;
        DeleteFile( ApplicationPath + 'tmps.slb');
      End;
      FFileName := aFileName;
    End;

    Procedure   TMimSlot.Load( Const aFileName: String); // Virtual;
    Var
      aDrawSheet: TMimDrawSheet;
      Sb        : TStream;
      St        : TStream;
    Begin
      Sb := TFileStream.Create( ApplicationPath + 'tmpl.slb', fmCreate);
      St := TFileStream.Create( aFileName, fmOpenRead);
      Try
        ObjectTextToResource( St, Sb);
      Finally
        Sb.Free;
        St.Free;
      End;
      Try
        aDrawSheet := TMimDrawSheet.Create( Self);
        ReadComponentResFile( ApplicationPath + 'tmpl.slb', aDrawSheet);
        If Assigned( FDrawSheet)
        Then FDrawSheet.Free;
        FDrawSheet := aDrawSheet;
        With FDrawSheet Do
        Begin
          Parent := Self;
          FSlot  := Self;
          FixupConnections;
          If Assigned( FModules) And ( FModules.Count > 0)
          Then ScrollModuleInView( FModules[ 0]);
          DoResize( Self);
        End;
      Finally
        DeleteFile( ApplicationPath + 'tmpl.slb');
      End;
      FFileName := aFileName;
    End;

    Procedure   TMimSlot.Dump( Var aFile: TextFile; anInd, aDepth: Integer); // Virtual;
    Begin
      WriteLn( aFile, Indent( anInd, 'Slot'));
      If Assigned( FDrawSheet)
      Then FDrawSheet.Dump( aFile, anInd, aDepth)
    End;

    Procedure   TMimSlot.AcceptMidiShort( aShortMidi: TShortMidi);
    Begin
      If Assigned( FDrawSheet) And Active
      Then FDrawSheet.AcceptMidiShort( aShortMidi);
    End;

    Procedure   TMimSlot.AcceptMidiLong( Const aLongMidi: TLongMidi);
    Begin
      If Assigned( FDrawSheet) And Active
      Then FDrawSheet.AcceptMidiLong( aLongMidi);
    End;

    Function    TMimSlot.CreateModule( aModuleClass: TMimModuleClass): TMimCustomModule;
    Begin
      If Assigned( FDrawSheet)
      Then Result := FDrawSheet.CreateModule( aModuleClass)
      Else Result := Nil;
    End;

    Procedure   TMimSlot.FreeModule( anIndex: Integer);
    Begin
      If Assigned( FDrawSheet)
      Then FDrawSheet.FreeModule( anIndex);
    End;

    Procedure   TMimSlot.FreeModules;
    Begin
      If Assigned( FDrawSheet)
      Then FDrawSheet.FreeModules;
    End;

    Procedure   TMimSlot.FreeSelectedModules;
    Begin
      If Assigned( FDrawSheet)
      Then FDrawSheet.FreeSelectedModules;
    End;

    Procedure   TMimSlot.CopyModulesToClipBoard;
    Begin
      If Assigned( FDrawSheet)
      Then FDrawSheet.CopyModulesToClipBoard;
    End;

    Procedure   TMimSlot.PasteModulesFromClipBoard;
    Begin
      If Assigned( FDrawSheet)
      Then FDrawSheet.PasteModulesFromClipBoard;
    End;

    Procedure   TMimSlot.SelectAllModules;
    Begin
      If Assigned( FDrawSheet)
      Then FDrawSheet.SelectAllModules;
    End;

    Procedure   TMimSlot.InvertModuleSelection;
    Begin
      If Assigned( FDrawSheet)
      Then FDrawSheet.InvertModuleSelection;
    End;

    Function    TMimSlot.ModuleCount: Integer;
    Begin
      If Assigned( FDrawSheet)
      Then Result := FDrawSheet.ModuleCount
      Else Result := 0;
    End;

    Procedure   TMimSlot.ScrollModuleInView( aModule: TMimCustomModule);
    var
      aRect: TRect;
    begin
      if Assigned( aModule)
      Then Begin
        aRect := aModule.ClientRect;
        Dec( aRect.Left  , HorzScrollBar.Margin);
        Inc( aRect.Right , HorzScrollBar.Margin);
        Dec( aRect.Top   , VertScrollBar.Margin);
        Inc( aRect.Bottom, VertScrollBar.Margin);
        aRect.TopLeft     := ScreenToClient( aModule.ClientToScreen( aRect.TopLeft    ));
        aRect.BottomRight := ScreenToClient( aModule.ClientToScreen( aRect.BottomRight));
        if aRect.Left < 0
        then
          with HorzScrollBar do
            Position := Position + aRect.Left
        else if aRect.Right > ClientWidth
        then begin
          if aRect.Right - aRect.Left > ClientWidth
          then aRect.Right := aRect.Left + ClientWidth;
          with HorzScrollBar do
            Position := Position + aRect.Right - ClientWidth;
        end;
        if aRect.Top < 0
        then with VertScrollBar do
          Position := Position + aRect.Top
        else if aRect.Bottom > ClientHeight
        then begin
          if aRect.Bottom - aRect.Top > ClientHeight
          then aRect.Bottom := aRect.Top + ClientHeight;
          with VertScrollBar do
            Position := Position + aRect.Bottom - ClientHeight;
        end;
      End;
    End;

    Function    TMimSlot.DragAnchor: TPoint;
    Begin
      Result := GAnchor;
    End;

    Function    TMimSlot.DragDistance: TPoint;
    Begin
      Result := GDistance;
    End;

    Function    TMimSlot.DragObject: TComponent;
    Begin
      Result := GDragged;
    End;

    Function    TMimSlot.LastDragObject: TComponent;
    Begin
      Result := GLastDragged;
    End;



{ ========
  TModulePageControl = Class( TPageControl)
  Private
      FOnCountChange       : TNotifyCountChange;
      FOnModuleCountChange : TNotifyCountChange;
      FOnConnectionMade    : TNotifyConnectionChange;
      FOnConnectionLost    : TNotifyConnectionChange;
      FOnActiveSlotChange  : TNotifyActiveSlotChange;
      FOnMidiOutShort      : TMidiOutShortProc;
      FOnMidiOutLong       : TMidiOutLongProc;
      FOnDetailClick       : TNotifyEvent;
      FOnDetailDblClick    : TNotifyEvent;
      FPageCount           : Integer;
  Public
    Property    ActiveSlot: TSlot  Read GetActiveSlot;
  Published
    Property    OnCountChange       : TNotifyCountChange      Read FOnCountChange       Write FOnCountChange;
    Property    OnModuleCountChange : TNotifyCountChange      Read FOnModuleCountChange Write FOnModuleCountChange;
    Property    OnConnectionMade    : TNotifyConnectionChange Read FOnConnectionMade    Write FOnConnectionMade;
    Property    OnConnectionLost    : TNotifyConnectionChange Read FOnConnectionLost    Write FOnConnectionLost;
    Property    OnActiveSlotChange  : TNotifyActiveSlotChange Read FOnActiveSlotChange  Write FOnActiveSlotChange;
    Property    OnMidiOutShort      : TMidiOutShortProc       Read FOnMidiOutShort      Write FOnMidiOutShort;
    Property    OnMidiOutLong       : TMidiOutLongProc        Read FOnMidiOutLong       Write FOnMidiOutLong;
    Property    OnDetailClick       : TNotifyEvent            Read FOnDetailClick       Write FOnDetailClick;
    Property    OnDetailDblClick    : TNotifyEvent            Read FOnDetailDblClick    Write FOnDetailDblClick;
  Private
}

    Function    TMimModulePageControl.GetActiveSlot: TMimSlot;
    Var
      i : Integer;
    Begin
      Result := Nil;
      If Assigned( ActivePage)
      Then
        With ActivePage Do
        Begin
          For i := 0 To ControlCount - 1 Do
            If Controls[ i] Is TMimSlot
            Then Begin
              Result := TMimSlot( Controls[ i]);
              Break;
            End;
        End;
    End;

    Procedure   TMimModulePageControl.DoPrivateCountChange( Sender: TObject; aCount: Integer);
    // Handles Count changes for Modules
    Begin
      If Sender Is TMimSlot
      Then PerformModuleCountChange( TMimSlot( Sender));
    End;

    Procedure   TMimModulePageControl.DoPrivateConnectionMade( aSender: TObject; aSource, aDestination: TMimConnector);
    Begin
      PerformConnectionMade( aSender, aSource, aDestination);
    End;

    Procedure   TMimModulePageControl.DoPrivateConnectionLost( aSender: TObject; aSource, aDestination: TMimConnector);
    Begin
      PerformConnectionLost( aSender, aSource, aDestination);
    End;

    Procedure   TMimModulePageControl.DoMidiOutShort( aSender: TObject; aShortMidi: TShortMidi);
    Begin
      If Assigned( FOnMidiOutShort) And Not( csDestroying In ComponentState)
      Then FOnMidiOutShort( aSender, aShortMidi);
    End;

    Procedure   TMimModulePageControl.DoMidiOutLong( aSender: TObject; Const aLongMidi: TLongMidi);
    Begin
      If Assigned( FOnMidiOutLong) And Not( csDestroying In ComponentState)
      Then FOnMidiOutLong( aSender, aLongMidi);
    End;

    Procedure   TMimModulePageControl.PerformCountChange;
    // Handles Count changes for Slots
    Begin
      If Assigned( FOnCountChange)
      Then Begin
        If csDestroying In ComponentState
        Then FOnCountChange( Self, 0)
        Else FOnCountChange( Self, PageCount);
      End;
      PerformModuleCountChange( ActiveSlot);
    End;

    Procedure   TMimModulePageControl.PerformModuleCountChange( Sender: TMimSlot);
    // Handles Count changes for Modules
    Begin
      If Assigned( FOnModuleCountChange)
      Then Begin
        If csDestroying In ComponentState
        Then FOnModuleCountChange( Self, 0)
        Else Begin
          If Assigned( Sender)
          Then FOnModuleCountChange( Sender, Sender.ModuleCount)
          Else FOnModuleCountChange( Self  , 0                 );
        End;
      End;
    End;

    Procedure   TMimModulePageControl.PerformConnectionMade( aSender: TObject; aSource, aDestination: TMimConnector);
    Begin
      If Assigned( FOnConnectionMade) And Not ( csDestroying In ComponentState)
      Then FOnConnectionMade( aSender, aSource, aDestination);
    End;

    Procedure   TMimModulePageControl.PerformConnectionLost( aSender: TObject; aSource, aDestination: TMimConnector);
    Begin
      If Assigned( FOnConnectionLost) And Not ( csDestroying In ComponentState)
      Then FOnConnectionLost( aSender, aSource, aDestination);
    End;

    Procedure   TMimModulePageControl.PerformSlotChange( aSender: TObject);
    Begin
      If Assigned( FOnActiveSlotChange)
      Then
        If ActiveSlot = Nil
        Then FOnActiveSlotChange( Nil, '')
        Else FOnActiveSlotChange( ActiveSlot, ActiveSlot.Caption);
    End;

    Procedure   TMimModulePageControl.PerformClick( aSender: TObject);
    Begin
      If Assigned( OnDetailClick)
      Then OnDetailClick( aSender);
    End;

    Procedure   TMimModulePageControl.PerformDblClick( aSender: TObject);
    Begin
      If Assigned( OnDetailDblClick)
      Then OnDetailDblClick( aSender);
    End;

    Procedure   TMimModulePageControl.Lock;
    Begin
      Inc( FLockCount);
    End;

    Procedure   TMimModulePageControl.UnLock;
    Begin
      If FLockCount > 0
      Then Dec( FLockCount);
    End;

{   Protected }

    Procedure   TMimModulePageControl.Change; // Override;
    Begin
      Inherited Change;
      PerformModuleCountChange( ActiveSlot);
    End;

{  Public }

    Constructor TMimModulePageControl.Create( anOwner: TComponent);  // Override;
    Begin
      Inherited Create( anOwner);
      OnChange := PerformSlotChange;
      HotTrack := True;
//      TabPosition := tpBottom;
    End;

    Procedure   TMimModulePageControl.Execute; // Virtual;
    Var
      i : Integer;
      j : Integer;
    Begin
      If FLockCount = 0
      Then
        For i := 0 To PageCount - 1 Do
          With Pages[ i] Do
            For j := 0 To ControlCount - 1 Do
              If Controls[ j] Is TMimSlot
              Then TMimSlot( Controls[ j]).Execute;
    End;

    Procedure   TMimModulePageControl.Recolorize;
    Var
      i : Integer;
      j : Integer;
    Begin
      For i := 0 To PageCount - 1 Do
        With Pages[ i] Do
          For j := 0 To ControlCount - 1 Do
            If Controls[ j] Is TMimSlot
            Then TMimSlot( Controls[ j]).Recolorize;
    End;

    Procedure   TMimModulePageControl.Save; // Virtual;
    Begin
      If Assigned( ActiveSlot)
      Then ActiveSlot.Save;
    End;

    Procedure   TMimModulePageControl.SaveAll; // Virtual;
    Var
      i : Integer;
      j : Integer;
    Begin
      For i := 0 To PageCount - 1 Do
        With Pages[ i] Do
          For j := 0 To ControlCount - 1 Do
            If Controls[ j] Is TMimSlot
            Then TMimSlot( Controls[ j]).Save;
    End;

    Procedure   TMimModulePageControl.SaveAs( Const aFileName: String); // Virtual;
    Begin
      If Assigned( ActiveSlot)
      Then Begin
        ActiveSlot.SaveAs( aFileName);
        ActivePage.Caption := CleanFileName( aFileName);
      End;
    End;

    Procedure   TMimModulePageControl.Load( Const aFileName: String); // Virtual;
    Begin
      Lock;
      Try
        AddSlot( CleanFileName( aFileName));
        If Assigned( ActiveSlot)
        Then Begin
          ActiveSlot.Load( aFileName);
          PerformModuleCountChange( ActiveSlot);
        End;
      Finally
        UnLock;
      End;
    End;

    Procedure   TMimModulePageControl.Dump( Var aFile: TextFile; anInd, aDepth: Integer); // Virtual;
    Var
      i : Integer;
      j : Integer;
    Begin
      WriteLn( aFile, Indent( anInd, ClassName));
      WriteLn( aFile, Indent( anInd, 'Begin'));
      If aDepth > 0
      Then Begin
        For i := 0 To PageCount - 1 Do
          With Pages[ i] Do
            For j := 0 To ControlCount - 1 Do
              If Controls[ j] Is TMimSlot
              Then TMimSlot( Controls[ j]).Dump( aFile, anInd + 1, aDepth - 1);
        WriteLn( aFile, Indent( anInd, 'End'));
      End;
    End;

    Procedure   TMimModulePageControl.AddSlot( Const aName: String);
    Var
      T : TTabSheet;
      S : String;
    Begin
      If aName = ''
      Then Begin
        Inc( FPageCount);
        S := Format( 'Slot%d', [ FPageCount]);
      End
      Else S := aName;
      T := TTabSheet.Create( Self);
      With T Do
      Begin
        Visible     := True;
        Caption     := S;
        PageControl := Self;
        With TMimSlot.Create( T) Do
        Begin
          Parent               := T;
          FFileName            := ApplicationPath + S + DefaultSlotExt;
          Align                := alClient;
          Color                := clBlack;
          BorderStyle          := bsNone;
          OnPrivateCountChange := DoPrivateCountChange;
          OnConnectionMade     := DoPrivateConnectionMade;
          OnConnectionLost     := DoPrivateConnectionLost;
          OnMidiOutShort       := DoMidiOutShort;
          OnMidiOutLong        := DoMidiOutLong;
          OnClick              := PerformClick;
          OnDblClick           := PerformDblClick;
        End;
      End;
      ActivePage := T;
      PerformCountChange;
    End;

    Procedure   TMimModulePageControl.AcceptMidiShort( aShortMidi: TShortMidi);
    Var
      i, j : Integer;
    Begin
      For i := 0 To PageCount - 1 Do
        With Pages[ i] Do
          For j := 0 To ControlCount - 1 Do
            If Controls[ j] Is TMimSlot
            Then TMimSlot( Controls[ j]).AcceptMidiShort( aShortMidi);
    End;

    Procedure   TMimModulePageControl.AcceptMidiLong( Const aLongMidi: TLongMidi);
    Var
      i, j : Integer;
    Begin
      For i := 0 To PageCount - 1 Do
        With Pages[ i] Do
          For j := 0 To ControlCount - 1 Do
            If Controls[ j] Is TMimSlot
            Then TMimSlot( Controls[ j]).AcceptMidiLong( aLongMidi);
    End;

    Procedure   TMimModulePageControl.FreeCurrentSlot;
    Begin
      Lock;
      Try
        If Assigned( ActivePage)
        Then Begin
          ActivePage.Free;
          If PageCount > 0
          Then ActivePage := Pages[ 0];
        End;
      Finally
        UnLock;
      End;
      PerformCountChange;
    End;

    Procedure   TMimModulePageControl.FreeAllSlots;
    Begin
      While Assigned( ActivePage) Do
        FreeCurrentSlot;
    End;

    Function    TMimModulePageControl.CreateModule( aModuleClass: TMimModuleClass): TMimCustomModule;
    Begin
      If Assigned( ActiveSlot)
      Then Begin
        Result := ActiveSlot.CreateModule( aModuleClass);
        With Result Do
        Begin
          Top  := 0;
          Left := 0;
        End;
      End
      Else Result := Nil;
    End;

    Procedure   TMimModulePageControl.FreeModule( anIndex: Integer);
    Begin
      Lock;
      Try
        If Assigned( ActiveSlot)
        Then ActiveSlot.FreeModule( anIndex);
      Finally
        UnLock;
      End;
    End;

    Procedure   TMimModulePageControl.FreeModules;
    Begin
      Lock;
      Try
        If Assigned( ActiveSlot)
        Then ActiveSlot.FreeModules;
      Finally
        UnLock;
      End;
    End;

    Procedure   TMimModulePageControl.FreeSelectedModules;
    Begin
      Lock;
      Try
        If Assigned( ActiveSlot)
        Then ActiveSlot.FreeSelectedModules;
      Finally
        UnLock;
      End;
    End;

    Procedure   TMimModulePageControl.CopyModulesToClipBoard;
    Begin
      If Assigned( ActiveSlot)
      Then ActiveSlot.CopyModulesToClipBoard;
    End;

    Procedure   TMimModulePageControl.PasteModulesFromClipBoard;
    Begin
      Lock;
      Try
        If Assigned( ActiveSlot)
        Then ActiveSlot.PasteModulesFromClipBoard;
      Finally
        UnLock;
      End;
    End;

    Procedure   TMimModulePageControl.SelectAllModules;
    Begin
      If Assigned( ActiveSlot)
      Then ActiveSlot.SelectAllModules;
    End;

    Procedure   TMimModulePageControl.InvertModuleSelection;
    Begin
      If Assigned( ActiveSlot)
      Then ActiveSlot.InvertModuleSelection;
    End;

    Function    TMimModulePageControl.ModuleCount: Integer;
    Begin
      If Assigned( ActiveSlot)
      Then Result := ActiveSlot.ModuleCount
      Else Result := 0;
    End;

    Function    TMimModulePageControl.DragAnchor: TPoint;
    Begin
      Result := GAnchor;
    End;

    Function    TMimModulePageControl.DragDistance: TPoint;
    Begin
      Result := GDistance;
    End;

    Function    TMimModulePageControl.DragObject: TComponent;
    Begin
      Result := GDragged;
    End;

    Function    TMimModulePageControl.LastDragObject: TComponent;
    Begin
      Result := GLastDragged;
    End;

{ ========
  TModuleButton = Class( TSpeedButton)
  Private
      FPageName    : String;
      FModuleClass : TMimModuleClass;
  Public
    Property    PageName   : String       Read FPageName;
    Property    ModuleClass: TMimModuleClass Read FModuleClass;
}

    Procedure   TMimModuleButton.LoadGlyph;
    Var
      B : TBitmap;
    Begin
      B := TBitmap.Create;
      Try
        With ModuleClass Do
          B.LoadFromFile( ApplicationPath + 'bitmaps\' + Copy( ClassName, 5, Length( ClassName)) + '.bmp');
      Except
        // If loading a glyph failed for any reason
        // try loading a default image (which could fail too...)
        On Exception Do
          Try
            B.LoadFromFile( ApplicationPath + 'bitmaps\Module.bmp');
          Except
            On Exception Do
              ; // Ignore exceptions, if we can't load, well that's it ...
          End;
      End;
      Glyph := B;
    End;

    Procedure   TMimModuleButton.Initialize( Const aPageName: String; aModuleClass: TMimModuleClass);
    Begin
      FPageName    := aPageName;
      FModuleClass := aModuleClass;
      LoadGlyph;
    End;


{ ========
  TModuleButtonClick = Procedure( aSender: TObject; aModuleClass: TMimModuleClass) Of Object;

  TmoduleSelector = Class( TPageControl)
  Private
      FRegisteredModules   : TList;
      FOnModuleButtonClick : TModuleButtonClick;
  Published
    Property    OnModuleButtonClick: TModuleButtonClick Read FonModuleButtonClick Write FOnModuleButtonClick;
  Private
}

    Procedure   TMimmoduleSelector.SetButtonsEnabled( aValue: Boolean);
    Var
      i : Integer;
    Begin
      If aValue <> FButtonsEnabled
      Then Begin
        FButtonsEnabled := aValue;
        With FRegisteredModules Do
          For i := 0 To Count - 1 Do
            TMimModuleButton( Items[ i]).Enabled := FButtonsEnabled;
      End;
    End;

    Function    TMimmoduleSelector.AddPage( Const aName: String): TTabSheet;
    Begin
      Result := TTabSheet.Create( Self);
      With Result Do
      Begin
        Visible     := True;
        Caption     := aName;
        PageControl := Self;
      End;
    End;

    Function    TMimmoduleSelector.PageButtonCount( Const aName: String): Integer;
    Var
      i : Integer;
    Begin
      Result := 0;
      If PageExists( aName)
      Then
        With PageByName( aName) Do
          For i := 0 To ControlCount - 1 Do
            If Controls[ i] Is TMimModuleButton
            Then Inc( Result);
    End;

    Function    TMimmoduleSelector.PageExists( Const aName: String): Boolean;
    Begin
      Result := PageByName( aName) <> Nil;
    End;

    Function    TMimmoduleSelector.PageByName( Const aName: String): TTabSheet;
    Var
      i : Integer;
    Begin
      Result := Nil;
      For i := 0 To PageCount - 1 Do
        If Pages[ i].Caption = aName
        Then Begin
          Result := Pages[ i];
          Break;
        End;
    End;

    Function    TMimmoduleSelector.FindModulePage( aModuleClass: TMimModuleClass): TTabSheet;
    Var
      i : Integer;
    Begin
      Result := Nil;
      With FRegisteredModules Do
        For i := 0 To Count - 1 Do
          With TMimModuleButton( Items[ i]) Do
            If ModuleClass = aModuleClass
            Then Begin
              Result := PageByName( PageName);
              Break;
            End;

    End;

    Function    TMimmoduleSelector.FindModuleButton( aModuleClass: TMimModuleClass): TMimModuleButton;
    Var
      aPage : TTabSheet;
      i     : Integer;
    Begin
      aPage := FindModulePage( aModuleClass);
      Result := Nil;
      If Assigned( aPage)
      Then
        With aPage Do
        Begin
          For i := 0 To ControlCount - 1 Do
            If Controls[ i] Is TMimModuleButton
            Then
              With TMimModuleButton( Controls[ i]) Do
                If ModuleClass = aModuleClass
                Then Begin
                  Result := TMimModuleButton( Controls[ i]);
                  Break;
                End;
        End;

    End;

    Procedure   TMimmoduleSelector.DoModuleMouseDown( aSender: TObject; aButton: TMouseButton; aShift: TShiftState; X, Y: Integer);
    Begin
      If Assigned( FOnModuleButtonClick) And ( aSender Is TMimModuleButton)
      Then
        FOnModuleButtonClick(
          TMimModuleButton( aSender),
          TMimModuleButton( aSender).ModuleClass
        );
    End;

    Procedure   TMimmoduleSelector.FormatPage( aPage: TTabSheet);
    Var
      LastLeft : Integer;
      i        : Integer;
    Begin
      Assert(
        Assigned( aPage),
        'TModuleSelecteor.FormatPage : can''t format an unassigned <NIL> page.'
      );
      If PageButtonCount( aPage.Caption) = 0
      Then Begin
        aPage.Free;
        If PageCount > 0
        Then ActivePage := Pages[ 0];
      End
      Else
        With aPage Do
        Begin
          LastLeft := 0;
          For i := 0 To ControlCount - 1 Do
            If Controls[ i] Is TMimModuleButton
            Then
              With TMimModuleButton( Controls[ i]) Do
              Begin
                Left := LastLeft;
                Inc( LastLeft, Width);
              End;
        End;
    End;

{  Protected }

    Procedure   TMimmoduleSelector.Notification( aComponent: TComponent; anOperation: TOperation); // Override;
    Begin
      If aComponent Is TMimModuleButton
      Then Begin
        If anOperation = opInsert
        Then FRegisteredModules.Add   ( aComponent)
        Else FRegisteredModules.Remove( aComponent);
      End;
    End;

{   Public }

    Constructor TMimmoduleSelector.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      FRegisteredModules := TList.Create;
      HotTrack := True;
    End;

    Destructor  TMimmoduleSelector.Destroy; // Override;
    Begin
      Inherited Destroy;
      FRegisteredModules.Free;
    End;

    Procedure   TMimmoduleSelector.RegisterModuleClass( Const aPageName: String; aModuleClass: TMimModuleClass);
    Var
      aPage         : TTabSheet;
      aModuleButton : TMimModuleButton;
    Begin
      aPage := PageByName( aPageName);
      If Not Assigned( aPage)
      Then aPage := AddPage( aPageName);
      ActivePage := aPage;
      ActivePage := Pages[ 0];
      aModuleButton := TMimModuleButton.Create( Self);
      With aModuleButton Do
      Begin
        Parent   := aPage;
        Left     := 0;
        Initialize( aPageName, aModuleClass);
        OnMouseDown := DoModuleMouseDown;
        FormatPage( aPage);
        Top      := 0;
        Flat     := True;
        Hint     := Copy( ModuleClass.ClassName, 5, Length( ModuleClass.ClassName));
        ShowHint := True;
        Enabled  := FButtonsEnabled;
      End;
    End;

    Procedure   TMimmoduleSelector.UnregisterModuleClass( aModuleClass: TMimModuleClass);
    Var
      aModuleButton : TMimModuleButton;
      aPage         : TTabSheet;
    Begin
      aModuleButton := FindModuleButton( aModuleClass);
      aPage         := FindModulePage  ( aModuleClass);
      If Assigned( aModuleButton) And Assigned( aPage)
      Then Begin
        aModuleButton.Free;
        FormatPage( aPage); // This will remove the page if it's empty.
      End;
    End;



Initialization

  RegisterClasses(
    [
      TMimDrawSheet,
      TMimCustomModule,
      TMimModule,
      TMimModule2,
      TMimModule4,
      TMimModule8,
      TMimMidiModule
    ]
  );

Finalization

  UnregisterClasses(
    [
      TMimDrawSheet,
      TMimCustomModule,
      TMimModule,
      TMimModule2,
      TMimModule4,
      TMimModule8,
      TMimMidiModule
    ]
  );

end.


